--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, smack = ...
local smackVersion = GetAddOnMetadata( "Smack", "Version" )
local compilerVersion = GetAddOnMetadata( "Smack_Compiler", "Version" )

local remove = table.remove

function smack.ClearCache()
    SmackCache =
    {
        Version = smackVersion,
        CompilerVersion = compilerVersion
    }
    SmackCache.CreationTime = GetServerTime()

    -- Remember which race/faction the character was when the cache was built,
    -- so that we rebuild it automatically in case of a race/faction change
    -- (since some compilation time conditionals depend on it).
    -- We also remember the class, because even though it's not possible
    -- to change class, it's possible to delete a character of a class
    -- and recreate a new character of the same name and a different class
    -- which would then pick up the saved variables of the former character.
    local _, playerClass = UnitClassBase( "player" )
    local _, _, playerRace = UnitRace( "player" )
    SmackCache.Class = playerClass
    SmackCache.Race = playerRace
    SmackCache.Faction = UnitFactionGroup( "player" )

    if not SmackDB.CacheClearTime then
        SmackDB.CacheClearTime = SmackCache.CreationTime
    end
end

function smack.InitCache()
    SmackCache = SmackCache or {}

    local _, playerClass = UnitClassBase( "player" )
    local _, _, playerRace = UnitRace( "player" )
    local playerFaction = UnitFactionGroup( "player" )

    if SmackCache.Version ~= smackVersion or
        SmackCache.CompilerVersion ~= compilerVersion or
        not SmackCache.CreationTime or
        SmackCache.Class ~= playerClass or
        SmackCache.Race ~= playerRace or
        SmackCache.Faction ~= playerFaction or
        not SmackDB.CacheClearTime or
        SmackDB.CacheClearTime > SmackCache.CreationTime then
        smack.ClearCache()
    end

    -- If there are more cached entries than actual macros, remove the
    -- extraneous ones. The remaining slots might not match the intended
    -- macros, but the macros timestamps should have been updated accordingly,
    -- so they will be recompiled as required.
    while #SmackCache > #SmackDB.Macros do
        remove( SmackCache, #SmackCache )
    end
end

local function GetCompiler()
    if not IsAddOnLoaded( "Smack_Compiler" ) then
        EnableAddOn( "Smack_Compiler" )
        LoadAddOn( "Smack_Compiler" )
    end

    return _G["Smack_Compiler"]
end

local callbacks =
{
    onError = function() end,
    onNotice = function() end,
    onSuspensionPoint = function() end
}

local function compile( timeslicer, source )
    local comp = GetCompiler()
    if not comp then
        return
    end

    callbacks.timeslicer = timeslicer
    return comp.CompileMacro( source, callbacks )
end

function smack.GetCompiledMacro( index, timeslicer )
    local macro = SmackDB.Macros[index]

    if not SmackCache[index] or not SmackCache[index][1] or SmackCache[index][1] < macro.lastModified then
        local compiledMacro = compile( timeslicer, SmackDB.Macros[index][1] )

        if compiledMacro then
            macro.hasErrors = nil
            SmackCache[index] = { GetServerTime(), compiledMacro }
        else
            macro.hasErrors = true
            SmackCache[index] = nil
            return
        end
    end

    return SmackCache[index][2]
end

function smack.SetMacroCachedCode( index, code )
    if code then
        SmackCache[index] = { GetServerTime(), code }
    else
        SmackCache[index] = nil
    end
end

function smack.IsMacroCompiled( index )
    return SmackCache[index] and SmackCache[index][2]
end
