--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Update the values of the insecure conditionals
-- that can change at runtime.
local _, smack = ...

-- Druid forms are a huge pain in the ass: there are many that can be
-- optionally learned (treant, mount form) or talented (moonkin, tree of life)
-- and each optional form shifts the subsequent form indices (as used in
-- form: conditionals) by 1 if learned.
-- So we just compute the index of each available form or set them to nil
-- if they're not available and inject this into the secure environment
-- so that form conditionals have some sane data to work with.
-- We don't do this for bear form, cat form and travel form because they
-- are always available and at the same indices.
local function updateDruidFormIndices()
    local haveMoonkinForm = IsSpellKnown( 24858 ) or IsSpellKnown( 197625 )
    local haveTreantForm = IsSpellKnown( 114282 )
    local haveMountForm = IsSpellKnown( 210053 )
    local haveToLForm = IsSpellKnown( 33891 )

    local index = 4
    if haveMoonkinForm then
        smack.SetGlobalAttribute( "moonkin", index )
        index = index + 1
    else
        smack.SetGlobalAttribute( "moonkin", nil )
    end

    if haveTreantForm then
        smack.SetGlobalAttribute( "treant", index )
        index = index + 1
    else
        smack.SetGlobalAttribute( "treant", nil )
    end

    if haveMountForm then
        smack.SetGlobalAttribute( "mount", index )
        index = index + 1
    else
        smack.SetGlobalAttribute( "mount", nil )
    end

    if haveToLForm then
        smack.SetGlobalAttribute( "tol", index )
    else
        smack.SetGlobalAttribute( "tol", nil )
    end
end

function smack.Events.PLAYER_SPECIALIZATION_CHANGED( unitTarget )
    if unitTarget ~= "player" then
        return
    end

    local specId = GetSpecializationInfo( GetSpecialization() )
    smack.SetGlobalAttribute( "spec", specId )
    smack.ReevaluateEnabledMacros()
end

function smack.Events.SPELLS_CHANGED()
    updateDruidFormIndices()
end

function smack.Events.COVENANT_CHOSEN()
    local cov = C_Covenants.GetActiveCovenantID()
    smack.SetGlobalAttribute( "cov", cov )
end

function smack.InitConditionals()
    local specId = GetSpecializationInfo( GetSpecialization() )
    smack.SetGlobalAttribute( "spec", specId )
    local cov = C_Covenants.GetActiveCovenantID()
    smack.SetGlobalAttribute( "cov", cov )
    updateDruidFormIndices()
end
