--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, smack = ...
local zu = smack.zu

local macroFrames = {}
local macroNames = {}
local macroUnderscoredNames = {}
local macroKeybinds = {}

local Macro = {}
smack.Macro = Macro

smack.MacroCreatedEvent = {}
smack.MacroDeletedEvent = {}
smack.MacroListUpdateEvent = {}

local remove = table.remove

smack.initFinishedEvent = {}

-- A common parent frame for all macro frames, which also routes click to
-- each specific macros by defining them as virtual buttons. This is so that
-- keybind actions to smack macros look nice and readable (like CLICK Smack:macroname)
local MainMacroFrame

-- Hook SetBindng to detect when someone rebind an action we are bound to,
-- so we can update accordingly.
local function onBindingChanged( key )
    local macro = macroKeybinds[key]
    if macro then
        macroKeybinds[key] = nil
        macro.keybind = nil
        smack.Raise( macro.updateEvent )
    end
end

hooksecurefunc( "SetBinding", onBindingChanged )
hooksecurefunc( "SetBindingMacro", onBindingChanged )
hooksecurefunc( "SetBindingSpell", onBindingChanged )
hooksecurefunc( "SetBindingItem", onBindingChanged )
hooksecurefunc( "SetOverrideBinding", onBindingChanged )
hooksecurefunc( "SetOverrideBindingMacro", onBindingChanged )
hooksecurefunc( "SetOverrideBindingSpell", onBindingChanged )
hooksecurefunc( "SetOverrideBindingItem", onBindingChanged )

hooksecurefunc( "SetOverrideBindingClick", function( _, _, key, frame, button )
    if frame == "Smack" and macroUnderscoredNames[button] then
        local macro = macroUnderscoredNames[button]
        macroKeybinds[key] = macro
        macro.keybind = key
        smack.Raise( macro.updateEvent )
    else
        onBindingChanged( key )
    end
end )

function smack.FindMacroFromKey( key )
    return macroKeybinds[key]
end

-- If name is already used by another macro, append or increment a number
-- at the end.
local function MakeNameUnique( name )
    if macroNames[name:lower()] == nil then
        return name
    end

    local index = 0
    for i = #name, 1, -1 do
        index = i
        local c = name:byte( i )
        if c < 48 or c > 57 then
            break
        end
    end

    if index == #name then
        return MakeNameUnique( name .. " 2" )
    end

    local prefix = name:sub( 1, index )
    local num = tonumber( name:sub( index + 1 ) )
    return MakeNameUnique( prefix .. tostring( num + 1 ) )
end

local function updateMacroActivation( self )
    local enabled = smack.SpecialTagsCheckMacroEnable( self.tags )

    if self.properties.enabled == enabled then
        smack.Raise( self.updateEvent )
        return
    end

    self.properties.enabled = enabled
    local index = self.index

    if not enabled then
        ClearOverrideBindings( macroFrames[index].frame )
        smack.ReleaseMacroFrame( macroFrames[index] )
        macroFrames[index] = { macro = self, empty = true }
        SmackCache[index] = nil

        if self.keybind then
            macroKeybinds[self.keybind] = nil
        end
    else
        macroFrames[index] = smack.CreateMacroFrame( self )

        if self.keybind then
            local underscoreName = self.name:gsub( " ", "_" )
            SetBinding( self.keybind )
            SetOverrideBindingClick( macroFrames[index].frame, false, self.keybind, "Smack", underscoreName )
        end
    end

    smack.Raise( self.updateEvent )
end

function smack.ReevaluateEnabledMacros()
    local macros = SmackDB.Macros
    for i=1,#macros,1 do
        local macro = macros[i]
        updateMacroActivation( macro )
    end

    SaveBindings(GetCurrentBindingSet())
end

local function initTagListeners( macro )
    local tae, tre, tue = {}, {}, {}

    smack.AddListener( tae, macro, function() updateMacroActivation( macro ) end )
    smack.AddListener( tre, macro, function() updateMacroActivation( macro ) end )
    smack.AddListener( tue, macro, function() updateMacroActivation( macro ) end )

    local tagMeta = {}
    tagMeta.tagAddedEvent = tae
    tagMeta.tagRemovedEvent = tre
    tagMeta.tagsUpdatedEvent = tue
    setmetatable( macro.tags, { __index = tagMeta } )
end

local function createAllMacros( timeslicer )
    local macros = SmackDB.Macros
    local errors = false

    for i=1,#macros,1 do
        local macro = macros[i]
        macro.enabled = nil

        if not macro.lastModified then
            macro.lastModified = GetServerTime()
        end

        macro.name = MakeNameUnique( macro.name )
        macroNames[macro.name:lower()] = macro

        local underscoreName = macro.name:gsub( " ", "_" )
        macroUnderscoredNames[underscoreName] = macro

        -- Since macros are directly stored as a serialized var,
        -- we can't pollute them with runtime stuff like events.
        -- But we can put that in metatables without affecting the
        -- serialization.
        local mt =
        {
            updateEvent = {},
            index = i,
            properties = { enabled = smack.SpecialTagsCheckMacroEnable( macro.tags ) }
        }
        setmetatable( mt, { __index = Macro } )
        setmetatable( macro, { __index = mt } )

        macro.tags = macro.tags or {}
        initTagListeners( macro )

        if macro.properties.enabled then
            macroFrames[i] = smack.CreateMacroFrame( macro, timeslicer )

            if macro.keybind then
                SetBinding( macro.keybind )
                SetOverrideBindingClick( macroFrames[i].frame, false, macro.keybind, "Smack", underscoreName )
            end
        else
            macroFrames[i] = { macro = macro, empty = true }
            smack.SetMacroCachedCode( i, nil )
        end

        if macro.hasErrors then
            errors = true
        end
    end

    smack.WaitOutOfCombat()
    SaveBindings(GetCurrentBindingSet())

    smack.SortMacros()
    smack.Raise( smack.initFinishedEvent )

    print( "|cff8533ffSmack|cffc299ff: finished initializing macros." )

    if errors then
        print( "|cff8533ffSmack|cffc299ff: |cffff0000warning: some of your macros have compilation errors." )
    end

    print( "|cff8533ffSmack|cffc299ff: |cffffe000use the |cffff8000/smackk|r command to open the gui." )
end

function smack.InitMacros()
    print( "|cff8533ffSmack|cffc299ff: initializing macros..." )

    wipe( macroFrames )
    wipe( macroNames )
    wipe( macroUnderscoredNames )
    wipe( macroKeybinds )

    if not MainMacroFrame then
        MainMacroFrame = CreateFrame( "Button", "Smack", nil, "SecureHandlerBaseTemplate,SecureActionButtonTemplate" )
        MainMacroFrame:SetAttribute( "type", "macro" )
        MainMacroFrame:RegisterForClicks( "AnyUp", "AnyDown" )
        MainMacroFrame:Show()
        smack.MainMacroFrame = MainMacroFrame
    end

    zu.TimeSlice( createAllMacros, 1 )
end

function Macro.Create( name, source, tags )
    local macros = SmackDB.Macros

    name = MakeNameUnique( name )

    local macro =
    {
        source,
        name = name,
        tags = {},
        lastModified = GetServerTime()
    }

    if tags then
        for i = 1, #tags, 1 do
            macro.tags[#macro.tags+1] = tags[i]
        end
    end

    macroNames[name:lower()] = macro
    macroUnderscoredNames[name:gsub( " ", "_" )] = macro

    macros[#macros+1] = macro
    local mt =
    {
        updateEvent = {},
        index = #macros,
        properties = { enabled = smack.SpecialTagsCheckMacroEnable( tags ) }
    }
    setmetatable( mt, { __index = Macro } )
    setmetatable( macro, { __index = mt } )

    initTagListeners( macro )

    if macro.properties.enabled then
        macroFrames[#macros] = smack.CreateMacroFrame( macro )
    else
        macroFrames[#macros] = { macro = macro, empty = true }
    end

    smack.SetMacroCachedCode( macro.index, "" )
    smack.SortMacros()

    smack.Raise( smack.MacroCreatedEvent, macro )

    return macro
end

function Macro:delete()
    local macros = SmackDB.Macros
    local index = self.index

    if macroFrames[index].frame then
        ClearOverrideBindings( macroFrames[index].frame )
    end
    smack.ReleaseMacroFrame( macroFrames[index] )
    remove( macroFrames, index )
    remove( macros, index )
    SmackCache[index] = nil
    macroNames[self.name:lower()] = nil
    macroUnderscoredNames[self.name:gsub( " ", "_" )] = nil

    if self.keybind then
        macroKeybinds[self.keybind] = nil
    end

    -- Update the indices of all macros following the
    -- one that was removed.
    for i=index,#macros,1 do
        local macro = macros[i]
        local mt = getmetatable( macro )
        mt.__index.index = i

        -- Update the changed timestamp to force the
        -- macro to be recompiled on other characters
        -- (because their cache won't match the macro's order anymore)
        local time = GetServerTime()
        macro.lastModified = time
        SmackCache[i-1] = SmackCache[i]
        SmackCache[i] = nil
    end

    smack.Raise( smack.MacroDeletedEvent, self )
end

function Macro:setMetadata( name, keybind )
    local oldName = self.name

    macroNames[oldName:lower()] = nil
    macroUnderscoredNames[oldName:gsub( " ", "_" )] = nil

    name = MakeNameUnique( name )
    local underscoreName = name:gsub( " ", "_" )

    macroNames[name:lower()] = self
    macroUnderscoredNames[ underscoreName ] = self
    self.name = name

    if self.keybind ~= keybind or oldName ~= name then
        if self.keybind and type( self.keybind ) == "string" and macroFrames[self.index].frame then
            ClearOverrideBindings( macroFrames[self.index].frame )
        end

        if self.keybind then
            macroKeybinds[self.keybind] = nil
        end

        self.keybind = keybind

        if keybind then
            SetBinding( keybind )
            SetOverrideBindingClick( macroFrames[self.index].frame, false, keybind, "Smack", underscoreName )
            macroKeybinds[keybind] = self

            SaveBindings(GetCurrentBindingSet())
        end
    end

    if oldName ~= name then
        smack.SortMacros()
        smack.Raise( smack.MacroListUpdateEvent )
    end

    smack.Raise( self.updateEvent )
end

function Macro:setSource( source, compiled )
    self[1] = source
    self.lastModified = GetServerTime()

    if compiled then
        self.hasErrors = nil
        if self.properties.enabled then
            smack.SetMacroCachedCode( self.index, compiled )
        end
    else
        self.hasErrors = true
    end

    smack.Raise( self.updateEvent )
end

function smack.GetMacroCount()
    return #SmackDB.Macros
end

function smack.GetMacro( index )
    return SmackDB.Macros[index]
end

local function dbSortPredicate( a, b )
    return a.name:lower() < b.name:lower()
end

local function cacheSortPredicate( a, b )
    if a.i == nil then
        return false
    elseif b.i == nil then
        return true
    end

    return a.i < b.i
end

local function framesSortPredicate( a, b )
    return a.macro.index < b.macro.index
end

function smack.SortMacros()
    local db = SmackDB.Macros
    local cache = SmackCache

    sort( db, dbSortPredicate )

    -- Since lua fucking sucks in so many ways we have to do this
    -- dumb thing where we store indices in the cache temporarily
    -- to be able to sort it in the same order as the db
    for i = 1, #db, 1 do
        local macro = db[i]

        if not cache[macro.index] then
            cache[macro.index] = {}
        end

        cache[macro.index].i = i

        -- If the macro has changed index, we need to update its timestamp
        -- so it will get recompiled on other chars
        if macro.index ~= i then
            local time = GetServerTime()
            macro.lastModified = time
            cache[macro.index][1] = time
        end

        local mt = getmetatable( macro ).__index
        mt.index = i
    end

    sort( cache, cacheSortPredicate )
    sort( macroFrames, framesSortPredicate )

    for i = 1, #cache, 1 do
        if not cache[i][1] then
            cache[i] = nil
        else
            cache[i].i = nil
        end
    end
end
