--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Deals with defering updates of macros and
-- insecure conditionals until the player is out of combat
local _, smack = ...

local yield = coroutine.yield
local resume = coroutine.resume
local running = coroutine.running

local pendingMacroUpdates = {}
local pendingAttributeUpdates = {}

smack.DeferedUpdatesQueuedEvent = {}
smack.DeferedUpdatesFlushedEvent = {}

local havePendingUpdates = false

local oocWaitingCoroutines = {}

function smack.Events.PLAYER_REGEN_ENABLED()
    for i=1, #oocWaitingCoroutines, 1 do
        resume( oocWaitingCoroutines[i] )
    end
    wipe( oocWaitingCoroutines )

    for attr, val in pairs( pendingAttributeUpdates ) do
        smack.MainMacroFrame:SetAttribute( attr, val )
    end
    wipe( pendingAttributeUpdates )

    for macro, _ in pairs( pendingMacroUpdates ) do
        macro:update()
    end
    wipe( pendingMacroUpdates )

    if havePendingUpdates then
        havePendingUpdates = false
        smack.Raise( smack.DeferedUpdatesFlushedEvent )
    end
end

-- Safely update an attribute on the main smack's frame, by
-- defering it until we're out of combat if necessary.

-- Used to inject the states of insecure conditionals into
-- the secure sandbox.
function smack.SetGlobalAttribute( attr, val )
    if InCombatLockdown() then
        pendingAttributeUpdates[attr] = val
    else
        smack.MainMacroFrame:SetAttribute( attr, val )
    end
end

-- Queue a macro to be updated as soon as we're out of combat.
function smack.QueueMacroForUpdate( macro )
    pendingMacroUpdates[macro] = true

    local shouldRaise = not havePendingUpdates
    havePendingUpdates = true

    if shouldRaise then
        smack.Raise( smack.DeferedUpdatesQueuedEvent )
    end
end

-- This lets a coroutine wait until we are out of combat.
-- Does nothing if called outside of a coroutine.
function smack.WaitOutOfCombat()
    local coro = running()

    if not coro or not InCombatLockdown() then
        return
    end

    local shouldRaise = not havePendingUpdates
    havePendingUpdates = true

    if shouldRaise then
        smack.Raise( smack.DeferedUpdatesQueuedEvent )
    end

    oocWaitingCoroutines[#oocWaitingCoroutines+1] = coro
    yield()
end

function smack.AreUpdatesPending()
    return havePendingUpdates
end
