--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Control flow graph. This is the intermediate
-- representation generated by the parser, on
-- which we then run some optimization and analysis
-- passes before handing it over to the lua code generator.
--
local _, compiler = ...
local CFG = {}
compiler.CFG = CFG

function CFG:new()
    local o = {}

    -- Unique id used to tag blocks already visited during a traversal,
    -- incremented for each new traversal.
    o.traversalId = 0
    o.edges = {}

    setmetatable( o, { __index = self } )
    o.entryPoints = { [1] = CFG.BasicBlock:new( o ) }
    o.currentInsertionPointBB = o.entryPoints[1]

    return o
end

function CFG:isEmpty()
    return self.entryPoints[1]:isEmpty()
end

-- The CFG is a graph of basic blocks.
-- A basic block is a sequence of 0 or more instructions
-- and one terminator instruction.
local BasicBlock = {}

function BasicBlock:new( cfg )
    local o = {}

    o.instructions = {}
    o.cfg = cfg
    o.predecessorCount = 0
    o.dominated = {}

    setmetatable( o, { __index = self } )
    return o
end

function BasicBlock:isEmpty()
    return self.terminator == nil and #self.instructions == 0
end

function BasicBlock:append( instruction )
    self.instructions[#self.instructions + 1] = instruction
end

-- Checks if self dominates b
function BasicBlock:dominates( b )
    return b == self or b.dominator == self or b.dominator and self:dominates( b.dominator )
end

CFG.BasicBlock = BasicBlock

-- There are four types of terminator instructions:
--  * return, which ends the program,
--  * branch, which continues the execution flow in another basic block,
--  * conditional branch, which continues the execution flow in either one
--    of two basic blocks depending on a boolean value,
--  * suspend, which defines a suspension point after which the program
--    will resume at a specified destination basic block.
CFG.TerminatorType =
{
    Return = 1,
    Branch = 2,
    CondBranch = 3,
    Suspend = 4
}

CFG.ExprOpCodes =
{
    Constant = 1,
    GetVar = 2,
    Not = 3,
    Or = 4,
    And = 5,
    StrCat = 6,
    Conditional = 7
}

CFG.OpCodes =
{
    SetVar = 1,
    Print = 2,
    Say = 3,
    Emote = 4,
    Cast = 5,
    Use = 6,
    UseEquippedItem = 7,
    Click = 8,
    CancelAura = 9,
    Target = 10,
    Focus = 11,
    SimpleSlashCmd = 12,
    SummonRandomMount = 13,
    PetAttack = 14,
    Way = 15
}
