--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The lua code generator.
--
local _, compiler = ...
local CodeGen = { opCodeHandlers = {}, exprOpCodeHandlers = {} }
compiler.CodeGen = CodeGen

local concat = table.concat

function CodeGen:new( context )
    local o = {}

    o.context = context
    o.indention = 0
    o.currentStateIndex = 1
    o.luaCode = {}

    setmetatable( o, { __index = self } )
    return o
end

function CodeGen:RegisterOpcodeHandler( opcode, handler )
    self.opCodeHandlers[opcode] = handler
end

function CodeGen:RegisterExprOpcodeHandler( opcode, handler )
    self.exprOpCodeHandlers[opcode] = handler
end

-- Common code gen handlers
CodeGen:RegisterExprOpcodeHandler( compiler.CFG.ExprOpCodes.Constant, function( _, expr )
    return tostring( expr[4] )
end )

-- Lua's operator precedence to be able to decide when to insert parentheses
local LuaPrecedence =
{
    [compiler.CFG.ExprOpCodes.Or] = 1,
    [compiler.CFG.ExprOpCodes.And] = 2,
    [compiler.CFG.ExprOpCodes.Not] = 3,
}

function CodeGen:buildUnaryOp( expr, luaOpName )
    local operand = self:generateExpr( expr[4] )

    local opPrec = LuaPrecedence[expr[4][3]]

    if opPrec and LuaPrecedence[expr[3]] > opPrec then
        return luaOpName  .. " ( " .. operand .. " )"
    end

    return luaOpName .. " " .. operand
end

function CodeGen:buildBinaryOp( expr, luaOpName )
    local lhs = self:generateExpr( expr[4] )
    local rhs = self:generateExpr( expr[5] )

    local lhsPrec = LuaPrecedence[expr[4][3]]
    local rhsPrec = LuaPrecedence[expr[5][3]]

    local s = lhs
    if lhsPrec and LuaPrecedence[expr[3]] > lhsPrec then
        s = "( " .. lhs .. " )"
    end

    s = s .. " " .. luaOpName .. " "

    if rhsPrec and LuaPrecedence[expr[3]] > rhsPrec then
        return s .. "( " .. rhs .. " )"
    end

    return s .. rhs
end

CodeGen:RegisterExprOpcodeHandler( compiler.CFG.ExprOpCodes.StrCat, function( codegen, expr )
    local lhs = codegen:generateStrExpr( expr[4] )
    local rhs = codegen:generateStrExpr( expr[5] )
    return lhs .. ".." .. rhs
end )

local function indent( self )
    local indention = ""
    for i = 1, self.indention do
        indention = indention .. "\t"
    end

    return indention
end

function CodeGen:emit( str )
    local luaCode = self.luaCode
    luaCode[#luaCode+1] = indent( self ) .. str
end

function CodeGen:generateExpr( expr )
    return self.exprOpCodeHandlers[expr[3]]( self, expr )
end

-- Generate an string expression to be inserted into a string.
-- Literal strings are returned as is, string expressions
-- are enclosed in '.. and ..'
function CodeGen:generateInStrExpr( expr )
    if expr[3] == compiler.CFG.ExprOpCodes.Constant then
        return expr[4]
    end

    return "'.."..self:generateExpr( expr ).."..'"
end

-- Generate an string expression to be used as a string.
-- Literal strings are enclosed with ', other
-- expressions are returned as is.
function CodeGen:generateStrExpr( expr )
    if expr[3] == compiler.CFG.ExprOpCodes.Constant
        and type( expr[4] ) == "string" then
        return "'"..expr[4].."'"
    end

    return self:generateExpr( expr )
end

function CodeGen:getContinuationBlock( bb )
    -- If we have an immediately dominated block with more than one predecessor,
    -- this is a continuation block.
    for _, v in pairs( bb.dominated ) do
        if v.predecessorCount > 1 then
            return v
        end
    end
end

-- Check if the provided BB is a dead-end that we don't need to generate:
-- It is if it has:
--   - no instructions
--   - a return terminator and we are in the entry state
--   - a suspend terminator that resumes to the same state as the current state
function CodeGen:isDeadEnd( bb )
    if #bb.instructions > 0 then
        return false
    end

    if bb.terminator[1] == compiler.CFG.TerminatorType.Return
        and self.currentStateIndex == 1 then
        return true
    end

    if bb.terminator[1] == compiler.CFG.TerminatorType.Suspend
        and self.currentStateIndex == bb.terminator[2].stateHeader then
        return true
    end

    return false
end

function CodeGen:generateBasicBlock( cfg, bb, forceEmitReturn )
    self.context.callbacks.timeslicer()

    -- Emit bb instructions
    local instructions = bb.instructions
    for i=1,#instructions,1 do
        local v = instructions[i]
        local handler = self.opCodeHandlers[v[1]]
        self:emit( handler( self, v ) )
    end

    local term = bb.terminator

    if term[1] == compiler.CFG.TerminatorType.Return then
        if #cfg.entryPoints > 1 or forceEmitReturn then
            -- Reset the state to 0 at the end of the program's execution.
            self:emit( "s = 0" )
        end
    elseif term[1] == compiler.CFG.TerminatorType.Suspend then
        if term[2].stateHeader ~= self.currentStateIndex then
            self:emit( "s = " .. term[2].stateHeader )
        end
    elseif term[1] == compiler.CFG.TerminatorType.Branch then
        -- Proceed directly with the code from the destination BB
        -- if we are its only predecessor.
        if term[2].predecessorCount == 1 then
            self:generateBasicBlock( cfg, term[2], forceEmitReturn )
        end
    elseif term[1] == compiler.CFG.TerminatorType.CondBranch then
        local trueBB = term[2]
        local falseBB = term[3]

        local cond

        -- If one of our destination BB has multiple predecessors, it means we need to build the other one first.
        -- If this happens to the trueBB we need to reverse the if.
        if trueBB.predecessorCount > 1 then
            trueBB, falseBB = falseBB, trueBB
            local condExpr = compiler.FoldConstants( { "bool", 0, compiler.CFG.ExprOpCodes.Not, term[4] } )
            cond = self:buildUnaryOp( condExpr, "not" )
        else
            cond = self:generateExpr( term[4] )
        end

        self:emit( "if " .. cond .. " then" )

        -- Generate the then block
        self.indention = self.indention + 1
        self:generateBasicBlock( cfg, trueBB, forceEmitReturn )
        self.indention = self.indention - 1

        local contBB = self:getContinuationBlock( bb )

        -- Only generate the else block if it has one predecessor (us),
        -- unless it is a dead end and we have no continuation.
        if falseBB.predecessorCount == 1 and ( contBB or not self:isDeadEnd( falseBB ) ) then
            self:emit( "else" )
            self.indention = self.indention + 1
            self:generateBasicBlock( cfg, falseBB, forceEmitReturn )
            self.indention = self.indention - 1
        end

        self:emit( "end" )

        if contBB then
            self:generateBasicBlock( cfg, contBB, forceEmitReturn )
        end
    end
end

function CodeGen:generate( cfg, forceEmitReturn )
    if #cfg.entryPoints == 1 then
        self:generateBasicBlock( cfg, cfg.entryPoints[1], forceEmitReturn )
        return self.luaCode
    end

    -- If we have more than one entry point, generate the state dispatcher
    self:emit( "if s == 0 then s = 1 end" )

    local entryPoints = cfg.entryPoints
    for i=1,#entryPoints,1 do
        local bb = entryPoints[i]
        self:emit( ( i == 1 and "if" or "elseif" ) .. " s == " .. i .. " then" )
        self.currentStateIndex = i
        self.indention = 1
        self:generateBasicBlock( cfg, bb, forceEmitReturn )
        self.indention = 0
    end

    self:emit( "end" )
end

function CodeGen:getGeneratedCode()
    return concat( self.luaCode, "\n" )
end
