--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- A dead code elimination pass on the CFG
-- to get rid of empty blocks and dead code
-- due to constant branching conditions.
--
local _, compiler = ...

-- Check if the provided block is empty and terminated by an unconditional branch.
local function IsEmptyBB( bb )
    if not bb.terminator then
        return false
    end

    return #bb.instructions == 0 and bb.terminator[1] == compiler.CFG.TerminatorType.Branch
end

-- Skip any empty blocks with unconditional branches until we find one that actually contains something
local function GetNextNonEmptyBB( cfg, bb )
    if bb.traversalId == cfg.traversalId then
        return bb
    end

    while IsEmptyBB( bb ) do
        bb = bb.terminator[2]
    end

    return bb
end

local function deadCodeElim( timeslicer, cfg, bb )
    -- Skip already traversed blocks
    if bb.traversalId == cfg.traversalId then
        return
    end

    timeslicer()
    bb.traversalId = cfg.traversalId

    if not bb.terminator then
        bb.terminator = { compiler.CFG.TerminatorType.Return }
    end

    if bb.terminator[1] == compiler.CFG.TerminatorType.Suspend then
        -- If we suspend into a resume block that just does a return,
        -- then this is a bogus suspend as the return will suspend anyway,
        -- so remove it.
        local resumeBB = GetNextNonEmptyBB( cfg, bb.terminator[2] )
        if #resumeBB.instructions == 0 and resumeBB.terminator and resumeBB.terminator[1] == compiler.CFG.TerminatorType.Return then
            bb.terminator = { compiler.CFG.TerminatorType.Return }
        else
            deadCodeElim( timeslicer, cfg, bb.terminator[2] )
        end
    elseif bb.terminator[1] == compiler.CFG.TerminatorType.CondBranch then
        -- CondBranch: check if the condition is constant,
        -- and replace it with an unconditional branch if it is.
        local trueBB = bb.terminator[2]
        local falseBB = bb.terminator[3]

        if bb.terminator[4][3] ~= compiler.CFG.ExprOpCodes.Constant then
            deadCodeElim( timeslicer, cfg, trueBB )
            deadCodeElim( timeslicer, cfg, falseBB )
        elseif bb.terminator[4][4] then
            bb.terminator = { compiler.CFG.TerminatorType.Branch, trueBB }
            deadCodeElim( timeslicer, cfg, trueBB )
        else
            bb.terminator = { compiler.CFG.TerminatorType.Branch, falseBB }
            deadCodeElim( timeslicer, cfg, falseBB )
        end
    elseif bb.terminator[2] then
        deadCodeElim( timeslicer, cfg, bb.terminator[2] )
    end

    -- Get rid of any empty blocks
    if bb.terminator[2] then
        bb.terminator[2] = GetNextNonEmptyBB( cfg, bb.terminator[2] )
    end

    if bb.terminator[3] then
        bb.terminator[3] = GetNextNonEmptyBB( cfg, bb.terminator[3] )
    end
end

function compiler.DeadCodeElim( timeslicer, cfg )
    -- We assume this pass will be run before splitting states
    -- so we should have only one entry point
    cfg.entryPoints[1] = GetNextNonEmptyBB( cfg, cfg.entryPoints[1] )
    deadCodeElim( timeslicer, cfg, cfg.entryPoints[1] )
    cfg.traversalId = cfg.traversalId + 1
end
