--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Compute the dominator tree for the CFG
-- by assigning indices to all reachable blocks,
-- generating an edge list, then running the
-- LT algorithm.
--
local _, compiler = ...

local function AddEdge( edges, from, to )
    local e = edges[from]
    e[#e+1] = to
end

local function BuildEdgeList( timeslicer, cfg, edges, blocks, bb )
    -- Skip already traversed blocks
    if bb.traversalId == cfg.traversalId then
        return
    end

    timeslicer()
    bb.traversalId = cfg.traversalId

    blocks[#blocks+1] = bb
    bb.index = #blocks

    edges[bb.index] = {}

    -- Don't traverse suspension points since we want to build the
    -- dom tree only for the current state
    if bb.terminator[1] == compiler.CFG.TerminatorType.Suspend then
        return
    end

    local trueBB = bb.terminator[2]
    local falseBB = bb.terminator[3]

    if trueBB then
        trueBB.predecessorCount = trueBB.predecessorCount + 1
        BuildEdgeList( timeslicer, cfg, edges, blocks, trueBB )
        AddEdge( edges, bb.index, trueBB.index )
    end

    if falseBB then
        falseBB.predecessorCount = falseBB.predecessorCount + 1
        BuildEdgeList( timeslicer, cfg, edges, blocks, falseBB )
        AddEdge( edges, bb.index, falseBB.index )
    end
end

function compiler.ComputeDominators( timeslicer, cfg )
    -- Perform the computation separately for each state's sub graph
    local entryPoints = cfg.entryPoints

    for i=1,#cfg.entryPoints,1 do
        local v = entryPoints[i]
        local blocks = {}
        local edges = {}
        BuildEdgeList( timeslicer, cfg, edges, blocks, v )
        cfg.traversalId = cfg.traversalId + 1

        local idom = compiler.LTComputeDominators( timeslicer, #blocks, edges )

        -- Store the dominator information directly inside the blocks
        for s=1,#idom,1 do
            local d = idom[s]
            if d > 0 then
                blocks[s].dominator = blocks[d]
                blocks[d].dominated[#blocks[d].dominated + 1] = blocks[s]
            end
        end
    end
end
