--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Detect loops occuring inside of a state,
-- which are not allowed, by traversing each state's
-- cfg and looking for branches to a dominator
-- block. Emit errors as necessary.
--
local _, compiler = ...

local function checkIfLooping( context, fromBB, toBB )
    if toBB:dominates( fromBB ) then
        context:error( toBB.loc, "This loop contains a non-suspending code path." )
    end
end

local function CheckLoops( context, cfg, bb )
    -- Skip already traversed blocks
    if bb.traversalId == cfg.traversalId then
        return
    end

    context.callbacks.timeslicer()
    bb.traversalId = cfg.traversalId

    -- Don't traverse suspension points.
    if bb.terminator[1] == compiler.CFG.TerminatorType.Suspend then
        return
    end

    local trueBB = bb.terminator[2]
    local falseBB = bb.terminator[3]

    if trueBB then
        checkIfLooping( context, bb, trueBB )
        CheckLoops( context, cfg, trueBB )
    end

    if falseBB then
        checkIfLooping( context, bb, falseBB )
        CheckLoops( context, cfg, falseBB )
    end
end

function compiler.CheckLoops( context, cfg )
    local entryPoints = cfg.entryPoints

    for i=1,#cfg.entryPoints,1 do
        local v = entryPoints[i]
        CheckLoops( context, cfg, v )
    end

    cfg.traversalId = cfg.traversalId + 1
end
