--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- An implementation of the Lengauer-Tarjan
-- dominator tree computation algorithm.
--
local _, compiler = ...

local function CreateArray( size )
    local a = {}
    a[0] = 0

    for i=1, size do
        a[i] = 0
    end

    return a
end

local function CreateState( n, edges )
    local s = {}

    s.dom = CreateArray( n )
    s.parent = CreateArray( n )
    s.ancestor = CreateArray( n )
    s.child = CreateArray( n )
    s.vertex = CreateArray( n )
    s.label = CreateArray( n )
    s.semi = CreateArray( n )
    s.size = CreateArray( n )

    s.succ = edges
    s.pred = {}
    s.bucket = {}

    s.n = 0
    return s
end

local function DFS( timeslicer, s, v )
    s.n = s.n + 1
    s.semi[v] = s.n
    s.vertex[s.n] = v s.label[v] = v
    s.ancestor[v] = 0 s.child[v] = 0
    s.size[v] = 1

    local succ = s.succ[v] or {}
    for i=1,#succ,1 do
        timeslicer()

        local w = succ[i]
        if s.semi[w] == 0 then
            s.parent[w] = v
            DFS( timeslicer, s, w )
        end

        local spw = s.pred[w] or {}
        spw[#spw+1] = v
        s.pred[w] = spw
    end
end

local function Compress( s, v )
    if s.ancestor[s.ancestor[v]] ~= 0 then
        Compress( s, s.ancestor[v] )

        if s.semi[s.label[s.ancestor[v]]] < s.semi[s.label[v]] then
            s.label[v] = s.label[s.ancestor[v]]
        end

        s.ancestor[v] = s.ancestor[s.ancestor[v]]
    end
end

local function Eval( s, v )
    if s.ancestor[v] == 0 then
        return s.label[v]
    end

    Compress( s, v )
    if s.semi[s.label[s.ancestor[v]]] >= s.semi[s.label[v]] then
        return s.label[v]
    end

    return s.label[s.ancestor[v]]
end

local function Link( timeslicer, st, v, w )
    local s = w

    while st.semi[st.label[w]] < st.semi[st.label[st.child[s]]]  do
        timeslicer()

        if ( st.size[s] + st.size[st.child[st.child[s]]] ) >= 2*st.size[st.child[s]] then
            st.ancestor[st.child[s]] = s
            st.child[s] = st.child[st.child[s]]
        else
            st.size[st.child[s]] = st.size[s]
            st.ancestor[s] = st.child[s]
            s = st.child[s]
        end
    end

    st.label[s] = st.label[w]
    st.size[v] = st.size[v] + st.size[w]

    if st.size[v] < 2*st.size[w] then
        s, st.child[v] = st.child[v], s
    end

    while s ~= 0 do
        st.ancestor[s] = v
        s = st.child[s]
    end
end

function compiler.LTComputeDominators( timeslicer, numBlocks, edges )
    local n = numBlocks
    local s = CreateState( n, edges )

    DFS( timeslicer, s, 1 )

    local i = n
    while i >= 2 do
        local w = s.vertex[i]

        local pred = s.pred[w] or {}
        for j=1,#pred,1 do
            timeslicer()

            local v = pred[j]
            local u = Eval( s, v )

            if s.semi[u] < s.semi[w] then
                s.semi[w] = s.semi[u]
            end
        end

        local sb = s.bucket[s.vertex[s.semi[w]]] or {}
        sb[#sb+1] = w
        s.bucket[s.vertex[s.semi[w]]] = sb

        Link( timeslicer, s, s.parent[w], w )

        local bucket = s.bucket[s.parent[w]] or {}
        for j=1,#bucket,1 do
            timeslicer()

            local v = bucket[j]
            local u = Eval( s, v )

            if s.semi[u] < s.semi[v] then
                s.dom[v] = u
            else
                s.dom[v] = s.parent[w]
            end
        end

        s.bucket[s.parent[w]] = nil
        i = i - 1
    end

    i = 2
    while i <= n do
        timeslicer()

        local w = s.vertex[i]

        if s.dom[w] ~= s.vertex[s.semi[w]] then
            s.dom[w] = s.dom[s.dom[w]]
        end

        i = i + 1
    end

    s.dom[1] = 0
    return s.dom
end
