--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Split the cfg into one separate sub graph for each state.
--
local _, compiler = ...

local function CopyBlocksReachableByState( timeslicer, cfg, bb, blockMap )
    blockMap = blockMap or {}

    -- Skip already traversed blocks, return their previously
    -- generated copy for this state.
    if bb.traversalId == cfg.traversalId then
        return blockMap[bb]
    end

    timeslicer()
    bb.traversalId = cfg.traversalId

    local newTerm = {}
    for k, v in pairs( bb.terminator ) do
        newTerm[k] = v
    end

    local newBB = { predecessorCount = 0, dominated = {}, terminator = newTerm }
    setmetatable( newBB, { __index = bb } )

    blockMap[bb] = newBB

    -- Don't traverse suspension points, we only want the blocks reachable within
    -- the same state
    if newBB.terminator[1] == compiler.CFG.TerminatorType.Suspend then
        return newBB
    end

    local trueBB = newBB.terminator[2]
    local falseBB = newBB.terminator[3]

    if trueBB then
        newBB.terminator[2] = CopyBlocksReachableByState( timeslicer, cfg, trueBB, blockMap )
    end

    if falseBB then
        newBB.terminator[3] = CopyBlocksReachableByState( timeslicer, cfg, falseBB, blockMap )
    end

    return newBB
end

local function CollectStates( context, cfg, bb )
    -- Skip already traversed blocks
    if bb.traversalId == cfg.traversalId then
        return
    end

    context.callbacks.timeslicer()
    bb.traversalId = cfg.traversalId

    if bb.terminator[1] == compiler.CFG.TerminatorType.Suspend then
        local resumeBB = bb.terminator[2]

        -- Notify user code of the presence of a suspension point,
        -- so the editor can decorate the line it's on
        context:suspensionPoint( bb.terminator.loc )

        -- Only record the state entry point if we didn't already do so,
        -- Since after dce we may have multiple suspension
        -- points resuming to the same state (which is much better
        -- than having redundant states with duplicate content)
        if resumeBB.stateHeader then
            return
        end

        resumeBB.stateHeader = #cfg.entryPoints + 1
        cfg.entryPoints[#cfg.entryPoints + 1] = resumeBB
        CollectStates( context, cfg, resumeBB )
    else
        local trueBB = bb.terminator[2]
        local falseBB = bb.terminator[3]

        if trueBB then
            CollectStates( context, cfg, trueBB )
        end

        if falseBB then
            CollectStates( context, cfg, falseBB )
        end
    end
end

function compiler.SplitStates( context, cfg )
    local entryPoints = cfg.entryPoints

    -- Collect all the states entry points
    CollectStates( context, cfg, entryPoints[1] )
    cfg.traversalId = cfg.traversalId + 1

    -- If we have only one state we have nothing
    -- to do.
    if #entryPoints == 1 then
        return
    end

    -- Build a separate graph for each entry point.
    for i=1,#entryPoints,1 do
        local v = entryPoints[i]
        entryPoints[i] = CopyBlocksReachableByState( context.callbacks.timeslicer, cfg, v )
        cfg.traversalId = cfg.traversalId + 1
    end
end
