--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, compiler = ...

local sort = table.sort

local Context = {}
compiler.Context = Context

function Context:new( callbacks )
    local o = {}

    o.callbacks = callbacks
    o.locs = compiler.Locations:new()

    setmetatable( o, { __index = self } )
    return o
end

-- This disables all further error reporting,
-- for situations where we know the parsing is going
-- to be broken by previous errors.
function Context:poison()
    self.poisoned = true
    self.failed = true
end

function Context:getLoc( loc )
    return unpack( self.locs[loc] )
end

function Context:error( loc, message )
    if self.poisoned then
        return
    end

    self.failed = true

    if loc ~= nil then
        self.callbacks.onError( message, unpack( self.locs[loc] ) )
    end
end

function Context:notice( loc, message )
    if loc ~= nil then
        self.callbacks.onNotice( message, unpack( self.locs[loc] ) )
    end
end

function Context:suspensionPoint( loc )
    if loc ~= nil then
        self.callbacks.onSuspensionPoint( unpack( self.locs[loc] ) )
    end
end

function Context:completion( provider, line, column )
    self.completionProvider, self.completionLine, self.completionColumn = provider, line, column
end

-- Suggestions can be emitted in additions to completions (to suggest things that can be appended
-- directly after the current word, like the . operator), but only to add to the list. The completion
-- position must be kept if it already exists.
function Context:suggestion( provider, line, column )
    self.suggestionProvider = provider
    if not self.completionLine then
        self.completionLine, self.completionColumn = line, column
    end
end

-- Suggestions can be emitted in additions to completions (to suggest things that can be appended
-- directly after the current word, like the . operator), but only to add to the list. The completion
-- position must be kept if it already exists.
function Context:help( text, line, column )
    self.helpText = text
    if not self.completionLine then
        self.completionLine, self.completionColumn = line, column
    end
end

-- Completion ordering: we group them by groupId (basically we group together things
-- that would get the same syntax colorisation), then sort alphabetically among those.
local function completionOrderingFunc( a, b )
    local typeA = type( a.groupId )
    local typeB = type( b.groupId )

    if typeA == "number" and typeB ~= "number" then
        return true
    end

    if typeA ~= "number" and typeB == "number" then
        return false
    end

    if a.groupId < b.groupId then
        return true
    end

    if a.groupId > b.groupId then
        return false
    end

    return a.name < b.name
end

function Context:emitPendingCompletions()
    if self.callbacks.onHelp and self.helpText then
        self.callbacks.onHelp( self.helpText, self.completionLine, self.completionColumn )
        return
    end

    if self.callbacks.onCompletion and self.completionProvider then
        sort( self.completionProvider, completionOrderingFunc )
        self.callbacks.onCompletion( self.completionProvider, self.completionLine, self.completionColumn )
    end

    if self.callbacks.onSuggestion and self.suggestionProvider then
        sort( self.suggestionProvider, completionOrderingFunc )
        self.callbacks.onSuggestion( self.suggestionProvider, self.completionLine, self.completionColumn )
    end
end
