--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, compiler = ...
Smack_Compiler = compiler

local concat = table.concat

local colorizerStateMeta =
{
    -- We don't compare currentIndention as this have no impact
    -- on the next line's coloring or indentation
    __eq = function( a, b )
        return
            a[1] == b[1]
            and a[3] == b[3]
            and a[4] == b[4]
            and a[5] == b[5]
    end
}

function compiler.Colorize( parserClass, sourceLine, state, palette )
    if not state then
        state = { 0, true, false, 0, 0 }  -- Comment nesting level, isEmpty, inside block, indentationHint, currentIndention
        setmetatable( state, colorizerStateMeta )
    end

    local insideBlock, indentationHint, currentIndention = state[3], 0, state[5]

    local context = compiler.Context:new()
    local tokenizer = compiler.Tokenizer:new( sourceLine, context, true )
    tokenizer.blockCommentDepth = state[1]

    -- Count the number of whitespaces at the line start, so we can cache
    -- it in the state for the benefit of the auto indenter
    for i=1,#sourceLine,1 do
        if sourceLine:byte( i ) ~= 32 then
            currentIndention = i - 1
            tokenizer.bytepos = i
            break
        end
    end

    local tok = tokenizer:readToken()
    local result = {}
    local lastConsumedIndex = 1

    local empty = tok == nil and true or false

    -- If we start within a comment from the previous line state,
    -- emit the comment color directly
    if state[1] > 0 then
        result[#result+1] = palette[compiler.SyntaxColor.Comment]
    end

    while tok do
        local _, fc, _, lc = context:getLoc( tok[2] )

        -- Get everything verbatim from the original string up to the current token's position
        if fc > lastConsumedIndex then
            result[#result+1] = sourceLine:sub( lastConsumedIndex, fc - 1 )
            lastConsumedIndex = fc
        end

        if tok[1] == compiler.TokenType.CommentBegin then
            result[#result+1] = palette[compiler.SyntaxColor.Comment]
        elseif tok[1] == compiler.TokenType.CommentEnd then
            result[#result+1] = "|r"
        elseif tok[1] == compiler.TokenType.OpenBrace then
            if not insideBlock then
                insideBlock = true
            end
                indentationHint = 1
        elseif tok[1] == compiler.TokenType.CloseBrace then
            insideBlock = true
        else
            -- For everything else, retrieve the color index to use from the
            -- rule registered for that token
            local infos = parserClass:getRuleInfos( tok )
            if infos then
                if infos[1] then
                    result[#result+1] = palette[infos[1]]
                    result[#result+1] = sourceLine:sub( fc, lc )
                    result[#result+1] = "|r"
                    lastConsumedIndex = lc + 1
                elseif infos.specificColor then
                    result[#result+1] = infos.specificColor
                    result[#result+1] = sourceLine:sub( fc, lc )
                    result[#result+1] = "|r"
                    lastConsumedIndex = lc + 1
                end

                -- If the rule consumes a literal string verbatim, do
                -- the same and give it a specific color
                if infos and infos.consumeLiteralString then
                    result[#result+1] = palette[compiler.SyntaxColor.SpellOrItemName]
                    result[#result+1] = tokenizer:readUnquotedString( true )
                    result[#result+1] = "|r"
                    lastConsumedIndex = tokenizer.bytepos
                end

                if infos.shouldIndent then
                    indentationHint = 1
                    insideBlock = false
                elseif infos.indentAsStatement then
                    if not insideBlock then
                        insideBlock = true
                        indentationHint = -1
                    end
                end
            end
        end

        tok = tokenizer:readToken()
    end

    -- Get the remaining of the line
    if lastConsumedIndex <= #sourceLine then
        result[#result+1] = sourceLine:sub( lastConsumedIndex )
    end

    -- If the line is empty, propagate the indentation hint from the previous line
    if empty then
        indentationHint = state[4]
    end

    state = { tokenizer.blockCommentDepth, empty, insideBlock, indentationHint, currentIndention }
    setmetatable( state, colorizerStateMeta )
    return concat( result ), state
end
