--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Completion/Suggestion handling
--
local _, compiler = ...

local completionProvider = {}

-- Returns the number of completion items
function completionProvider:count()
    return #self
end

-- Returns the text and the display text for the completion item of index
function completionProvider:getItem( index )
    return self[index]
end

local function initOrGetCompletionProvider( p )
    local list = p.completionProvider or { encounteredTypes = {} }
    setmetatable( list, { __index = completionProvider } )
    p.completionProvider = list
    return list
end

local function addRelevantCompletions( p, complTok, ruleGen )
    local word = complTok.value
    local list = initOrGetCompletionProvider( p )

    local currentCC = p.completionContext
    local newResults = 0

    while( true ) do
        p.context.callbacks.timeslicer()

        local name, infos = ruleGen()
        if name == nil then
            break
        end

        local lname = name:lower()

        if not p.appendSugestionsOnly or infos.append then
            local shouldInclude = false
            local ruleCC = infos and infos.completionContext

            if ruleCC then
                for k,_ in pairs( currentCC ) do
                    if ruleCC[k] then
                        shouldInclude = true
                        break
                    end
                end
            end

            if shouldInclude and infos.completionPredicate then
                shouldInclude = infos.completionPredicate( p )
            end

            if shouldInclude and word and not infos.append then
                shouldInclude = lname:find( word )
            end

            if shouldInclude then
                if lname == word then
                    name = name .. ' '
                end

                local groupId = infos[1] or infos.class or 0
                list[#list+1] =
                {
                    name = name,
                    append = infos.append,
                    groupId = groupId,
                    doc = compiler.documentation[lname]
                }

                newResults = newResults + 1
            end
        end
    end

    if word then
        p.appendSugestionsOnly = true
    end

    if newResults == 0 then
        return
    end

    local l, c = p.context:getLoc( complTok[2] )

    if word then
        if newResults > 1 or list[1].name:lower() ~= word then
            p.context:completion( p.completionProvider, l, c )
        end
    else
        p.context:suggestion( p.completionProvider, l, c )
    end
end

local function endCompletion( p )
    p.prefixCompletionCallBack = nil
    p.onPushValue = nil
    p.onRuleApplied = nil
    p.onReadUnquotedString = nil
end

-- When we need to read an unquoted string during a completion,
-- then we need to cancel any solution we may have already emitted:
-- unquoted strings are open ended and there are no possible completions for
-- them (until I decide to do something stupid and try to complete spell names),
-- but the tokenizer may already have recognized the unquoted string as a partial
-- identifier, so we may have already emitted complations for it. So clear them.
local function cancelCompletion( p, s )
    if s ~= "" then
        p.context.completionProvider = nil
        p.context.suggestionProvider = nil
    end

    p.prefixCompletionCallBack = nil
    p.onPushValue = nil
    p.onRuleApplied = nil
    p.onReadUnquotedString = nil
end

local function prefixCompletion( p, complTok )
    addRelevantCompletions( p, complTok, p:getPrefixRules() )
end

local function infixCompletion( p, complTok )
    local cc = p.completionContext
    p.completionContext = {}

    -- We can get called for multiple pushValues() of the same type,
    -- so only emit solutions the first time we see any of these types.
    -- Example: the cursor right at the end of target.pet would cause a completion
    -- call for the unit value pushed by "pet" and then again for the new unit value
    -- pushed by the dot operator, but we don't want to suggest the dot operator twice as
    -- a possible continuation of the expression.
    local list = initOrGetCompletionProvider( p )
    if p.lhsVal and p.lhsVal[1] and not list.encounteredTypes[p.lhsVal[1]] then
        p.completionContext[p.lhsVal[1]] = 1
        list.encounteredTypes[p.lhsVal[1]] = 1
    end

    addRelevantCompletions( p, complTok, p:getInfixRules() )
    p.completionContext = cc
end

function compiler.SetupCompletion( parser )
    parser:registerPrefixRule( compiler.TokenType.Completion, nil,
        function( p, precedence, complTok )
            p.onRuleApplied = endCompletion
            p.onReadUnquotedString = cancelCompletion
            p.prefixCompletionCallBack = function()
                prefixCompletion( p, complTok )
            end

            return p:parseExpr( precedence )
        end
    )

    parser:registerInfixRule( compiler.TokenType.Completion, nil,
        function( _, prec )
            return prec + 1
        end,

        function( p, _, complTok )
            p.onRuleApplied = endCompletion
            p.onReadUnquotedString = cancelCompletion
            p.prefixCompletionCallBack = function()
                prefixCompletion( p, complTok )
            end

            infixCompletion( p, complTok )

            p.onPushValue = function( _, val )
                local ccc = p.completionContext
                p.completionContext = {}
                if val[1] then
                    p.completionContext[val[1]] = 1
                end

                infixCompletion( p, complTok )
                p.completionContext = ccc
            end

            return true
        end
    )
end
