--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, compiler = ...

compiler.documentation = {}

compiler.documentation["if"] =
[[$code$if <condition> <code> [else <code>]$$
$margin$The $c$if$$ statement lets you execute some code conditionally. It is followed by a boolean expression, and the code to execute.

The code can be comprised of a single statement, or multiple statements (including nested $c$if$$ or other control statements), enclosed in curly braces.

$c$else$$ can optionally be used to define an alternative piece of code to execute if the condition is false.$$
Example

$margin$$code$if CurrentTarget Exists and CurrentTarget Alive
{
    Emote lol
    Cast Sap
}
else
    Print no target$$$$]]

compiler.documentation["while"] =
[[$code$while <condition> <code>$$
$margin$The $c$while$$ statement executes a piece of code repeatedly, as long as the provided boolean expression is true.

However, it is not possible to actually execute multiple loops during one invocation of the macro. Therefore, the code inside the loop must always go through a suspension point (either a $c$Cast$$ statement with a spell that is on the global cooldown or a $c$suspend$$ statement).

Suspension points are indicated by an horizontal line. They are a place where the macro execution will stop, and the execution will be resumed there when invoking the macro again.$$
Example

$margin$$code$while CurrentTarget Exists and CurrentTarget Alive
{
    Cast Shuriken Toss
    Cast Loot-a-Rang
}$$
This is valid because Shuriken Toss is on the GCD and casting it implicitely inserts a suspension point.$$]]

compiler.documentation.either =
[[$code$either { <statements> }$$
$margin$The $c$either$$ statement executes the first of multiple $c$Cast,$$ $c$Use,$$ or $c$UseEquippedItem$$ statements.

The normal behavior of Smack, when invoking multiple spell/items one after another that are on the GCD, is to cast/use them in sequence with each invocation of the macro.

However, sometimes it is useful to cast the first spell or item available among several. Enclosing multiple statements in an $c$either$$ block is equivalent to simply putting them one after another in regular WoW macros.$$
Example

$margin$$code$either
{
    Use Conjured Mana Bun
    Use Ethereal Pomegranate
}$$$$]]

compiler.documentation.suspend =
[[$code$suspend$$
$margin$The $c$suspend$$ statement causes the execution of the macro to halt. It is resumed upon the next invocation of the macro.$$
Example

$margin$$code$Say This is a good addon
suspend
Say For a clown!
suspend
Say To play WoW at the circus$$
This will cause the three sentences to be said in sequence, one after each macro invocation. The sequence will then restart from the beginning.$$]]

compiler.documentation["return"] =
[[$code$return$$
$margin$The $c$return$$ statement ends the execution of the macro.

It is useful when used inside of a $c$if$$ to conditionally end the macro early.$$
Example

$margin$$code$if not Stealthed and not Combat
{
    Cast Stealth
    return
}

Cast Sap$$$$]]

compiler.documentation.var =
[[$code$var <name> = <expression>$$
$margin$Declares a variable of the given name and initialize it with the given expression.

The variable can be assigned a new value using $c$=$$ and can appear at any place where an expression of the corresponding type may appear.

The variable type is determined by the type of the expression used to initialize it and cannot be changed. A variable declared inside of a brace block is only visible inside of that block.$$
Example

$margin$$code$var tgt = CurrentTarget
if RCtrl tgt = CurrentFocus
Cast @tgt Blind$$
This will cause Blind to be cast at target, unless right ctrl is held, in which case it will be cast at focus.$$]]

compiler.documentation["break"] =
[[$code$break$$
$margin$The $c$break$$ statement exits the enclosing $c$while.$$$$]]

compiler.documentation.continue =
[[$code$continue$$
$margin$The $c$continue$$ statement proceeds with the next loop of the enclosing $c$while.$$$$]]

compiler.documentation["not"] =
[[$code$not <boolean expression>$$
$margin$The not operator negates a boolean expression.

It has priority over both the $c$or$$ and the $c$and$$ operators.

Parentheses can be used to modify the order of operations.$$]]

compiler.documentation["or"] =
[[$code$<boolean expression> or <boolean expression>$$
$margin$The or operator is a boolean expression that is true if either of its operand is true.

Both the $c$or$$ and the $c$and$$ operators have priority over $c$or.$$

Parentheses can be used to modify the order of operations.$$]]

compiler.documentation["and"] =
[[$code$<boolean expression> and <boolean expression>$$
$margin$The and operator is a boolean expression that is true only if both of its operands are true.

It has priority over the $c$or$$ operator, but the $c$not$$ operator have priority over $c$and$$.

Parentheses can be used to modify the order of operations.$$]]

compiler.documentation.print =
[[$code$Print <text>$$
$margin$The $c$Print$$ statement prints the provided text in the chat window.

This can be useful to debug macros with complex execution flows, by printing messages indicating which execution path have been taken.$$]]

compiler.documentation.emote =
[[$code$Emote <emote name>$$
$margin$The $c$Emote$$ statement plays the indicated emote.

Currently, any text can be used for the emote name. In a future version, this will accept only valid emote names, which will be suggested by autocompletion.$$]]

compiler.documentation.humanoid =
[[$code$Humanoid (boolean)$$
$margin$The $c$Humanoid$$ conditional is true if you are currently neither in a shapeshifted form. This is always true for classes other than druid and shaman.$$]]

compiler.documentation.bear =
[[$code$Bear (boolean)$$
$margin$$c$Bear$$ is true if you are currently a druid in bear form.$$]]

compiler.documentation.cat =
[[$code$Cat (boolean)$$
$margin$$c$Cat$$ is true if you are currently a druid in cat form.$$]]

compiler.documentation.travel =
[[$code$Travel (boolean)$$
$margin$$c$Travel$$ is true if you are currently a druid in travel form.$$]]

compiler.documentation.moonkin =
[[$code$Moonkin (boolean)$$
$margin$$c$Moonkin$$ is true if you are currently a druid in moonkin form.$$]]

compiler.documentation.treant =
[[$code$Treant (boolean)$$
$margin$$c$Treant$$ is true if you are currently a druid in treant form.$$]]

compiler.documentation.ghostwolf =
[[$code$GhostWolf (boolean)$$
$margin$$c$GhostWolf$$ is true if you are currently a shaman and in ghost wolf form.$$]]

compiler.documentation.shift =
[[$code$Shift (boolean)$$
$margin$$c$Shift$$ is true if either the left or right shift key is currently pressed.$$]]

compiler.documentation.alt =
[[$code$Alt (boolean)$$
$margin$$c$Alt$$ is true if either the left or right alt key is currently pressed.$$]]

compiler.documentation.ctrl =
[[$code$Ctrl (boolean)$$
$margin$$c$Ctrl$$ is true if either the left or right control key is currently pressed.$$]]

compiler.documentation.lshift =
[[$code$LShift (boolean)$$
$margin$$c$LShift$$ is true if the left shift key is currently pressed.$$]]

compiler.documentation.lalt =
[[$code$LAlt (boolean)$$
$margin$$c$LAlt$$ is true if the left alt key is currently pressed.$$]]

compiler.documentation.lctrl =
[[$code$LCtrl (boolean)$$
$margin$$c$LCtrl$$ is true if the left control key is currently pressed.$$]]

compiler.documentation.rshift =
[[$code$RShift (boolean)$$
$margin$$c$RShift$$ is true if the right shift key is currently pressed.$$]]

compiler.documentation.ralt =
[[$code$RAlt (boolean)$$
$margin$$c$RAlt$$ is true if the right alt key is currently pressed.$$]]

compiler.documentation.rctrl =
[[$code$RCtrl (boolean)$$
$margin$$c$RCtrl$$ is true if the right control key is currently pressed.$$]]

compiler.documentation.havepet =
[[$code$HavePet (boolean)$$
$margin$$c$HavePet$$ is true if you currently have a combat pet.$$]]


compiler.documentation.cast =
[[$code$Cast [@<unit expression>] <spell name>$$
$margin$The $c$Cast$$ statement casts a spell or use an item. If there are both an item and a spell available with the same name, this casts the spell.

It can optionally be followed immediately by $c$@$$ and a specific unit to target the spell to, instead of using the current target.

If the spell or item is known to be on the global cooldown and other statements follow the cast, a suspension point is automatically inserted, as indicated visually by an horizontal line.

Note: automatic suspension is not currently supported for items and toys. Some spells on the GCD may also not be currently detected as such. In these cases, a suspension point can be inserted explicitely using the $c$suspend$$ statement.$$]]

compiler.documentation.noform =
[[$code$NoForm (boolean)$$
$margin$$c$NoForm$$ is true if you are currently neither in a shapeshifted form, priest shadowform, or rogue stealth.$$]]

compiler.documentation.mountform =
[[$code$MountForm (boolean)$$
$margin$$c$MountForm$$ is true if you are currently a druid in mount form.$$]]

compiler.documentation.treeoflife =
[[$code$TreeOfLife (boolean)$$
$margin$$c$TreeOfLife$$ is true if you are currently a druid in tree of life form.$$]]

compiler.documentation.stealthed =
[[$code$Stealthed (boolean)$$
$margin$$c$Stealthed$$ is true if you are currently stealthed.

If you are a rogue, this is true not only when you are using your own stealth, but also if you are stealthed for other reasons, such as being shrouded by another rogue without being in stealth yourself.$$]]

compiler.documentation.roguestealth =
[[$code$RogueStealth (boolean)$$
$margin$$c$RogueStealth$$ is true if you are currently a rogue using stealth.

Note that this is different from being stealthed from other reasons, such as being shrouded by another rogue.

You can use the $c$Stealthed$$ conditional for a more generic and class agnostic stealth check.$$]]

compiler.documentation.pseudostealth =
[[$code$PseudoStealth (boolean)$$
$margin$$c$PseudoStealth$$ is true if you are currently a rogue and in a state considered as stealth for the purpose of using abilities, such as Subterfuge and Shadow Dance.

Note that unfortunately this is also true if you are a Night Fae rogue in SoulShape form, as for macro purpose it is considered to be the same form index as the above. Blame Blizzard for the clown show that are macro form indices.$$]]

compiler.documentation.canusestealthskills =
[[$code$CanUseStealthSkills (boolean)$$
$margin$$c$CanUseStealthSkills$$ is true if you are currently a rogue and can use stealth skills.

This is merely a convenient shortcut for $c$(RogueStealth or Pseudostealth).$$$$]]

compiler.documentation.shadowform =
[[$code$ShadowForm (boolean)$$
$margin$$c$ShadowForm$$ is true if you are currently a shadow priest and in shadow form.$$]]

compiler.documentation.spiritofredemption =
[[$code$SpiritOfRedemption (boolean)$$
$margin$$c$SpiritOfRedemption$$ is true if you are currently a holy priest and in spirit of redemption form.$$]]

compiler.documentation.vehiclebar =
[[$code$VehicleBar (boolean)$$
$margin$$c$VehicleBar$$ is true if a vehicle action bar is currently active.

Not to be confused with the override bar, which is exactly the same thing without the vehicle.$$]]

compiler.documentation.overridebar =
[[$code$OverrideBar (boolean)$$
$margin$$c$OverrideBar$$ is true if an override action bar is currently active.

Not to be confused with the vehicle bar, which is exactly the same thing but you're in a vehicle.$$]]

compiler.documentation.shittygimmickbar =
[[$code$ShittyGimmickBar (boolean)$$
$margin$$c$ShittyGimmickBar$$ is true if your action bars are currently replaced with some tedious minigame actions, regardless of being in a vehicle or not.

This is a convenience shortcut for $c$(VehicleBar or OverrideBar).$$$$
Example

$margin$$code$while ShittyGimmickBar
{
    click ActionButton1
    suspend
    if not ShittyGimmickBar break
    click ActionButton2
    suspend
    if not ShittyGimmickBar break
    click ActionButton3
    suspend
    if not ShittyGimmickBar break
    click ActionButton4
    suspend
}$$
Inserting the code above at the beginning of any other macro will allow you to spam it to optimally use any currently active vehicle or override ability in most such situations.
Looking at the screen is optional in some cases.$$]]
