--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The conditionals. This contains both the equivalents of
-- the builtin macro conditionals, and some custom ones.
--
-- To avoid having to define a gazillion expression opcodes,
-- we use a generic one that holds a string containing the
-- expression to use at runtime.
--
local _, compiler = ...

local format = string.format

local unitConditionalInfos =
{
    compiler.SyntaxColor.Conditional, completionContext = { unit=1, rhsUnit=1 },

    -- Unit conditionals are postfix operators that take a unit expression and return a boolean
    -- expression. Knowing that we are in a unit expression context is not enough to decide
    -- if they can be suggested by completion, we also need to be sure that whatever
    -- requested this expression to be parsed wanted a boolean result.
    completionPredicate = function( p )
        return p.expectedExpressionType == "bool"
    end
}

function compiler.SetupConditionals( parser )
    parser:createPostfixOperatorRule( "Exists", compiler.Precedence.ConditionalOps, unitConditionalInfos,
        function( loc, lhsVal, p )
            if lhsVal[1] ~= "unit" then
                p.context:error( lhsVal[2], "Exists: expected a unit expression." )
                return true
            end

            loc = p.context.locs:createSpanningLocation( loc, lhsVal[2] )
            p:pushValue( { "bool", loc, compiler.CFG.ExprOpCodes.Conditional, "UnitExists(%s)", lhsVal } )
            return true
        end
    )

    parser:createPostfixOperatorRule( "Dead", compiler.Precedence.ConditionalOps, unitConditionalInfos,
        function( loc, lhsVal, p )
            if lhsVal[1] ~= "unit" then
                p.context:error( lhsVal[2], "Dead: expected a unit expression." )
                return true
            end

            loc = p.context.locs:createSpanningLocation( loc, lhsVal[2] )
            p:pushValue( { "bool", loc, compiler.CFG.ExprOpCodes.Conditional, "UnitIsDead(%s)", lhsVal } )
            return true
        end
    )

    parser:createPostfixOperatorRule( "Alive", compiler.Precedence.ConditionalOps, unitConditionalInfos,
        function( loc, lhsVal, p )
            if lhsVal[1] ~= "unit" then
                p.context:error( lhsVal[2], "Alive: expected a unit expression." )
                return true
            end

            loc = p.context.locs:createSpanningLocation( loc, lhsVal[2] )
            local deadExpr = { "bool", lhsVal[2], compiler.CFG.ExprOpCodes.Conditional, "UnitIsDead(%s)", lhsVal }
            p:pushValue( { "bool", lhsVal[2], compiler.CFG.ExprOpCodes.Not, deadExpr } )
            return true
        end
    )

    parser:createPostfixOperatorRule( "Hostile", compiler.Precedence.ConditionalOps, unitConditionalInfos,
        function( loc, lhsVal, p )
            if lhsVal[1] ~= "unit" then
                p.context:error( lhsVal[2], "Hostile: expected a unit expression." )
                return true
            end

            loc = p.context.locs:createSpanningLocation( loc, lhsVal[2] )
            p:pushValue( { "bool", lhsVal[2], compiler.CFG.ExprOpCodes.Conditional,
                "SecureCmdOptionParse( '[@%s,harm]' )", lhsVal, true } )
            return true
        end
    )

    parser:createPostfixOperatorRule( "Friendly", compiler.Precedence.ConditionalOps, unitConditionalInfos,
        function( loc, lhsVal, p )
            if lhsVal[1] ~= "unit" then
                p.context:error( lhsVal[2], "Friendly: expected a unit expression." )
                return {}
            end

            loc = p.context.locs:createSpanningLocation( loc, lhsVal[2] )
            p:pushValue( { "bool", lhsVal[2], compiler.CFG.ExprOpCodes.Conditional,
                "SecureCmdOptionParse( '[@%s,noharm]' )", lhsVal, true } )
            return true
        end
    )

    local prefixConditionalInfos = { compiler.SyntaxColor.Conditional, completionContext = { bool=1 } }

    parser:registerPrefixRule( "Stealthed", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsStealthed()" } )
            return true
        end
    )

    parser:registerPrefixRule( "Combat", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "SecureCmdOptionParse( '[combat]' )" } )
            return true
        end
    )

    parser:registerPrefixRule( "Shift", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsShiftKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "Alt", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsAltKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "Ctrl", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsControlKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "LShift", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsLeftShiftKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "LAlt", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsLeftAltKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "LCtrl", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsLeftControlKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "RShift", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsRightShiftKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "RAlt", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsRightAltKeyDown()" } )
            return true
        end
    )

    parser:registerPrefixRule( "RCtrl", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsRightControlKeyDown()" } )
            return true
        end
    )

    -- TODO pet types
    parser:registerPrefixRule( "HavePet", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "SecureCmdOptionParse( '[pet]' )" } )
            return true
        end
    )

    parser:registerPrefixRule( "IsMounted", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsMounted()" } )
            return true
        end
    )

    parser:registerPrefixRule( "IsFlying", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "IsFlying()" } )
            return true
        end
    )

    parser:registerPrefixRule( "VehicleBar", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "HasVehicleActionBar()" } )
            return true
        end
    )

    parser:registerPrefixRule( "OverrideBar", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "HasOverrideActionBar()" } )
            return true
        end
    )

    parser:registerPrefixRule( "ShittyGimmickBar", prefixConditionalInfos,
        function( p, _, tok )
            local veh = { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "HasVehicleActionBar()" }
            local ov = { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "HasOverrideActionBar()" }
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Or, veh, ov } )
            return true
        end
    )
end

compiler.CodeGen:RegisterExprOpcodeHandler( compiler.CFG.ExprOpCodes.Conditional, function( codegen, expr )
    if expr[5] then
        return format( expr[4], expr[6] and
            codegen:generateInStrExpr( expr[5] ) or codegen:generateStrExpr( expr[5] ) )
    else
        return expr[4]
    end
end )
