--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Player related conditionals: class, race, faction, spec
--
local _, compiler = ...

local prefixConditionalInfos = { compiler.SyntaxColor.Conditional, completionContext = { bool=1 } }

function compiler.SetupPlayerConditionals( parser )
    -- Class
    local smack = _G["Smack_Main"]
    local classes = smack.ClassNames
    local _, playerClass = UnitClassBase( "player" )

    for i=1,#classes,1 do
        local cname = classes[i]
        local _, _, _, color = GetClassColor( cname:upper() )

        parser:registerPrefixRule( cname, { specificColor = '|c' .. color .. '|r', class=cname, completionContext = { bool=1 } },
            function( p, _, tok )
                p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Constant, playerClass == i } )
                return true
            end
        )
    end

    -- Race
    local _, _, playerRace = UnitRace( "player" )

    -- There is no convenient api function that returns the number of races,
    -- also the race list includes non player races. So we are stuck with this ugly shit.
    local NumRaces = 37
    local NonPlayerRacesRange = { 12, 21 }

    for i=1, NumRaces, 1 do
        if ( i < NonPlayerRacesRange[1] or i > NonPlayerRacesRange[2] ) and i ~= 25 and i ~= 26 and i ~= 33 then
            local rinfo = C_CreatureInfo.GetRaceInfo( i )

            if rinfo then
                local rname = rinfo.clientFileString

                -- FFS maybe I should just bite the bullet and also make my own table for this
                if rname == "Scourge" then rname = "Undead" end

                local isPlayersRace = playerRace == i

                -- Special case for pandaren because there are 3 pandaren race ids
                -- depending on their allegiance
                if i == 24 then
                    isPlayersRace = playerRace >= 24 and playerRace <= 26
                end

                parser:registerPrefixRule( rname, prefixConditionalInfos,
                    function( p, _, tok )
                        p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Constant, isPlayersRace } )
                        return true
                    end
                )
            end
        end
    end

    -- Faction
    local playerFaction = UnitFactionGroup( "player" )

    parser:registerPrefixRule( "Horde", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Constant, playerFaction == "Horde" } )
            return true
        end
    )

    parser:registerPrefixRule( "Alliance", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Constant, playerFaction == "Alliance" } )
            return true
        end
    )

    parser:registerPrefixRule( "Neutral", prefixConditionalInfos,
        function( p, _, tok )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Constant, playerFaction == "Neutral" } )
            return true
        end
    )

    -- Specializations
    local specs = smack.SpecsDB

    for i = 1, #specs, 1 do
        local s = specs[i]
        local name = s[1]
        local classes = s[2]

        local infos

        -- If there is only one class with a spec of that name, setup its syntax coloring to
        -- that class color. Otherwise, use the generic conditional color.

        -- We can't even easily get the number of non array entries in a table
        -- shit-ass language
        local numClasses = 0
        for _,_ in pairs( classes ) do
            numClasses = numClasses + 1
            if numClasses > 1 then
                break
            end
        end

        if numClasses == 1 then
            local classIndex = next( classes )
            local cinfo = C_CreatureInfo.GetClassInfo( classIndex )
            local cname = cinfo.classFile:lower()
            local color = C_ClassColor.GetClassColor( cname )
            infos = { specificColor = color:GenerateHexColorMarkup(), class=cname, completionContext = { bool=1 } }
        else
            infos = prefixConditionalInfos
        end

        -- If the spec doesn't exist for the current player class,
        -- register the false constant at its name. Otherwise, register
        -- a check of the current player spec id, which will be stored as
        -- a property in the parent frame of the macro.
        local specId = classes[playerClass]

        if specId then
            parser:registerPrefixRule( name, infos,
                function( p, _, tok )
                    p:addInit( "local spec = self:GetParent():GetAttribute('spec')" )
                    p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "spec == " .. specId } )
                    return true
                end
            )
        else
            parser:registerPrefixRule( name, infos,
                function( p, _, tok )
                    p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Constant, false } )
                    return true
                end
            )
        end
    end

    -- Covenants
    parser:registerPrefixRule( "Kyrian", prefixConditionalInfos,
        function( p, _, tok )
            p:addInit( "local cov = self:GetParent():GetAttribute('cov')" )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "cov == 1" } )
            return true
        end
    )

    parser:registerPrefixRule( "Venthyr", prefixConditionalInfos,
        function( p, _, tok )
            p:addInit( "local cov = self:GetParent():GetAttribute('cov')" )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "cov == 2" } )
            return true
        end
    )

    parser:registerPrefixRule( "NightFae", prefixConditionalInfos,
        function( p, _, tok )
            p:addInit( "local cov = self:GetParent():GetAttribute('cov')" )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "cov == 3" } )
            return true
        end
    )

    parser:registerPrefixRule( "NecroLord", prefixConditionalInfos,
        function( p, _, tok )
            p:addInit( "local cov = self:GetParent():GetAttribute('cov')" )
            p:pushValue( { "bool", tok[2], compiler.CFG.ExprOpCodes.Conditional, "cov == 4" } )
            return true
        end
    )
end
