--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Unit specification expressions.
--
-- Unit literals:
-- player, target, focus, pet, mouseover (to be completed)
-- me is an alias for player.
--
-- The dot operator, applied on a target literal, allows
-- to modify it. Example:
-- target.target, target.pet, etc.
-- Only "target" and "pet" are allowed to the right of the dot
-- operator, but chaining multiple dot operators is allowed
-- like so: mouseover.target.pet.target
-- (no idea why anyone would want to do that, I'm just reflecting
-- the capabilities of the api)
--
-- TODO: all the other things, and the ability to explicitely
-- specify a player name.
--
local _, compiler = ...

-- Unit literals can be offered by completion either when an unit expression is expected, or if a bool
-- expression is expected (since unit conditionals such as "target exists" begin with an unit expression)
local unitLiteralsInfos = { compiler.SyntaxColor.Conditional, completionContext = { bool=1, unit=1 } }

-- For the subset of unit literals that are allowed to the right of the dot operator
local rhsUnitLiteralsInfos = { compiler.SyntaxColor.Conditional, completionContext = { bool=1, unit=1, rhsUnit=1 } }

function compiler.SetupUnitExpr( parser )

    parser:registerPrefixRule( "CurrentTarget", rhsUnitLiteralsInfos,
        function( p, _, tok )
            p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "target" } )
            return true
        end
    )

    parser:registerPrefixRule( "CurrentFocus", unitLiteralsInfos,
        function( p, _, tok )
            p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "focus" } )
            return true
        end
    )

    parser:registerPrefixRule( "Player", unitLiteralsInfos,
        function( p, _, tok )
            p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "player" } )
            return true
        end
    )

    parser:registerPrefixRule( "Me", unitLiteralsInfos,
        function( p, _, tok )
            p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "player" } )
            return true
        end
    )

    parser:registerPrefixRule( "Pet", rhsUnitLiteralsInfos,
        function( p, _, tok )
            p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "pet" } )
            return true
        end
    )

    parser:registerPrefixRule( "MouseOver", unitLiteralsInfos,
        function( p, _, tok )
            p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "mouseover" } )
            return true
        end
    )

    for i=1,5,1 do
        parser:registerPrefixRule( "Arena"..i, unitLiteralsInfos,
            function( p, _, tok )
                p:pushValue( { "unit", tok[2], compiler.CFG.ExprOpCodes.Constant, "arena"..i } )
                return true
            end
        )
    end

    local dotOpInfos =
    {
        completionContext = { unit=1, rhsUnit=1 },
        rhsCompletionContext = { rhsUnit=1 },
        append = true
    }

    parser:createLeftAssInfixOperatorRule( ".", compiler.Precedence.DotOp, dotOpInfos,
        function( loc, lhsVal, rhsVal, p )
            if lhsVal[1] ~= "unit" then
                p.context:error( lhsVal[2], "Dot: expected a unit expression on the left hand side." )
                return {}
            end

            if rhsVal[1] ~= "unit" then
                p.context:error( rhsVal[2], "Dot: expected a unit expression on the right hand side." )
                return {}
            end

            if rhsVal[3] ~= compiler.CFG.ExprOpCodes.Constant or rhsVal[4] ~= "target" and rhsVal[4] ~= "pet" then
                p.context:error( rhsVal[2], "Only 'target' and 'pet' are allowed on the right of the dot operator." )
                return {}
            end

            loc = p.context.locs:createSpanningLocation( lhsVal[2], rhsVal[2] )

            if lhsVal[3] == compiler.CFG.ExprOpCodes.Constant then
                p:pushValue( { "unit", loc, compiler.CFG.ExprOpCodes.Constant, lhsVal[4] .. rhsVal[4] } )
            else
                p:pushValue( { "unit", loc, compiler.CFG.ExprOpCodes.StrCat, lhsVal, rhsVal } )
            end

            return true
        end
    )

end
