--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Variable declaration, expression, and assignment.
--
local _, compiler = ...

function compiler.SetupVarRules( parser )
    -- Variable declaration statement
    compiler.createStatement( parser, "var", { compiler.SyntaxColor.ControlStmt, indentAsStatement = true },
        function( p, _, tok )
            local kwLoc = tok[2]
            tok = p.tokenizer:readToken()
            if not tok or tok[1] ~= compiler.TokenType.Identifier then
                p.context:error( kwLoc, "Identifier expected after 'var'." )
                return
            end

            local nameLoc = tok[2]
            local varName = tok.value
            if p:isReservedWord( varName ) then
                p.context:error( nameLoc, "'" .. varName .. "' is a reserved word and can't be used as a variable name." )
                return {}
            end

            tok = p.tokenizer:readToken()
            if not tok or tok[1] ~= compiler.TokenType.Identifier or tok.value ~= "=" then
                local loc = p.context.locs:createSpanningLocation( kwLoc, nameLoc )
                p.context:error( loc, "'=' expected after 'var " .. varName .. "'." )
                return
            end

            local success = p:parseExpr( compiler.Precedence.Statement )
            local initVal = p:popValue()

            if not success or not initVal or initVal[1] == nil then
                local loc
                if initVal then
                    loc = initVal[2]
                else
                    loc = p.context.locs:createSpanningLocation( kwLoc, tok[2] )
                end
                p.context:error( loc, "Expression expected after 'var " .. varName .. " ='." )
                return
            end

            local varIndex = p.varIndices[varName]
            if varIndex == nil then
                p.numVars = p.numVars + 1
                varIndex = p.numVars
                p.varIndices[varName] = varIndex
            end

            if not p.scopeStack:declVar( varName, { varIndex, initVal[1] } ) then
                p.context:error( nameLoc, "A variable named '" .. varName .. "' already exists in this scope." )
                return {}
            end

            if initVal[1] == "unit" then
                p.context:notice( nameLoc, "Note: this variable only contains a unit specification, which will be re-evaluated everytime it's used as the input for a conditional or as the target of a cast." )
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.SetVar, varIndex, initVal } )
            return {}
        end )

    -- Variable expression
    parser:registerPrefixRule( compiler.TokenType.Identifier, { compiler.SyntaxColor.Variable },
        function( p, prec, tok )
            local var = p.scopeStack:lookUp( tok.value )

            if not var then
                p.context:error( tok[2], "Undefined identifier '" .. tok.value .. "'" )
                p.context:poison()

                -- Partial recovery: we want to preserve the current completion context, because this may
                -- just be a valid but incomplete identifier
                return p:parseExpr( prec, parser.completionContext )
            end

            if var[2] == "unit" then
                p.context:notice( tok[2], "Note: this variable only contains a unit specification, which will be re-evaluated everytime it's used as the input for a conditional or as the target of a cast." )
            end

            p:pushValue( { var[2], tok[2], compiler.CFG.ExprOpCodes.GetVar, var[1] } )
            return true
        end )

    -- Assignment operator
    parser:createLeftAssInfixOperatorRule( "=", compiler.Precedence.AssignmentOp, { indentAsStatement = true },
        function( loc, lhsVal, rhsVal, p )
            if p.cfg.currentInsertionPointBB == nil then
                loc = p.context.locs:createSpanningLocation( lhsVal[2], rhsVal[2] )
                p.context:error( loc, "Unreachable code." )
                p.context:poison()
                return true
            end

            if lhsVal[3] ~= compiler.CFG.ExprOpCodes.GetVar then
                p.context:error( lhsVal[2], "=: expected a variable name on the lhs." )
                return false
            end

            if lhsVal[1] ~= rhsVal[1] then
                loc = p.context.locs:createSpanningLocation( lhsVal[2], rhsVal[2] )
                p.context:error( loc, "Can't assign an expression type of a different type than the one that the variable was initialized with." )
                return true
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.SetVar, lhsVal[4], rhsVal } )
            return true
        end )
end

-- GetVar/SetVar opcode handlers.
compiler.CodeGen:RegisterExprOpcodeHandler( compiler.CFG.ExprOpCodes.GetVar, function( _, expr )
    return "v[" .. expr[4] .. "]"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.SetVar, function( codegen, instr )
    return "v[" .. instr[2] .. "] = " .. codegen:generateStrExpr( instr[3] )
end )
