--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The parser's core. This is a generic Pratt parser,
-- whose actual parsing rules are defined separately.
--
local _, compiler = ...
local Parser = {}
compiler.Parser = Parser

local yield = coroutine.yield
local wrap = coroutine.wrap

function Parser:new( tokenizer, context )
    local o = {}

    o.prefixRules = {}
    o.infixRules = {}
    o.tokenizer = tokenizer
    o.context = context
    o.completionContext = {}

    setmetatable( o, { __index = self } )
    return o
end

----------------------------------------------------------------------------------------
-- Parsing rules management
----------------------------------------------------------------------------------------
function Parser:registerPrefixRule( tokTypeOrIdentifier, infos, parseFunc )
    local r = { infos = infos, parse = parseFunc }

    if type( tokTypeOrIdentifier ) == "string" then
        r.name = tokTypeOrIdentifier
        tokTypeOrIdentifier = tokTypeOrIdentifier:lower()
    end

    self.prefixRules[tokTypeOrIdentifier] = r
end

function Parser:registerInfixRule( tokTypeOrIdentifier, infos, precFunc, parseFunc )
    local r = { infos = infos, prec = precFunc, parse = parseFunc }

    if type( tokTypeOrIdentifier ) == "string" then
        r.name = tokTypeOrIdentifier
        tokTypeOrIdentifier = tokTypeOrIdentifier:lower()
    end

    self.infixRules[tokTypeOrIdentifier] = r
end

local function findPrefixRule( self, token )
    local rule = nil

    if token[1] == compiler.TokenType.Identifier then
        rule = self.prefixRules[token.value]
    end

    return rule or self.prefixRules[token[1]]
end

local function findInfixRule( self, token )
    local rule = nil

    if token[1] == compiler.TokenType.Identifier then
        rule = self.infixRules[token.value]
    end

    return rule or self.infixRules[token[1]]
end

-- Rule query api
function Parser:isReservedWord( str )
    return self.prefixRules[str] ~= nil or self.infixRules[str] ~= nil
end

function Parser:getRuleInfos( tok )
    local rule = findInfixRule( self, tok ) or findPrefixRule( self, tok )
    return rule and rule.infos or nil
end

function Parser:getPrefixRules()
    local function gen()
        for k,v in pairs( self.prefixRules ) do
            if type(k) == "string" then
                yield( v.name or k, v.infos )
            end
        end
    end

    return wrap( gen )
end

function Parser:getInfixRules()
    local function gen()
        for k,v in pairs( self.infixRules ) do
            if type(k) == "string" then
                yield( v.name or k, v.infos, v )
            end
        end
    end

    return wrap( gen )
end

----------------------------------------------------------------------------------------
-- Core parsing logic
----------------------------------------------------------------------------------------
local function parsePrefix( self, precedence )
    if self.prefixCompletionCallBack then
        self.prefixCompletionCallBack( self )
    end

    local tok = self.tokenizer:readToken()
    if not tok then
        return
    end

    local rule = findPrefixRule( self, tok )

    if not rule then
        return
    end

    if self.onRuleApplied then
        self.onRuleApplied( self )
    end

    return rule.parse( self, precedence, tok )
end

function Parser:parseExpr( precedence, completionContext )
    local prec = precedence or 0

    local savedCompletionContext = self.completionContext
    self.completionContext = completionContext or self.completionContext

    local success = parsePrefix( self, prec )
    if not success then
        -- We have to make sure to restore what needs to be restored on every
        -- exit path of the function by hand like cavemen. Thanks as always for being
        -- a bad language, lua
        self.completionContext = savedCompletionContext
        return false
    end

    while true do
        self.context.callbacks.timeslicer()

        local nextTok = self.tokenizer:lookAhead()
        if not nextTok then
            self.completionContext = savedCompletionContext
            return true
        end

        local rule = findInfixRule( self, nextTok )
        if not rule then
            self.completionContext = savedCompletionContext
            return true
        end

        if rule.prec( self, prec ) <= prec then
            self.completionContext = savedCompletionContext
            return true
        end

        if self.onRuleApplied then
            self.onRuleApplied( self )
        end

        -- Now that we confirmed that we have a parsing rule to apply to it,
        -- consume the token.
        self.tokenizer:readToken()
        success = rule.parse( self, prec, nextTok )
        if not success then
            self.completionContext = savedCompletionContext
            return
        end
    end

    self.completionContext = savedCompletionContext
end

function Parser:pushValue( val )
    if self.lhsVal and self.onDropValue then
        self.onDropValue( self, self.lhsVal )
    end

    self.lhsVal = val
    if val and self.onPushValue then
        self.onPushValue( self, val )
    end
end

function Parser:popValue()
    local val = self.lhsVal
    self.lhsVal = nil
    return val
end

function Parser:readUnquotedString()
    local s = self.tokenizer:readUnquotedString()

    if self.onReadUnquotedString then
        self.onReadUnquotedString( self, s )
    end

    return s
end

----------------------------------------------------------------------------------------
-- Helpers to create operator rules (prefix, postfix, left or right associative infix)
----------------------------------------------------------------------------------------
function Parser:createPrefixOperatorRule( identifier, precedence, infos, func )
    local function parseFunc( parser, _, tok )
        local success = parser:parseExpr( precedence + 1, infos.completionContext )
        local rhsVal = parser:popValue()

        if not success or not rhsVal then
            parser.context:error( tok[2], "Sub-expression expected after prefix operator '" .. identifier .. "'." )
            return
        end

        return func( tok[2], rhsVal, parser )
    end

    self:registerPrefixRule( identifier, infos, parseFunc )
end

function Parser:createPostfixOperatorRule( identifier, precedence, infos, func )
    local function precFunc()
        return precedence
    end

    local function parseFunc( parser, _, tok )
        local lhsVal = parser:popValue()
        return func( tok[2], lhsVal, parser )
    end

    self:registerInfixRule( identifier, infos, precFunc, parseFunc )
end

function Parser:createLeftAssInfixOperatorRule( identifier, precedence, infos, func )
    local function precFunc()
        return precedence
    end

    local function parseFunc( parser, _, tok )
        local lhsVal = parser:popValue()
        if not lhsVal then
            parser.context:error( tok[2], "Sub-expression expected to the left of infix operator '" .. identifier .. "'." )
            return false
        end

        local success = parser:parseExpr( precedence + 1, infos.rhsCompletionContext or infos.completionContext )
        local rhsVal = parser:popValue()

        if not success or not rhsVal then
            parser.context:error( tok[2], "Sub-expression expected to the right of infix operator '" .. identifier .. "'." )
            return false
        end

        return func( tok[2], lhsVal, rhsVal, parser )
    end

    self:registerInfixRule( identifier, infos, precFunc, parseFunc )
end

function Parser:createRightAssInfixOperatorRule( identifier, precedence, infos, func )
    local function precFunc()
        return precedence
    end

    local function parseFunc( parser, _, tok )
        local lhsVal = parser:popValue()
        if not lhsVal then
            parser.context:error( tok[2], "Sub-expression expected to the left of infix operator '" .. identifier .. "'." )
            return false
        end

        local success = parser:parseExpr( precedence, infos.rhsCompletionContext or infos.completionContext )
        local rhsVal = parser:popValue()

        if not success or not rhsVal then
            parser.context:error( tok[2], "Sub-expression expected to the right of infix operator '" .. identifier .. "'." )
            return false
        end

        return func( tok[2], lhsVal, rhsVal, parser )
    end

    self:registerInfixRule( identifier, infos, precFunc, parseFunc )
end
