--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The Smack parser, which instantiates the parser core
-- and registers all of the language's parsing rules.
--
local _, compiler = ...

-- Setup the parser prototype instance with all the
-- parsing rules for the Smack language
local SmackParser = compiler.Parser:new()
compiler.SmackParser = SmackParser

-- When dropping a value (ie when not using the result of an expression),
-- check if it is legit (dropping a pending if that wans't consumed by a else),
-- or invalid (dropping an expression that was just around by itself and therefore
-- useless since expressions don't have side effects)
SmackParser.onDropValue = function( parser, val )
    if val[1] == "pendingIf" then
        return
    end

    if val[1] == "blockEnd" then
        parser.context:error( val[2], "Unexpected '}'." )
    else
        parser.context:error( val[2], "Unexpected "..val[1].." expression." )
    end
end

-- Parsing rules for invalid tokens
SmackParser:registerPrefixRule( compiler.TokenType.Error, nil,
    function( parser, prec, tok )
        parser.context:error( tok[2], "Unrecognized token." )
        return parser:parseExpr( prec )
    end
)

-- Parsing rules for parenthesized sub expressions
SmackParser:registerPrefixRule( compiler.TokenType.OpenParen, nil,
    function( parser )
        local completionContext = {}
        for k,_ in pairs( parser.completionContext ) do
            if k ~= "stmt" then
                completionContext[k]=1
            end
        end

        parser.lhsVal = nil

        local result = parser:parseExpr( nil, completionContext )
        if not result then
            return
        end

        local tok = parser.tokenizer:readToken()
        if not tok or tok[1] ~= compiler.TokenType.CloseParen then
            parser.context:error( tok and tok[2] or parser.tokenizer:getCurrentLoc(), "Missing ')'." )
            return
        end

        return result
    end
)

SmackParser:registerPrefixRule( compiler.TokenType.CloseParen, nil,
    function( parser, _, tok )
        parser.context:error( tok[2], "Unexpected ')'." )
        return false
    end
)

-- Parsing rules for brace blocks
local stmtParsingContext = { stmt=1 }

SmackParser:registerPrefixRule( compiler.TokenType.OpenBrace, nil,
    function( parser )
        parser.scopeStack:push()

        while parser:parseExpr( nil, stmtParsingContext ) do
            if parser.lhsVal and parser.lhsVal[1] == "blockEnd" then
                parser:popValue()
                return true
            else
                parser:pushValue()
            end
        end

        parser.context:error( parser.tokenizer:getCurrentLoc(), "Missing '}'." )
        return false
    end
)

SmackParser:registerPrefixRule( compiler.TokenType.CloseBrace, nil,
    function( parser, _, tok )
        parser:pushValue( { "blockEnd", tok[2] } )
        return true
    end
)

-- All variable related parsing rules
compiler.SetupVarRules( SmackParser )

-- Install the bool literals and bool operators parsing rules
compiler.SetupBooleans( SmackParser )

-- Install the control statements.
compiler.SetupIfStatement( SmackParser )
compiler.SetupElseStatement( SmackParser )
compiler.SetupWhileStatement( SmackParser )
compiler.SetupControlStatements( SmackParser )
compiler.SetupEitherStatement( SmackParser )
compiler.SetupOnStatement( SmackParser )

-- Install the action statements
compiler.SetupActionStatements( SmackParser )

-- Install the unit expressions (target, target.target and such).
compiler.SetupUnitExpr( SmackParser )

-- Install slot names
compiler.SetupSlotLiterals( SmackParser )

compiler.SetupConditionals( SmackParser )
compiler.SetupPlayerConditionals( SmackParser )
compiler.SetupFormConditionals( SmackParser )

compiler.SetupCompletion( SmackParser )

function SmackParser:new( str, context, cursorLine, cursorCol )
    local o = {}

    o.tokenizer = compiler.Tokenizer:new( str, context )
    o.tokenizer:setCursorPosition( cursorLine, cursorCol )
    o.context = context

    o.scopeStack = compiler.ScopeStack:new()
    o.varIndices = {}
    o.numVars = 0
    o.inits = {}
    o.eventHandlers = {}

    -- Push a scope for the top level
    o.scopeStack:push()

    -- Create a CFG with an entry point basic block,
    -- and store it in the parser so the parsing rules
    -- can find it to append code to it.
    o.cfg = compiler.CFG:new()

    setmetatable( o, { __index = self } )
    return o
end

function SmackParser:parse()
    -- The top level parsing loop.
    --
    -- TODO report a syntax error if we stopped before consuming
    -- all the input. For instance a stray ')' would just cause
    -- the parser to stop without saying anything currently.

    while self:parseExpr( nil, stmtParsingContext ) do
        self:pushValue()
    end

    if self.cfg.currentInsertionPointBB then
        self.cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Return }
    end
end

function SmackParser:addInit( init )
    self.inits[init] = true
end
