--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Action statements. (cast, use and such)
--
local _, compiler = ...

function compiler.SetupActionStatements( parser )
    compiler.createStatement( parser, "Print", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            -- TODO this should take an actual string expression,
            -- once we have at least string literals...
            local sayStr = p:readUnquotedString()
            if sayStr == "" then
                p.context:error( tok[2], "print: expected a string." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Print, sayStr } )
            return {}
        end
    )

    compiler.createStatement( parser, "Say", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            -- TODO this should take an actual string expression,
            -- once we have at least string literals...
            local sayStr = p:readUnquotedString()
            if sayStr == "" then
                p.context:error( tok[2], "say: expected a string." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Say, sayStr } )
            return {}
        end
    )

    compiler.createStatement( parser, "Way", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            local wayStr = p:readUnquotedString()
            if wayStr == "" then
                p.context:error( tok[2], "way: expected a string." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Way, wayStr } )
            return {}
        end
    )

    compiler.createStatement( parser, "Emote", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeIdentifier = true },
        function( p, _, tok )
            local nextTok = p.tokenizer:lookAhead()
            if not nextTok then
                p.context:error( tok[2], "emote: expected an identifier." )
                return {}
            end

            if nextTok[1] ~= compiler.TokenType.Identifier then
                p.context:error( nextTok[2], "emote: expected an identifier." )
                return {}
            end

            p.tokenizer:readToken()
            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Emote, nextTok.value } )
            return {}
        end
    )

    compiler.createStatement( parser, "Cast", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            -- Action statements that can suspend execution have to be on their own line
            -- so that later on we'll be able to display suspension points in the editor
            -- as an horizontal line.
            if not tok.atNewLine then
                p.context:error( tok[2], "Cast statements have to be on their own line." )
                return {}
            end

            local unitVal

            local tokState = p.tokenizer:saveState()
            local nextTok = p.tokenizer:lookAhead()

            if nextTok and nextTok[1] == compiler.TokenType.Completion then
                if not nextTok.value then -- Ignore word completion token
                    local l, c = p.context:getLoc( nextTok[2] )
                    p.context:help( "You can type the spell's name directly, or use '@' first to specify a target.", l, c )
                end

                p.tokenizer:readToken()
                tokState = p.tokenizer:saveState()
                nextTok = p.tokenizer:lookAhead()

            end

            if nextTok and nextTok[1] == compiler.TokenType.At then
                p.tokenizer:readToken()

                local cc = p.completionContext
                p.completionContext = {}
                local success = p:parseExpr( compiler.Precedence.Statement, { unit=1, rhsUnit=1 } )
                p.completionContext = cc

                unitVal = p:popValue()

                if not success or not unitVal or unitVal[1] ~= "unit" then
                    local loc = nextTok[2]
                    if unitVal then
                        loc = unitVal[2]
                    end

                    p.context:error( loc, "Unit expression expected after '@'." )
                    unitVal = nil
                end
            else
                p.tokenizer:restoreState( tokState )
            end

            local spellName = p:readUnquotedString()
            if spellName == "" then
                p.context:error( tok[2], "Cast: expected a spell name or Id." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Cast, spellName, at = unitVal } )

            if not p.inEitherBlock then
                -- Automatically insert a suspension point after spells that require it
                -- (basically spells off the gcd)
                -- TODO: need to handle this for items and toys
                -- TODO: spells like milling that require an item target
                -- can be made to work on a specific item with /use. Figure out
                -- a better way to express that than using a "either" block.
                local spellId = select( 7, GetSpellInfo( spellName ) )
                if spellId and not compiler.NonSuspensiveSpellsDB[spellId]  then
                    compiler.EmitSuspensionPoint( p.cfg, tok[2] )
                end
            end

            return {}
        end
    )

    -- TODO: target spec? suspension?
    compiler.createStatement( parser, "Use", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            -- Action statements that can suspend execution have to be on their own line
            -- so that later on we'll be able to display suspension points in the editor
            -- as an horizontal line.
            if not tok.atNewLine then
                p.context:error( tok[2], "Use statements have to be on their own line." )
                return {}
            end

            local itemName = p:readUnquotedString()
            if itemName == "" then
                p.context:error( tok[2], "Use: expected an item name." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Use, itemName } )
            return {}
        end
    )

    local slotContext = { slot=1 }
    compiler.createStatement( parser, "UseEquippedItem", { compiler.SyntaxColor.ActionStmt },
        function( p, _, tok )

            local success = p:parseExpr( compiler.Precedence.Statement, slotContext )
            local slotVal = p:popValue()

            if not success or not slotVal or slotVal[1] ~= "slot" then
                local loc = tok[2]
                if slotVal then
                    loc = slotVal[2]
                end

                p.context:error( loc, "Slot expression expected after 'UseEquippedItem'." )
                slotVal = {}
                return
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.UseEquippedItem, slotVal } )
            return {}
        end
    )

    compiler.createStatement( parser, "Click", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            -- TODO this should take an actual string expression,
            -- once we have at least string literals...
            local clickStr = p:readUnquotedString()
            if clickStr == "" then
                p.context:error( tok[2], "click: expected a string." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Click, clickStr } )
            return {}
        end
    )

    compiler.createStatement( parser, "CancelAura", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            -- TODO this should take an actual string expression,
            -- once we have at least string literals...
            local caString = p:readUnquotedString()
            if caString == "" then
                p.context:error( tok[2], "cancelaura: expected a string." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.CancelAura, caString } )
            return {}
        end
    )

    compiler.createStatement( parser, "Target", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            local targetStr = p:readUnquotedString()
            if targetStr == "" then
                p.context:error( tok[2], "Target: expected a target name." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Target, targetStr } )
            return {}
        end
    )

    -- TODO: we can just write "focus target" but it could make sense to be able to take
    -- a unit spec value. Decide what to do later
    compiler.createStatement( parser, "Focus", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true, consumeLiteralString = true },
        function( p, _, tok )
            local targetStr = p:readUnquotedString()
            if targetStr == "" then
                p.context:error( tok[2], "Focus: expected a target name." )
                return {}
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.Focus, targetStr } )
            return {}
        end
    )

    compiler.createStatement( parser, "RandomMount", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true },
        function( p )
            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.SummonRandomMount } )
            return {}
        end
    )

    compiler.createStatement( parser, "PetAttack", { compiler.SyntaxColor.ActionStmt, indentAsStatement = true },
        function( p )
            local unitVal

            local tokState = p.tokenizer:saveState()
            local nextTok = p.tokenizer:lookAhead()
            if nextTok and nextTok[1] == compiler.TokenType.At then
                p.tokenizer:readToken()

                p.expectedExpressionType = "unit"
                local success = p:parseExpr( compiler.Precedence.Statement )
                unitVal = p:popValue()
                p.expectedExpressionType = nil

                if not success or not unitVal or unitVal[1] ~= "unit" then
                    local loc = nextTok[2]
                    if unitVal then
                        loc = unitVal[2]
                    end

                    p.context:error( loc, "Unit expression expected after '@'." )
                    unitVal = nil
                end
            else
                p.tokenizer:restoreState( tokState )
            end

            p.cfg.currentInsertionPointBB:append( { compiler.CFG.OpCodes.PetAttack, at = unitVal } )
            return {}
        end
    )

    compiler.createSimpleSlashCommandStatement( parser, "StartAttack" )
    compiler.createSimpleSlashCommandStatement( parser, "StopAttack" )
    compiler.createSimpleSlashCommandStatement( parser, "TargetEnemy" )
    compiler.createSimpleSlashCommandStatement( parser, "TargetEnemyPlayer" )
    compiler.createSimpleSlashCommandStatement( parser, "ClearTarget" )
    compiler.createSimpleSlashCommandStatement( parser, "ClearFocus" )
    compiler.createSimpleSlashCommandStatement( parser, "TargetLastTarget" )
    compiler.createSimpleSlashCommandStatement( parser, "LeaveVehicle" )
    compiler.createSimpleSlashCommandStatement( parser, "PetDismiss" )
    compiler.createSimpleSlashCommandStatement( parser, "PetFollow" )
    compiler.createSimpleSlashCommandStatement( parser, "PetStay" )
    compiler.createSimpleSlashCommandStatement( parser, "PetMoveTo" )
    compiler.createSimpleSlashCommandStatement( parser, "PetAssist" )
    compiler.createSimpleSlashCommandStatement( parser, "PetPassive" )
    compiler.createSimpleSlashCommandStatement( parser, "PetDefensive" )
    compiler.createSimpleSlashCommandStatement( parser, "PetAutoCastOn" )
    compiler.createSimpleSlashCommandStatement( parser, "PetAutoCastOff" )
    compiler.createSimpleSlashCommandStatement( parser, "PetAutoCastToggle" )
end

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Print, function( _, instr )
    return "print( '" .. instr[2]:gsub( "'", "\\'" ) .. "' )"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Say, function( _, instr )
    return "m=m..'/s " .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Way, function( _, instr )
    return "m=m..'/way " .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Emote, function( _, instr )
    return "m=m..'/" .. instr[2] .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Cast, function( codegen, instr )
    local tgt = instr.at and "[@" .. codegen:generateInStrExpr( instr.at ) .. "]" or ""
    return "m=m..'/cast " .. tgt .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Use, function( _, instr )
    return "m=m..'/use " .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.UseEquippedItem, function( codegen, instr )
    return "m=m..'/use " .. codegen:generateInStrExpr( instr[2] ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Click, function( _, instr )
    return "m=m..'/click " .. instr[2]:gsub( "'", "\\'" ) .. " LeftButton\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.CancelAura, function( _, instr )
    return "m=m..'/cancelaura " .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Target, function( _, instr )
    return "m=m..'/target " .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.Focus, function( _, instr )
    return "m=m..'/focus " .. instr[2]:gsub( "'", "\\'" ) .. "\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.SummonRandomMount, function()
    return "m=m..'/script C_MountJournal.SummonByID(0)\\n'"
end )

compiler.CodeGen:RegisterOpcodeHandler( compiler.CFG.OpCodes.PetAttack, function( codegen, instr )
    local result = "m=m..'/petattack"

    if instr.at then
        result = result .. " [@" .. codegen:generateInStrExpr( instr.at ) .. "]"
    end

    return result .. "\\n'"
end )
