--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Misc control statements (suspend, break, continue)
--
local _, compiler = ...

function compiler.EmitSuspensionPoint( cfg, loc )
    local resumeBB = compiler.CFG.BasicBlock:new( cfg )
    cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Suspend, resumeBB, loc = loc }
    cfg.currentInsertionPointBB = resumeBB
end

local function suspendPredicate( p )
    return not p.inEitherBlock
end

local function breakPredicate( p )
    return not p.inEitherBlock and p.breakBB
end

local function continuePredicate( p )
    return not p.inEitherBlock and p.continueBB
end

function compiler.SetupControlStatements( parser )
    local suspendInfos =
    {
        compiler.SyntaxColor.ControlStmt,
        indentAsStatement = true,
        completionPredicate = suspendPredicate
    }

    compiler.createStatement( parser, "suspend", suspendInfos,
        function( p, _, tok )
            if p.inEitherBlock then
                p.context:error( tok[2], "Suspend can't be used inside of an \"either\" block." )
                p.context:poison()
                return {}
            end

            if p.inOnBlock then
                p.context:error( tok[2], "Suspend can't be used inside of an \"on\" block." )
                p.context:poison()
                return {}
            end

            -- Action statements that can suspend execution have to be on their own line
            -- so that later on we'll be able to display suspension points in the editor
            -- as an horizontal line.
            if not tok.atNewLine then
                p.context:error( tok[2], "Suspend statements have to be on their own line." )
                return {}
            end

            local trail = p:readUnquotedString()
            if trail ~= "" then
                p.context:error( tok[2], "Suspend statements have to be on their own line." )
                p.context:poison()
                return {}
            end

            compiler.EmitSuspensionPoint( p.cfg, tok[2] )
            return {}
        end
    )

    local breakInfos =
    {
        compiler.SyntaxColor.ControlStmt,
        indentAsStatement = true,
        completionPredicate = breakPredicate
    }

    compiler.createStatement( parser, "break", breakInfos,
        function( p, _, tok )
            if not p.breakBB then
                p.context:error( tok[2], "Break is only allowed inside of loops." )
                return {}
            end

            if p.inEitherBlock then
                p.context:error( tok[2], "Break can't be used inside of an \"either\" block." )
                p.context:poison()
                return {}
            end

            p.cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Branch, p.breakBB }
            p.cfg.currentInsertionPointBB = nil
            return {}
        end
    )

    local continueInfos =
    {
        compiler.SyntaxColor.ControlStmt,
        indentAsStatement = true,
        completionPredicate = continuePredicate
    }

    compiler.createStatement( parser, "continue", continueInfos,
        function( p, _, tok )
            if not p.continueBB then
                p.context:error( tok[2], "Continue is only allowed inside of loops." )
                return {}
            end

            if p.inEitherBlock then
                p.context:error( tok[2], "Continue can't be used inside of an \"either\" block." )
                p.context:poison()
                return {}
            end

            p.cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Branch, p.continueBB }
            p.cfg.currentInsertionPointBB = nil
            return {}
        end
    )

    compiler.createStatement( parser, "return", { compiler.SyntaxColor.ControlStmt, indentAsStatement = true },
        function( p )
            p.cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Return }
            return {}
        end
    )
end
