--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The on statement.
--
local _, compiler = ...

local boolContext = { bool=1 }
local stmtContext = { stmt=1 }

local function onPredicate( p )
    return p.cfg:isEmpty() and not p.inEitherBlock and not p.inOnBlock
end

function compiler.SetupOnStatement( parser )
    local onInfos =
    {
        compiler.SyntaxColor.ControlStmt,
        shouldIndent = true,
        completionPredicate = onPredicate
    }

    compiler.createStatement( parser, "on", onInfos,
        function( p, _, tok )
            if not onPredicate( p ) then
                p.context:error( tok[2], "On can only be used at the beginning of a macro or after another on statement." )
                p.context:poison()
                return {}
            end

            -- Save the main cfg
            local mainCFG = p.cfg
            p.inOnBlock = true

            -- Create the event handler cfg
            p.cfg = compiler.CFG:new()
            p.eventHandlers[#p.eventHandlers+1] = p.cfg

            -- Parse the condition into the entry point block
            p.expectedExpressionType = "bool"
            local success = p:parseExpr( compiler.Precedence.Statement, boolContext )
            local condVal = p:popValue()
            p.expectedExpressionType = nil

            if not success or not condVal or condVal[1] ~= "bool" then
                local loc = tok[2]
                if condVal then
                    loc = condVal[2]
                end

                p.context:error( loc, "Boolean expression expected after 'on'." )

                -- No RAII, garbage language strikes again, we have to be careful to restore
                -- shit on every exit path by hand like cavemen
                p.inOnBlock = false
                p.cfg = mainCFG
                return
            end

            -- Get the last block of the header. If we do some crazy things
            -- in some operators like shortcut evaluation,
            -- the header might actually span several basic blocks.
            local headerLastBB = p.cfg.currentInsertionPointBB

            -- Create the body BB
            local bodyBB = compiler.CFG.BasicBlock:new( p.cfg )
            p.cfg.currentInsertionPointBB = bodyBB

            -- Create the continuation BB
            local contBB = compiler.CFG.BasicBlock:new( p.cfg )

            -- Parse the body
            p.scopeStack:push()
            local ok = p:parseExpr( compiler.Precedence.Statement, stmtContext )
            p.scopeStack:pop()

            if not ok then
                p.context:poison()
                return
            end

            -- Wire up the cond branch that either goes to the body or to the continuation.
            headerLastBB.terminator = { compiler.CFG.TerminatorType.CondBranch, bodyBB, contBB, condVal }

            p.inOnBlock = false
            p.cfg = mainCFG
            return {}
        end
    )
end
