--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The while statement.
--
local _, compiler = ...

local boolContext = { bool=1 }
local stmtContext = { stmt=1 }

local function whilePredicate( p )
    return not p.inEitherBlock and not p.inOnBlock
end

function compiler.SetupWhileStatement( parser )
    local whileInfos =
    {
        compiler.SyntaxColor.ControlStmt,
        shouldIndent = true,
        completionPredicate = whilePredicate
    }

    compiler.createStatement( parser, "while", whileInfos,
        function( p, _, tok )
            if p.inEitherBlock then
                p.context:error( tok[2], "While can't be used inside of an \"either\" block." )
                p.context:poison()
                return {}
            end

            if p.inOnBlock then
                p.context:error( tok[2], "While can't be used inside of an \"on\" block." )
                p.context:poison()
                return {}
            end

            -- Create the header BB
            local headerBB = compiler.CFG.BasicBlock:new( p.cfg )

            p.cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Branch, headerBB }
            p.cfg.currentInsertionPointBB = headerBB

            -- Parse the condition into the header block
            p.expectedExpressionType = "bool"
            local success = p:parseExpr( compiler.Precedence.Statement, boolContext )
            local condVal = p:popValue()
            p.expectedExpressionType = nil

            if not success or not condVal or condVal[1] ~= "bool" then
                local loc = tok[2]
                if condVal then
                    loc = condVal[2]
                end

                p.context:error( loc, "Boolean expression expected after 'while'." )
                return
            end

            -- Tag the header with a location for the loop checker to use
            -- if it has errors to report.
            local loc = p.context.locs:createSpanningLocation( tok[2], condVal[2] )
            headerBB.loc = loc

            -- Get the last block of the header. If we do some crazy things
            -- in some operators like shortcut evaluation,
            -- the header might actually span several basic blocks.
            local headerLastBB = p.cfg.currentInsertionPointBB

            -- Create the body BB
            local bodyBB = compiler.CFG.BasicBlock:new( p.cfg )
            p.cfg.currentInsertionPointBB = bodyBB

            -- Create the continuation BB
            local contBB = compiler.CFG.BasicBlock:new( p.cfg )

            -- Setup the basic blocks to jump to when encountering a break or a continue
            -- Save the current ones first to restore them once we're done
            local breakBB, continueBB = p.breakBB, p.continueBB
            p.breakBB, p.continueBB = contBB, headerBB

            -- Parse the body
            p.scopeStack:push()
            local ok = p:parseExpr( compiler.Precedence.Statement, stmtContext )
            p.scopeStack:pop()

            p.breakBB, p.continueBB = breakBB, continueBB

            if not ok then
                p.context:poison()
                return
            end

            -- Loop back from the end of the body to the header.
            if p.cfg.currentInsertionPointBB and not p.cfg.currentInsertionPointBB.terminator then
                p.cfg.currentInsertionPointBB.terminator = { compiler.CFG.TerminatorType.Branch, headerBB }
            end

            -- Wire up the cond branch that either goes to the body or to the continuation.
            headerLastBB.terminator = { compiler.CFG.TerminatorType.CondBranch, bodyBB, contBB, condVal }

            p.cfg.currentInsertionPointBB = contBB
            return {}
        end
    )
end
