 --
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, compiler = ...

compiler.TokenType =
{
    Error = 0,
    OpenParen = 1,
    CloseParen = 2,
    OpenBrace = 3,
    CloseBrace = 4,
    At = 5,
    Identifier = 6,
    CommentBegin = 7,
    CommentEnd = 8,
    Completion = 9
}

local Tokenizer = {}
compiler.Tokenizer = Tokenizer

function Tokenizer:new( source, context, wantCommentTokens )
    local o = {}

    o.context = context
    o.str = source
    o.strlen = #source
    o.bytepos = 1
    o.lineNum = 1
    o.lineStartOffset = 1
    o.atNewLine = true
    o.blockCommentDepth = 0
    o.lastCommentLine = 0
    o.wantCommentTokens = wantCommentTokens

    setmetatable( o, { __index = self } )
    return o
end

function Tokenizer:setCursorPosition( l, c )
    self.cursorLine = l
    self.cursorCol = c
end

function Tokenizer:saveState()
    return
    {
        self.bytepos,
        self.lineNum,
        self.lineStartOffset,
        self.atNewLine,
        self.lastCommentLine,
        self.blockCommentDepth,
        self.nextToken,
        self.pendingIdentifier,
        self.backTrackingState
    }
end

function Tokenizer:restoreState( state )
    self.bytepos,
    self.lineNum,
    self.lineStartOffset,
    self.atNewLine,
    self.lastCommentLine,
    self.blockCommentDepth,
    self.nextToken,
    self.pendingIdentifier,
    self.backTrackingState
        = unpack( state )
end

function Tokenizer:setSource( source )
    self.str = source
    self.strlen = #source
end

local function eos( self )
    return self.bytepos > self.strlen
end

local function isIdentifierStartChar( c )
    -- A-Z a-z _
    return c >= 65 and c <= 90
        or c >= 97 and c <= 122
        or c == 95
        or c == 61 -- =
        or c == 46 -- .
end

local function isIdentifierChar( c )
    -- A-Z a-z _ 0-9
    return c >= 65 and c <= 90
        or c >= 97 and c <= 122
        or c == 95
        or c >= 48 and c <= 57
end

local function markLocStart( self )
    self.locStartLine = self.lineNum
    self.locStartCol = self.bytepos - self.lineStartOffset + 1
end

local function loc( self, trim )
    trim = trim or 0
    return self.context.locs:create( self.locStartLine, self.locStartCol,
        self.lineNum, self.bytepos - self.lineStartOffset + 1 - trim )
end

function Tokenizer:getCurrentLoc()
    return self.context.locs:create( self.lineNum, self.bytepos - self.lineStartOffset + 1,
        self.lineNum, self.bytepos - self.lineStartOffset + 1 )
end

local function readIdentifier( self )
    local ident = ""

    local idStartBytePos = self.bytepos

    local c = self.str:byte( self.bytepos )
    self.bytepos = self.bytepos + 1
    ident = ident .. string.char( c )

    if c == 61 then -- =
        return { compiler.TokenType.Identifier, loc( self, 1 ), value = "=" }
    end

    if c == 46 then -- .
        return { compiler.TokenType.Identifier, loc( self, 1 ), value = "." }
    end

    while not eos( self ) do
        c = self.str:byte( self.bytepos )
        if not isIdentifierChar( c ) then
            break
        end

        self.bytepos = self.bytepos + 1
        ident = ident .. string.char( c )
    end

    local location = loc( self, 1 )
    local identifier = ident:lower()

    local identTok = { compiler.TokenType.Identifier, location, value = identifier }
    if self.atNewLine then
        identTok.atNewLine = true
        self.atNewLine = false
    end

    if self.lineNum == self.cursorLine and self.bytepos - self.lineStartOffset + 1 == self.cursorCol then
        self.pendingIdentifier = identTok

        -- We don't want the completion token to advance the byte pos yet, it will be advanced
        -- when actually emitting the actual identifier token. This avoids issues when saving
        -- the context after getting the completion token, but before reading the identifier token
        -- and then deciding to retrieve the rest of the line as an unquoted string.
        self.bytepos = idStartBytePos
        return { compiler.TokenType.Completion, location, value = identifier }
    end

    return identTok
end

-- Skip comments. Handles any overlapping and nested combination of line comments
-- and block comments.
local function consumeComment( self )
    -- We are inside of a comment whenever we are still on the same line where
    -- we found a // comment introducer, or the current nesting level of comment blocks is > 0.
    --
    -- Note that we still look for /* comments introducers inside of single comment lines,
    -- and for single comment line inside of block comments.
    --
    -- This avoids two common issues:
    -- 1. // /*
    --    blah */
    --   Where the second line is not recognized as a comment because the // "cancels" the /*
    --
    -- 2. /*
    --    // blah */ stuff
    --   Where "stuff" is not recognized as commented out because the // was ignored by being inside of a comment block
    local function isInsideComment( self )
        return self.blockCommentDepth > 0 or self.lastCommentLine == self.lineNum
    end

    while not eos( self ) and isInsideComment( self ) do
        local c = self.str:byte( self.bytepos )
        self.bytepos = self.bytepos + 1

        if eos( self ) then
            return
        end

        if c == 10 then -- newline
            self.atNewLine = true
            self.lineNum = self.lineNum + 1
            self.lineStartOffset = self.bytepos
        elseif c == 47 then -- /
            c = self.str:byte( self.bytepos )
            self.bytepos = self.bytepos + 1

            if c == 47 then -- /
                self.lastCommentLine = self.lineNum
            elseif c == 42 then -- *
                self.blockCommentDepth = self.blockCommentDepth + 1
            elseif c == 10 then -- newline
                self.atNewLine = true
                self.lineNum = self.lineNum + 1
                self.lineStartOffset = self.bytepos
            end
        elseif c == 42 then -- *
            c = self.str:byte( self.bytepos )
            self.bytepos = self.bytepos + 1

            if c == 47 then -- /
                self.blockCommentDepth = self.blockCommentDepth - 1
            elseif c == 10 then -- newline
                self.atNewLine = true
                self.lineNum = self.lineNum + 1
                self.lineStartOffset = self.bytepos
            end
        end
    end
end

local function readNextToken( self )
    local pi = self.pendingIdentifier
    if pi then
        self.pendingIdentifier = nil
        self.bytepos = self.bytepos + #pi.value
        return pi
    end

    -- Skip whitespace and comments
    while not eos( self ) do
        if self.lineNum == self.cursorLine and self.bytepos - self.lineStartOffset + 1 == self.cursorCol then
            local tok = { compiler.TokenType.Completion, self:getCurrentLoc() }
            self.cursorLine = nil
            return tok
        end

        local c = self.str:byte( self.bytepos )

        if self.lastCommentLine == self.lineNum or self.blockCommentDepth > 0 then
            consumeComment( self )
            if self.wantCommentTokens and self.lastCommentLine ~= self.lineNum and self.blockCommentDepth == 0 then
                return { compiler.TokenType.CommentEnd, self:getCurrentLoc() }
            end
        elseif c == 47 and self.bytepos < self.strlen then -- /
            local cc = self.str:byte( self.bytepos + 1 )
            if cc == 47 or cc == 42 then -- / or *
                self.lastCommentLine = cc == 47 and self.lineNum or 0
                self.blockCommentDepth = cc == 42 and 1 or 0

                if self.wantCommentTokens then
                    local tok = { compiler.TokenType.CommentBegin, self:getCurrentLoc() }
                    self.bytepos = self.bytepos + 2
                    return tok
                end

                self.bytepos = self.bytepos + 2
            else
                break
            end
        elseif c == 10 then -- newline
            self.atNewLine = true
            self.lineNum = self.lineNum + 1
            self.bytepos = self.bytepos + 1
            self.lineStartOffset = self.bytepos
        elseif c == 32 or c == 13 or c == 9 then -- either newline, space, or tab
            self.bytepos = self.bytepos + 1
        else
            break
        end
    end

    if self.lineNum == self.cursorLine and self.bytepos - self.lineStartOffset + 1 == self.cursorCol then
        local tok = { compiler.TokenType.Completion, self:getCurrentLoc() }
        self.cursorLine = nil
        return tok
    end

    if eos( self ) then
        return
    end

    markLocStart( self )
    local c = self.str:byte( self.bytepos )

    if c == 40 then -- (
        self.bytepos = self.bytepos + 1
        return { compiler.TokenType.OpenParen, loc( self, 1 ) }
    end

    if c == 41 then -- )
        self.bytepos = self.bytepos + 1
        return { compiler.TokenType.CloseParen, loc( self, 1 ) }
    end

    if c == 123 then -- {
        self.bytepos = self.bytepos + 1
        return { compiler.TokenType.OpenBrace, loc( self, 1 ) }
    end

    if c == 125 then -- }
        self.bytepos = self.bytepos + 1
        return { compiler.TokenType.CloseBrace, loc( self, 1 ) }
    end

    if c == 64 then -- @
        self.bytepos = self.bytepos + 1
        return { compiler.TokenType.At, loc( self, 1 ) }
    end

    if isIdentifierStartChar( c ) then
        return readIdentifier( self )
    end

    self.bytepos = self.bytepos + 1
    return { compiler.TokenType.Error, loc( self, 1 ) }
end

function Tokenizer:lookAhead()
    if not self.nextToken then
        self.backTrackingState = self:saveState()
        self.nextToken = readNextToken( self )
    end

    return self.nextToken
end

function Tokenizer:readToken()
    if self.nextToken then
        local tok = self.nextToken
        self.nextToken = nil
        self.backTrackingState = nil
        return tok
    end

    return readNextToken( self )
end

-- Read characters verbatim until the next newline and return
-- them as a string. Whitespaces at the beginning and end are trimmed.
function Tokenizer:readUnquotedString( dontStrip )
    -- No completion can happen after an unquoted string, so if the cursor
    -- is on the current line, forget about it
    if self.lineNum == self.cursorLine then
        self.cursorLine = nil
    end

    -- Clear the cached lookahead token
    self.nextToken = nil

    -- Backtrack from the lookahead token
    if self.backTrackingState then
        self:restoreState( self.backTrackingState )
        self.backTrackingState = nil
    end

    local str = ""

    while not eos( self ) do
        local c = self.str:byte( self.bytepos )

        if c == 47 then -- /
            break
        end

        self.bytepos = self.bytepos + 1

        if c == 10 then -- newline
            self.atNewLine = true
            self.lineNum = self.lineNum + 1
            self.lineStartOffset = self.bytepos
            break
        end

        str = str .. string.char( c )
    end

    return dontStrip and str or strtrim( str )
end

Tokenizer = Tokenizer
