--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, ui = ...
local zu = ui.zu

local itGet = zu.BTree.itGet

local MacroEditorClass = {}
ui.MacroEditorClass = MacroEditorClass
setmetatable( MacroEditorClass, { __index = zu.CompositeClass } )

function MacroEditorClass:new( tpl )
    local o = {}

    o.onError = function( ... ) o:error( ... ) end
    o.onNotice = function( ... ) o:notice( ... ) end
    o.onSuspensionPoint = function( ... ) o:suspensionPoint( ... ) end
    o.onCompletion = function( ... ) o:completion( ... ) end
    o.onHelp = function( ... ) o:help( ... ) end
    o.decorations = {}
    o.pendingCompletion = {}
    o.pendingHelp = {}
    o.errorCount = 0

    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function ui.MacroEditor( tpl )
    return MacroEditorClass:new( tpl )
end

local function GetCompiler()
    if not IsAddOnLoaded( "Smack_Compiler" ) then
        EnableAddOn( "Smack_Compiler" )
        LoadAddOn( "Smack_Compiler" )
    end

    return _G["Smack_Compiler"]
end

local callbacks = {}

local function compile( timeslicer, self, updateMacro, reason )
    local comp = GetCompiler()
    if not comp then
        return
    end

    callbacks.timeslicer = timeslicer
    callbacks.onError = self.onError
    callbacks.onNotice = self.onNotice
    callbacks.onSuspensionPoint = self.onSuspensionPoint
    callbacks.onCompletion = nil
    callbacks.onSuggestion = nil
    callbacks.onHelp = nil

    local l, c = self.editor:getPrimaryCursorPosition()
    if reason == "character"
        or reason == "paste"
        or reason == "enter" then
        callbacks.onCompletion = self.onCompletion
        callbacks.onSuggestion = self.onCompletion
        callbacks.onHelp = self.onHelp
    elseif self.completionList.shown and not self.editor:isEmpty() then
        -- If the completion list is currently shown,
        -- keep displaying completions, but ignore
        -- suggestions (the intent is that we keep
        -- showing completions for the current word even
        -- after typing backspace(s) to correct it,
        -- but we hide the completion list if the entire
        -- word is erased)
        callbacks.onCompletion = self.onCompletion
    end

    self.errorCount = 0
    wipe( self.pendingCompletion )

    local source = self.editor:getText( false, timeslicer )
    local compiled = comp.CompileMacro( source, callbacks, l, c )

    -- If the macro is empty, display completions right away. Otherwise delay them
    -- to avoid shoving them in the user's face all the time which makes edition
    -- actually annoying
    if #source == 0 then
        self.completionList.apparitionDelay = 0
        self.helpPopup.apparitionDelay = 0
    else
        self.completionList.apparitionDelay = 1
        self.helpPopup.apparitionDelay = 1

        if reason == "new" then
            updateMacro = true
        end
    end

    if updateMacro then
        -- When updating the macro in the saved data, get the cleaned version
        -- of the code (trim whitespace & empty lines at end)
        --
        -- However, for compilation we don't want the cleaned version, because
        -- then the compiler won't be able to emit suggestions inside of
        -- trailing whitespace since the tokenizer will never encounter
        -- the cursor position in that case.
        local cleanedSource = self.editor:getText( true, timeslicer )
        self.macro:setSource( cleanedSource, compiled )
    end

    if self.pendingHelp[1] then
        self.helpPopup:setText( self.pendingHelp[1] )
        self.helpPopup:show( self.pendingHelp[2], self.pendingHelp[3] )
        self.pendingHelp[1] = nil
    elseif self.pendingCompletion[1] then
        self.completionList:setProvider( self.pendingCompletion[1] )
        self.completionList:show( self.pendingCompletion[2], self.pendingCompletion[3] )
    else
        self.completionList:hide()
        self.helpPopup:hide()
    end

    self.editor.model:clearDecorations()

    local decs = self.decorations
    for i=1,#decs,1 do
        self.editor.model:addDecoration( decs[i] )
    end
    wipe( decs )

    self.editor.model:updateDecorations()
end

local errorMessageTpl = { minimize = true }
local popupTpl = { style = "tooltip", channel = "tooltip" }
local errorTpl =
{
    r = 1, g = 0.1, b = 0.2, a = 1
}

function MacroEditorClass:error( message, fl, fc, ll, lc )
    if self.errorCount >= 10 then
        return
    end

    self.errorCount = self.errorCount + 1

    local text = zu.CreateWidget( zu.ErrorClass, errorMessageTpl )
    text.text = message

    local popup = zu.CreateWidget( zu.PopupClass, popupTpl )
    popup.skin = self:getSkin()
    popup.content = text

    local dec = zu.CreateWidget( zu.SquiggleDecorationClass, errorTpl )

    dec.firstLine = fl
    dec.firstCol = fc
    dec.lastLine = ll
    dec.lastCol = lc
    dec.popup = popup

    self.decorations[#self.decorations+1] = dec
end

local noticeMessageTpl = { minimize = true }
local noticeTpl =
{
    r = 1, g = 0.8, b = 0, a = 1
}

function MacroEditorClass:notice( message, fl, fc, ll, lc )
    local text = zu.CreateWidget( zu.WarningClass, noticeMessageTpl )
    text.text = message

    local popup = zu.CreateWidget( zu.PopupClass, popupTpl )
    popup.skin = self:getSkin()
    popup.content = text

    local dec = zu.CreateWidget( zu.SquiggleDecorationClass, noticeTpl )

    dec.firstLine = fl
    dec.firstCol = fc
    dec.lastLine = ll
    dec.lastCol = lc
    dec.popup = popup

    self.decorations[#self.decorations+1] = dec
end

local spTpl = { r = 0.76, g = 0.6, b = 1 }

function MacroEditorClass:suspensionPoint( line )
    local dec = zu.CreateWidget( zu.HRuleDecorationClass, spTpl )
    dec:setLine( line )
    self.decorations[#self.decorations+1] = dec
end

function MacroEditorClass:completion( provider, l, c )
    self.pendingCompletion[1] = provider
    self.pendingCompletion[2] = l
    self.pendingCompletion[3] = c
end

function MacroEditorClass:help( text, l, c )
    self.pendingHelp[1] = text
    self.pendingHelp[2] = l
    self.pendingHelp[3] = c
end

local function recompile( self, updateMacro, reason )
    -- If we already have a compilation in progress, cancel it
    if zu.compTaskHandle then
        zu.CancelTimeSlicedFunction( zu.compTaskHandle )
    end

    zu.compTaskHandle = zu.TimeSlice( compile, 1, self, updateMacro, reason )
end

local function macroSourceChanged( self, reason )
    recompile( self, true, reason )
end

local function ColorizeTag( tag )
    local smack = _G["Smack_Main"]
    local ti = smack.GetSpecialTagInfos( tag )
    return ti and ti.displayName or tag
end

local deleteDialogTpl =
{
    okLabel = "Delete",
}

local it = {}
local dropLineDecTpl = { r = 0.4, g = 0, b = 1, above = true, thickness = 4 }
local function mouseHoverUpdate( self, hover, l, line, relativeYPosition )
    if not line then
        return
    end

    local dec = self.dropLineDec

    -- If the cursor is in the bottom half of the line, take the next line.
    -- This makes the drop line follow the mouse cursor more closely and feels
    -- better.
    if relativeYPosition > ( line.height * 0.5 ) then
        l = l + 1
        l = min( l, self.editor.model.btree:size() )
    end

    if dec and l ~= dec:line() then
        self.editor.model:removeDecoration( dec )
        zu.ReleaseWidget( dec )
        self.dropLineDec = nil
        dec = nil
    end

    local infoType = GetCursorInfo()
    local wantDrop = hover and infoType and ( infoType == "spell" or infoType == "item" or infoType == "mount" )
    if not wantDrop then
        if dec then
            self.editor.model:removeDecoration( dec )
            zu.ReleaseWidget( dec )
            self.dropLineDec = nil
        end
    elseif not dec then
        dec = zu.CreateWidget( zu.HRuleDecorationClass, dropLineDecTpl )
        dec:setLine( l )
        self.editor.model:addDecoration( dec )
        self.dropLineDec = dec

        if l > 1 then
            l = l - 1
            -- Directly accessing it, fuck it, cba
            -- making a bunch of nice accessors this time around
            self.editor.model.btree:findByIndex( l, it )
            local prevLine = itGet( it )
            self.dropColumn = strlenutf8( prevLine[2] ) + 1
        else
            l = 0
            self.dropColumn = 1
        end

        self.dropLine = l
    end

    self.editor.model:updateDecorations()
end

local function handleDrop( self )
    if not self.dropLineDec then
        return
    end

    local l = self.dropLine
    local c = self.dropColumn

    local infoType, itemID, mountIndex, spellID = GetCursorInfo()
    if not infoType then
        return
    end

    self.editor:setPrimaryCursorPosition( max( l, 1 ), c )

    local textToPaste

    if infoType == "spell" then
        local spellName = GetSpellInfo( spellID )
        textToPaste = "Cast " .. spellName
    elseif infoType == "item" then
        local itemName = C_Item.GetItemNameByID(itemID)
        textToPaste = "Use " .. itemName
    elseif infoType == "mount" then
        local MountID = C_MountJournal.GetMountIDs()[mountIndex]
        local mountName = C_MountJournal.GetMountInfoByID(MountID)
        textToPaste = "Use " .. mountName
    end

    if textToPaste then
        if l > 0 then
            textToPaste = "\n" .. textToPaste
        else
            textToPaste = textToPaste .. "\n"
        end

        self.editor:pasteText( textToPaste, "dragdrop" )
        self.editor:focus()
    end

    ClearCursor()
end

-- If we gain focus and the editor is empty,
-- run the compiler to generate suggestions
local function gainFocus( self )
    if self.editor:isEmpty() then
        recompile( self, false, "enter" )
    end
end

local function hidePopupsOnMouseDown( self )
    if self.editor:haveFocus() and self.editor:isEmpty() then
        return
    end
    self.completionList:hide()
    self.helpPopup:hide()
end

local function hidePopups( self )
    self.completionList:hide()
    self.helpPopup:hide()
end

function MacroEditorClass:create( parentFrame )
    if not self.content then
        self.editor = zu.Editor
        {
            colorizer = GetCompiler().MacroColorizer( self:getSkin().syntaxColorPalette ),
            indenter = GetCompiler().MacroIndenter(),
            mouseHoverUpdate = function( hover, l, line, relativeYPosition )
                mouseHoverUpdate( self, hover, l, line, relativeYPosition )
            end,
            onMouseUp = function() handleDrop( self ) end,
            onGainFocus = function() gainFocus( self ) end,
            onMouseDown = function() hidePopupsOnMouseDown( self ) end,
            onLoseFocus = function() hidePopups( self ) end,
            onViewScrolled = function() hidePopups( self ) end
        }

        self.completionList = ui.MacroCompletionList:new( self.editor )
        self.completionList.skin = self:getSkin()

        self.helpPopup = ui.MacroHelp:new( self.editor )
        self.helpPopup.skin = self:getSkin()

        self.nameEditBox = zu.EditBox
        {
            onTextEditionFinished = function( text )
                if self.macro then
                    self.macro:setMetadata( text, self.macro.keybind )
                end
            end
        }

        self.keybind = ui.KeybindButton
        {
            onKeyBindChanged = function( kb )
                self.macro:setMetadata( self.macro.name, kb )
            end
        }

        self.keybindHider = zu.Hider
        {
            content = self.keybind,
            contentHidden = self.macro and not self.macro.properties.enabled
        }

        self.tagEditor = zu.TagEditor { colorizer = ColorizeTag }

        self.deleteConfirmationDialog = zu.Dialog( deleteDialogTpl )

        self.deleteConfirmationDialog.onOk = function()
            self.listController:deleteElement( self.macro )
            self:getWindow():setOverlay( nil )
        end

        self.deleteConfirmationDialog.onCancel = function()
            self:getWindow():setOverlay( nil )
        end

        zu.AddListener( self.editor.textChangedEvent, self, macroSourceChanged )

        self.content = zu.VerticalLayout
        {
            zu.HorizontalLayout
            {
                zu.Button
                {
                    type = "trash",
                    onClick = function()
                        local hardSpacesName = self.macro.name:gsub( " ", "\194\160" )
                        local text = string.format( "Do you really want to delete %s?",
                            hardSpacesName )
                        self.deleteConfirmationDialog:setText( text )
                        self:getWindow():setOverlay( self.deleteConfirmationDialog )
                    end
                },
                self.nameEditBox,
                self.keybindHider
            },
            self.tagEditor,
            self.editor
        }

        self.content.parent = self
    end

    self.content:create( parentFrame )
end

function MacroEditorClass:hide()
    hidePopups( self )
    zu.CompositeClass.hide( self )
end

function MacroEditorClass:freeze()
    hidePopups( self )
    zu.CompositeClass.freeze( self )
end

local function update( self )
    self.nameEditBox:setText( self.macro.name )
    self.keybind:setKeybind( self.macro.keybind, self.macro.name )

    if self.macro.properties.enabled then
        self.keybindHider:showContent()
    else
        self.keybindHider:hideContent()
    end
end

function MacroEditorClass:setMacro( macro, isNew )
    if self.macro then
        zu.RemoveListener( self.macro.updateEvent, self )
    end

    if macro == nil then
        self.nameEditBox:setText( "" )
        self.editor:setText( "" )
        self.keybind:setKeybind()
        self.tagEditor:setModel()
        return
    end

    self.macro = macro
    self.nameEditBox:setText( self.macro.name )
    self.editor:setText( self.macro[1] )
    self.keybind:setKeybind( self.macro.keybind, self.macro.name )
    self.tagEditor:setModel( self.macro.tags )

    if self.macro.properties.enabled then
        self.keybindHider:showContent()
    else
        self.keybindHider:hideContent()
    end

    zu.AddListener( self.macro.updateEvent, self, update )

    -- Always recompile macros when they are opened by the editor
    -- so that we recreate any relevant text decoration (not only errors
    -- but hints as well).
    -- However, don't update the macro in that case to avoid displaying
    -- spurious "changes are pending until out of combat" messages
    -- when browsing through macros while in combat.
    recompile( self, false, isNew and "new" or nil )
end

function MacroEditorClass:selectName()
    self.nameEditBox:select()
end
