--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, ui = ...
local zu = ui.zu

local KeybindButtonClass = {}
ui.KeybindButtonClass = KeybindButtonClass
setmetatable( KeybindButtonClass, { __index = zu.CompositeClass } )

function KeybindButtonClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function ui.KeybindButton( tpl )
    return KeybindButtonClass:new( tpl )
end

local function applyKeybind( self, key )
    self.key = key
    self.button:setLabel( key or "not bound" )

    if self.onKeyBindChanged then
        self.onKeyBindChanged( key )
    end
end

local function onClick( self )
    self.binding = true

    local actionHardSpacesName = self.action:gsub( " ", "\194\160" )

    local text = string.format( "Press a new key to bind to %s (currently bound to %s)",
        actionHardSpacesName, self.key or "nothing" )
    self.bindingDialog:setText( text )

    self:getWindow():setOverlay( self.bindingDialog )

    -- Directly attaching events to the frame created by the text widget contained
    -- in the dialog. This is ugly as hell but there is literally no point in
    -- making something nicer for this.
    self.bindCatcherFrame = self.bindingDialog.textWidget.frame
    self.bindCatcherFrame:EnableKeyboard( true )
    self.bindCatcherFrame:SetPropagateKeyboardInput( false )
    self.bindCatcherFrame:SetScript( "OnKeyDown", function( _, key ) self:onKeyDown( key ) end )
end

function KeybindButtonClass:create( parentFrame )
    if not self.content then
        self.bindingDialog = zu.Dialog
        {
            okLabel = "Clear binding",

            onCancel = function()
                self:getWindow():setOverlay( nil )
            end,

            onOk = function()
                applyKeybind( self )
                self:getWindow():setOverlay( nil )
            end
        }

        self.rebindConfirmationDialog = zu.Dialog
        {
            okLabel = "Rebind",

            onOk = function()
                applyKeybind( self, self.pendingKey )
                self:getWindow():setOverlay( nil )
            end,

            onCancel = function()
                applyKeybind( self, self.key )
                self:getWindow():setOverlay( nil )
            end
        }

        self.button = zu.Button
        {
            type = "keybind",
            label = self.key or "not bound",
            onClick = function() onClick( self ) end
        }

        self.content = self.button

        self.content.parent = self
        self.content:create( parentFrame )

        if InCombatLockdown() then
            self.button:disable()
        end

        self.button.frame:RegisterEvent( "PLAYER_REGEN_DISABLED" )
        self.button.frame:RegisterEvent( "PLAYER_REGEN_ENABLED" )
        self.button.frame:SetScript( "OnEvent", function( _, e )
            if e == "PLAYER_REGEN_DISABLED" then
                self:getWindow():setOverlay( nil )
                self.binding = false
                self.pendingKey = nil
                self.button:disable()
            elseif e == "PLAYER_REGEN_ENABLED" then
                self.button:enable()
            end
        end )
    end
end

function KeybindButtonClass:setKeybind( key, action )
    self.key = key
    self.button:setLabel( key or "not bound" )
    self.action = action
end

local function validateKeybind( self, key )
    local smack = _G["Smack_Main"]

    -- Conflict detection
    local oldAction = ""

    local oldMacro = smack.FindMacroFromKey( key )
    if oldMacro then
        oldAction = oldMacro.name
    else
        oldAction = GetBindingAction( key )
    end

    if oldAction ~= "" and oldAction ~= self.action then
        self.pendingKey = key

        local oldHardSpacesName = oldAction:gsub( " ", "\194\160" )
        local newHardSpacesName = self.action:gsub( " ", "\194\160" )

        local text = string.format( "%s is currently bound to %s.\nDo you really want to rebind it to %s?",
            key, oldHardSpacesName, newHardSpacesName )

        self.rebindConfirmationDialog:setText( text )
        self:getWindow():setOverlay( self.rebindConfirmationDialog )
    else
        self:getWindow():setOverlay( nil )
        applyKeybind( self, key )
    end
end

function KeybindButtonClass:onKeyDown( key )
    if not self.binding then
        self.bindCatcherFrame:SetPropagateKeyboardInput( true )
        return
    end

    local keyPressed = GetConvertedKeyOrButton( key )

    if not IsKeyPressIgnoredForBinding( keyPressed ) then
        keyPressed = CreateKeyChordStringUsingMetaKeyState( keyPressed )

        if self.key ~= keyPressed then
            self.binding = false
            validateKeybind( self, keyPressed )
        end
    end
end
