--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, ui = ...
local zu = ui.zu

local MacroListClass = {}
ui.MacroListClass = MacroListClass
setmetatable( MacroListClass, { __index = zu.CompositeClass } )

function MacroListClass:new( macroListModel, macroListController )
    local o =
    {
        macroListModel = macroListModel,
        macroListController = macroListController
    }
    setmetatable( o, { __index = self } )
    return o
end

function ui.MacroList( macroListModel, macroListController )
    return MacroListClass:new( macroListModel, macroListController )
end

local function updateFilterVisibility( self )
    if #self.macroListModel.allTags > 0 then
        self.filterHider:showContent()
    else
        self.filterHider:hideContent()
    end
end

local function ColorizeTag( tag )
    local smack = _G["Smack_Main"]
    local ti = smack.GetSpecialTagInfos( tag )
    return ti and ti.displayName or tag
end

local function GetActionAndNameFromCursor()
    local infoType, itemID, mountIndex, spellID = GetCursorInfo()
    if not infoType then
        return
    end

    local name = ""
    local body = ""

    if infoType == "spell" then
        name = GetSpellInfo( spellID )
        action = "Cast"
    elseif infoType == "item" then
        name = C_Item.GetItemNameByID(itemID)
        action = "Use"
    elseif infoType == "mount" then
        local MountID = C_MountJournal.GetMountIDs()[mountIndex]
        name = C_MountJournal.GetMountInfoByID(MountID)
        action = "Use"
    end

    return name, action
end

local function mouseHoverUpdate( self, hover )
    local name = GetActionAndNameFromCursor()

    self.wantDrop = hover and name

    if self.wantDrop then
        self.popup.content:setText( 'Create "'..name..'" macro' )

        local x, y = GetCursorPosition()
        y = 768 - y

        local uiScale = UIParent:GetEffectiveScale()
        x = x / uiScale
        y = y / uiScale

        local xoffset = 20
        local yoffset = -10

        self.popup:show( zu.Scale( x + xoffset ), zu.Scale( y + yoffset ), nil, "BOTTOMLEFT", nil, false )
    else
        self.popup:hide()
    end

    return self.wantDrop
end

local function mouseUp( self )
    if not self.wantDrop then
        return
    end

    local name, action = GetActionAndNameFromCursor()
    if not name then
        return
    end

    self.macroListController:createElement( name, action .. " " .. name )

    self.popup:hide()
    ClearCursor()
end

function MacroListClass:create( parentFrame )
    if not self.content then
        local tagSelector = zu.TagSelector { colorizer = ColorizeTag }
        tagSelector:setModel( self.macroListModel.allTags )
        tagSelector:setOutputModel( self.macroListModel.filterTags )

        local popupText = zu.Plus { minimize = true }
        self.popup = zu.Popup{ style = "tooltip", channel = "tooltip", apparitionDelay = 0 }
        self.popup.skin = self:getSkin()
        self.popup.content = popupText

        self.filterHider = zu.Hider
        {
            contentHidden = true,
            content = zu.Collapsible
            {
                name = "FilterCollapsible",
                title = "Filter by tag",
                tagSelector,
                collapsed = true
            }
        }

        local onlyCurSpecHider = zu.Hider
        {
            content = zu.CheckBox
            {
                name = "OnlyCurrentSpec",
                label = "current spec only",
                checked = true,
                onClick = function( checked )
                    self.macroListModel:setOnlyCurrentSpec( checked )
                end
            }
        }

        zu.AddListener( self.macroListModel.allTags.tagAddedEvent, self, updateFilterVisibility )
        zu.AddListener( self.macroListModel.allTags.tagRemovedEvent, self, updateFilterVisibility )

        self.macroListController.mouseHoverUpdate = function( _, hover )
            return mouseHoverUpdate( self, hover )
        end

        self.macroListController.mouseUp = function()
            mouseUp( self )
        end

        self.content = zu.VerResizeGroup
        {
            name = "MacroBrowserResizeGroup",

            zu.EditBox
            {
                prompt = "Filter by name",
                onTextChanged = function( text )
                    self.macroListModel:setSearchFilter( text )
                end
            },

            zu.CheckBox
            {
                name = "OnlyUnbound",
                label = "only not bound",
                checked = false,
                onClick = function( checked )
                    self.macroListModel:setOnlyUnbound( checked )
                end
            },

            zu.CheckBox
            {
                name = "OnlyCurrentClass",
                label = "current class only",
                checked = true,
                onClick = function( checked )
                    self.macroListModel:setOnlyCurrentClass( checked )

                    if checked then
                        onlyCurSpecHider.content:setChecked( false )
                        onlyCurSpecHider:showContent()
                    else
                        onlyCurSpecHider:hideContent()
                    end
                end
            },

            onlyCurSpecHider,
            self.filterHider,

            zu.List
            {
                model = self.macroListModel,
                controller = self.macroListController
            }
        }

        self.content.parent = self
    end

    self.content:create( parentFrame )
end

function MacroListClass:release()
    zu.RemoveListener( self.macroListModel.allTags.tagAddedEvent, self )
    zu.RemoveListener( self.macroListModel.allTags.tagRemovedEvent, self )
end
