--
-- Smack, a structured macro language for WoW
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, ui = ...
local zu = ui.zu

local MacroListModel = {}
ui.MacroListModel = MacroListModel

local function update( self )
    zu.Raise( self.updateEvent )
end

-- NOTE: possible performance hazard there, because SpecialTagsCheckIsForCurrentClass
-- and SpecialTagsCheckForSpecificClass both enumerate all the tags.
-- We could cache their result to avoid this. I'm just lazy so
-- I'll just wait until some crazy guy with a million macros and a thousands tags
-- starts complaining I guess
local function sortingPredicate( a, b )
    local smack = _G["Smack_Main"]

    -- Current class macros come first, enabled or not
    local acc = smack.SpecialTagsCheckIsForCurrentClass( a.tags )
    local bcc = smack.SpecialTagsCheckIsForCurrentClass( b.tags )

    if acc and not bcc then
        return true
    end

    if bcc and not acc then
        return false
    end

    -- Enabled macros come first
    if a.properties.enabled ~= b.properties.enabled then
        return a.properties.enabled
    end

    -- Class specific macros come first
    local ati = smack.SpecialTagsCheckForSpecificClass( a.tags )
    local bti = smack.SpecialTagsCheckForSpecificClass( b.tags )

    if ati and not bti then
        return true
    end

    if bti and not ati then
        return false
    end

    -- Classes are grouped together, by alphabetical order
    if ati and bti and ati.cname ~= bti.cname then
        return ati.cname < bti.cname
    end

    -- With everything else equal, macro are sorted by their indices
    -- (which are already kept in alphabetical order by the main addon)
    return a.index < b.index
end

local function sortList( self )
    local sortedList = self.sortedList
    wipe( sortedList )

    local smack = _G["Smack_Main"]
    for i=1,smack.GetMacroCount() do
        local macro = smack.GetMacro( i )
        sortedList[#sortedList+1] = macro
    end

    sort( sortedList, sortingPredicate )

    wipe( self.sortedIndices )
    for i=1,#sortedList,1 do
        local macro = sortedList[i]
        self.sortedIndices[macro] = i
    end
end

local function checkSearchFilter( self, macro )
    local searchFilter = self.searchFilter
    local n = macro.name:lower()

    for i = 1, #searchFilter, 1 do
        local searchTerm = searchFilter[i]

        if not n:find( searchTerm, 1, true ) then
            return false
        end
    end

    return true
end

local function isFilteringOn( self )
    return self.onlyUnbound or self.onlyCurrentClass or self.onlyCurrentClass or #self.filterTags > 0 or #self.searchFilter > 0
end

local function filterList( self )
    self.needFilterUpdate = false
    wipe( self.filteredList )
    wipe( self.filteredIndices )

    local sortedList = self.sortedList
    local smack = _G["Smack_Main"]

    for i = 1, #sortedList do
        local macro = sortedList[i]

        if  ( not self.onlyUnbound or macro.keybind == nil ) and
            ( not self.onlyCurrentClass or smack.SpecialTagsCheckIsForCurrentClass( macro.tags ) ) and
            ( not self.onlyCurrentSpec or macro.properties.enabled ) and
            self.filterTags:isSubSetOf( macro.tags ) and checkSearchFilter( self, macro ) then
            local index = #self.filteredList + 1
            self.filteredList[index] = macro
            self.filteredIndices[macro] = index
        end
    end
end

local function updateFilters( self )
    filterList( self )
    update( self )
end

local function addTagToAllTags( self, tag )
    local lc = tag:lower()
    local count = self.tagCounters[lc]

    if count == nil then
        self.tagCounters[lc] = 1
        self.allTags:add( lc )
        self.allTags:sort()
    else
        self.tagCounters[lc] = count + 1
    end
end

local function tagAdded( self, _, tag )
    addTagToAllTags( self, tag )
end

local function tagRemoved( self, _, tag )
    local lc = tag:lower()
    local count = self.tagCounters[lc]

    if count == 1 then
        self.tagCounters[lc] = nil
        self.allTags:remove( lc )
    else
        self.tagCounters[lc] = count - 1
    end
end

local function macroUpdated( self )
    sortList( self )
    filterList( self )
    update( self )
end

local function macroCreated( self, macro )
    local tags = macro.tags
    zu.Tags:new( tags )

    zu.AddListener( tags.tagAddedEvent, self, tagAdded )
    zu.AddListener( tags.tagRemovedEvent, self, tagRemoved )
    zu.AddListener( macro.updateEvent, self, macroUpdated )

    for i = 1, #tags, 1 do
        addTagToAllTags( self, tags[i] )
    end

    sortList( self )
    filterList( self )
    update( self )
end

local function macroDeleted( self, macro )
    local tags = macro.tags
    local needFilterUpdate = self.filterTags:isSubSetOf( tags ) or checkSearchFilter( self, macro.name )

    for i = 1, #tags, 1 do
        tagRemoved( self, _, tags[i] )
    end

    sortList( self )

    if needFilterUpdate then
        filterList( self )
    end

    update( self )
end

function MacroListModel:new()
    local o = {}
    o.updateEvent = {}

    o.sortedList = {}
    o.sortedIndices = {}

    o.filteredList = {}
    o.filteredIndices = {}

    o.allTags = zu.Tags:new {}
    o.filterTags = zu.Tags:new {}

    o.tagCounters = {}

    o.searchFilter = {}
    o.onlyUnbound = false
    o.onlyCurrentClass = true
    o.onlyCurrentSpec = true

    zu.AddListener( o.filterTags.tagAddedEvent, o, updateFilters )
    zu.AddListener( o.filterTags.tagRemovedEvent, o, updateFilters )

    local smack = _G["Smack_Main"]
    o.modelResetEvent = smack.CacheClearedEvent

    setmetatable( o, { __index = self } )
    return o
end

local function mergeTags( self, tags )
    for i = 1, #tags, 1 do
        tagAdded( self, _, tags[i] )
    end
end

function MacroListModel:create()
    if self.initDone then
        return
    end

    self.initDone = true

    local smack = _G["Smack_Main"]
    smack.AddListener( smack.MacroCreatedEvent, self, macroCreated )
    smack.AddListener( smack.MacroDeletedEvent, self, macroDeleted )
    smack.AddListener( smack.MacroListUpdateEvent, self, update )

    for i = 1, smack.GetMacroCount() do
        local macro = smack.GetMacro( i )
        self[#self+1] = macro

        local tags = macro.tags
        mergeTags( self, tags )

        zu.Tags:new( tags )
        zu.AddListener( tags.tagAddedEvent, self, tagAdded )
        zu.AddListener( tags.tagRemovedEvent, self, tagRemoved )
        zu.AddListener( macro.updateEvent, self, macroUpdated )
    end

    sortList( self )
    filterList( self )
end

function MacroListModel:release()
    local smack = _G["Smack_Main"]
    smack.RemoveListener( smack.MacroCreatedEvent, self )
    smack.RemoveListener( smack.MacroDeletedEvent, self )
    smack.RemoveListener( smack.MacroListUpdateEvent, self )
    wipe( self )
end

function MacroListModel:count()
    return isFilteringOn( self ) and #self.filteredList or #self.sortedList
end

function MacroListModel:getElementSize( view )
    local w = zu.CreateWidget( ui.MacroItemClass )
    w.parent = view
    w.model = self
    w:create( view.frame )
    w:calcMinMax()

    local height, minWidth, prefWidth, maxWidth, levelRange
        = w.minHeight, w.minWidth, w.preferredWidth, w.maxWidth, w.levelRange
    zu.ReleaseWidget( w )

    return height, minWidth, prefWidth, maxWidth, levelRange
end

function MacroListModel:getWidget( index )
    local w = zu.CreateWidget( ui.MacroItemClass, {} )
    w.model = self
    w:setMacro( self:getElement( index ) )
    return w
end

function MacroListModel:getElement( index )
    return isFilteringOn( self ) and self.filteredList[index] or self.sortedList[index]
end

function MacroListModel:getElementIndex( element )
    if isFilteringOn( self ) then
        return self.filteredIndices[element] or 0
    else
        return self.sortedIndices[element] or 0
    end
end

function MacroListModel:createElement( name, body )
    local smack = _G["Smack_Main"]
    body = body or ""

    local filterTags = self.filterTags
    filterTags:sort()

    local tags = zu.Tags:new()

    if self.onlyCurrentClass then
        local _, playerClass = UnitClassBase( "player" )
        local cinfo = C_CreatureInfo.GetClassInfo( playerClass )
        local cname = cinfo.classFile:lower()

        tags:add( cname )
    end

    if self.onlyCurrentSpec then
        local specId = GetSpecializationInfo( GetSpecialization() )
        local sname = smack.SpecNames[specId]:lower()
        if sname and sname ~= "initial" then
            tags:add( sname )
        end
    end

    for i=1,#filterTags,1 do
        tags:add( filterTags[i] )
    end

    return smack.Macro.Create( name, body, tags )
end

function MacroListModel:deleteElement( element )
    element:delete()
end

function MacroListModel:setSearchFilter( text )
    local searchFilter = self.searchFilter
    wipe( searchFilter )

    local wordList = { strsplit( " ", text ) }

    for i = 1, #wordList, 1 do
        local word = strtrim( wordList[i] )
        if word ~= "" then
            searchFilter[#searchFilter+1] = word:lower()
        end
    end

    updateFilters( self )
end

function MacroListModel:setOnlyUnbound( onlyUnbound )
    if self.onlyUnbound == onlyUnbound then
        return
    end

    self.onlyUnbound = onlyUnbound

    if self.initDone then
        updateFilters( self )
    end
end

function MacroListModel:setOnlyCurrentClass( onlyCurrentClass )
    if self.onlyCurrentClass == onlyCurrentClass then
        return
    end

    self.onlyCurrentClass = onlyCurrentClass
    if not onlyCurrentClass then
        self.onlyCurrentSpec = false
    end

    if self.initDone then
        updateFilters( self )
    end
end

function MacroListModel:setOnlyCurrentSpec( onlyCurrentSpec )
    if self.onlyCurrentSpec == onlyCurrentSpec then
        return
    end

    self.onlyCurrentSpec = onlyCurrentSpec
    if onlyCurrentSpec then
        self.onlyCurrentClass = true
    end

    if self.initDone then
        updateFilters( self )
    end
end
