--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor

local SliderClass = {}
zu.SliderClass = SliderClass
setmetatable( SliderClass, { __index = zu.Widget } )
SliderClass.levelRange = 2

SliderClass.total = 10
SliderClass.visible = 1
SliderClass.value = 0
SliderClass.bigIncrement = 1        -- Amount to move when clicking in the slider's background.
SliderClass.minKnobSize = 16
SliderClass.background = true
SliderClass.fade = false            -- User requested fadeout
SliderClass.autoFadeOut = false     -- Automatic fade out when the amount visible is less or equal than the total
SliderClass.fadeOutState = false
SliderClass.alpha = 1
SliderClass.onValueUpdated = function() end

function SliderClass:new( tpl )
    local o = {}

    o.startDraggingEvent = {}
    o.endDraggingEvent = {}

    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function zu.Slider( s )
    return SliderClass:new( s )
end

function zu.HorizontalSlider( s )
    return SliderClass:new( s )
end

function zu.VerticalSlider( s )
    s.vertical = true
    return SliderClass:new( s )
end

local function updateKnob( self )
    if self.total < self.visible then
        self.value = 0
    else
        self.value = min( self.value, self.total - self.visible )
    end

    local width = nil
    local length = nil

    if self.vertical then
        width, length = self.frame:GetSize()
    else
        length, width = self.frame:GetSize()
    end

    local kPos = 0
    local kLength = length

    if self.total > self.visible then
        kLength = ( self.visible * length ) / self.total
        kLength = floor( kLength + 0.5 )
        kLength = max( kLength, self.minKnobSize )

        kPos = ( self.value * ( length - kLength ) ) / ( self.total - self.visible )
        kPos = floor( kPos + 0.5 )
        kPos = min( kPos, length - kLength )
    end

    if self.vertical then
        self.kFrame:SetSize( width, kLength )
        self.kFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 0, -kPos )
    else
        self.kFrame:SetSize( kLength, width )
        self.kFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", kPos, 0 )
    end
end

local function shouldBeFadedOut( self )
    return not self.isDragging and ( self.fade or self.autoFadeOut and self.visible >= self.total )
end

local function updateFading( self )
    local shouldFade = shouldBeFadedOut( self )
    local isFaded = self.fadeOutState

    if shouldFade and not isFaded then
        self.fadeOutState = true
        self.showAnim:Stop()
        self.hideAnim:Play()
    elseif not shouldFade and isFaded then
        self.fadeOutState = false
        self.hideAnim:Stop()
        self.showAnim:Play()
    end
end

function SliderClass:create( parentFrame )
    if not self.decorator then
        self.decorator = self:getSkin().SliderDecorator:new( self )
    end

    if self.frame then
        self.frame:SetParent( parentFrame )
    else
        self.frame = CreateFrame( "Frame", nil, parentFrame )
        self.kFrame = CreateFrame( "Frame", nil, self.frame )

        if self.background then
            self.bgtex = self.frame:CreateTexture( nil, "BACKGROUND" )
            self.bgtex:SetAllPoints( self.frame )
        end

        self.ktex = self.kFrame:CreateTexture( nil, "BACKGROUND" )
        self.ktex:SetAllPoints( self.kFrame )

        if not self.showAnim then
            self.showAnim = self.kFrame:CreateAnimationGroup()
            self.showAnim:SetToFinalAlpha( true )
            local anim = self.showAnim:CreateAnimation( "Alpha" )
            anim:SetFromAlpha( 0 )
            anim:SetToAlpha( 1 )
            anim:SetDuration( 0.1 )
        end

        if not self.hideAnim then
            self.hideAnim = self.kFrame:CreateAnimationGroup()
            self.hideAnim:SetToFinalAlpha( true )
            local anim = self.hideAnim:CreateAnimation( "Alpha" )
            anim:SetFromAlpha( 1 )
            anim:SetToAlpha( 0 )
            anim:SetDuration( 1 )
        end

        self.decorator:apply()

        -- Background click handler
        self.frame:EnableMouse( true )
        self.frame:SetScript( "OnMouseDown",
            function()
                if self.total < self.visible then
                    return
                end

                local kX, kY = self.kFrame:GetScaledRect()
                local mouseX, mouseY = GetCursorPosition()
                local newValue = self.value

                if self.vertical then
                    if mouseY < kY then
                        newValue = newValue + self.bigIncrement
                    else
                        newValue = newValue - self.bigIncrement
                    end
                else
                    if mouseX > kX then
                        newValue = newValue + self.bigIncrement
                    else
                        newValue = newValue - self.bigIncrement
                    end
                end

                newValue = max( newValue, 0 )
                newValue = min( newValue, self.total - self.visible )

                if self.value ~= newValue then
                    self.value = newValue
                    self.onValueUpdated( newValue )
                end

                updateKnob( self )
            end )

        -- knob click handler
        local kMouseOffset = 0
        local startPos = nil
        local length = nil
        local kLength = nil
        local scale = nil
        local maxPos = nil

        local kUpdateFunction = function()
            local mouseX, mouseY = GetCursorPosition()
            mouseY = 768 - mouseY
            local mousePos = nil

            if self.vertical then
                mousePos = mouseY
            else
                mousePos = mouseX
            end

            local kPos = ( mousePos - kMouseOffset - startPos ) * scale
            kPos = floor( kPos + 0.5 )

            kPos = max( kPos, 0 )
            kPos = min( kPos, maxPos )

            if self.vertical then
                self.kFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 0, -kPos )
            else
                self.kFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", kPos, 0 )
            end

            local newValue = 0
            if length > kLength then
                newValue = floor( kPos * ( self.total - self.visible ) / ( length - kLength ) + 0.5 )
            end

            if self.value ~= newValue then
                self.value = newValue
                self.onValueUpdated( newValue )
            end
        end

        self.kFrame:EnableMouse( true )
        self.kFrame:SetScript( "OnMouseDown",
            function( frame )
                self:getWindow():freeze()

                local width, height = self.frame:GetSize()
                local kWidth, kHeight = self.kFrame:GetSize()
                local kX, kY, _, kH = self.kFrame:GetScaledRect()
                local x, y, _, h = self.frame:GetScaledRect()
                local mouseX, mouseY = GetCursorPosition()
                scale = 1 / self.kFrame:GetEffectiveScale()

                mouseY = 768 - mouseY
                kY = 768 - ( kY + kH )
                y = 768 - ( y + h )

                if self.vertical then
                    kMouseOffset = mouseY - kY
                    startPos = y
                    length = height
                    kLength = kHeight
                else
                    kMouseOffset = mouseX - kX
                    startPos = x
                    length = width
                    kLength = kWidth
                end

                maxPos = length - kLength
                self.isDragging = true
                frame:SetScript( "OnUpdate", kUpdateFunction )

                zu.Raise( self.startDraggingEvent )
            end )

        self.kFrame:SetScript( "OnMouseUp",
            function( frame )
                frame:SetScript( "OnUpdate", nil )
                self.isDragging = false
                self:getWindow():unfreeze()
                self.decorator:release()
                zu.Raise( self.endDraggingEvent )
            end )
    end

    if shouldBeFadedOut( self ) then
        self.fadeOutState = true
        self.kFrame:SetAlpha( 0 )
    else
        self.fadeOutState = false
        self.kFrame:SetAlpha( 1 )
    end
end

function SliderClass:fadeIn()
    self.fade = false
    updateFading( self )
end

function SliderClass:fadeOut()
    self.fade = true
    updateFading( self )
end

function SliderClass:calcMinMax()
    if self.vertical then
        self.minWidth = zu.Scale( 8 )
        self.maxWidth = self.minWidth
        self.minHeight = zu.Scale( 2 )
    else
        self.minHeight = zu.Scale( 8 )
        self.maxHeight = self.minHeight
        self.minWidth = zu.Scale( 2 )
    end

    self.preferredWidth = self.minWidth
    self.preferredHeight = self.minHeight
end

function SliderClass:setLevel( level )
    zu.Widget.setLevel( self, level )
    self.frame:SetFrameLevel( level )
    self.kFrame:SetFrameLevel( level + 1 )
end

function SliderClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )

    local oldW, oldH = self.frame:GetSize()
    self.frame:SetSize( width, height )
    self.frame:SetPoint( "TOPLEFT", self.frame:GetParent(), "TOPLEFT", x, -y )

    local oldKW, oldKH = self.kFrame:GetSize()
    local _, _, _, oldKx, oldKy = self.kFrame:GetPoint( 1 )

    if not oldKx then
        updateKnob( self )
        return
    end

    if self.vertical then
        local kLength = height

        if self.total > self.visible then
            kLength = ( self.visible * height ) / self.total
            kLength = floor( kLength + 0.5 )
            kLength = max( kLength, zu.Scale( self.minKnobSize ) )
        end

        local kPos = 0
        if oldH > oldKH then
            kPos = -oldKy * ( height - kLength ) / ( oldH - oldKH )
            kPos = floor( kPos + 0.5 )
            kPos = min( kPos, height - kLength )
        end

        self.kFrame:SetSize( width, kLength )
        self.kFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 0, -kPos )
    else
        local kLength = width

        if self.total > self.visible then
            kLength = ( self.visible * width ) / self.total
            kLength = floor( kLength + 0.5 )
            kLength = max( kLength, zu.Scale( self.minKnobSize ) )
        end

        local kPos = 0
        if oldW > oldKW then
            kPos = oldKx * ( width - kLength ) / ( oldW - oldKW )
            kPos = floor( kPos + 0.5 )
            kPos = min( kPos, width - kLength )
        end

        self.kFrame:SetSize( kLength, height )
        self.kFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", kPos, 0 )
    end
end

function SliderClass:setTotalAndVisible( total, visible, bigIncrement )
    self.total = total
    self.visible = visible
    self.bigIncrement = bigIncrement or self.bigIncrement
    updateKnob( self )
    updateFading( self )
end

function SliderClass:setValue( value )
    self.value = value
    updateKnob( self )
end
