--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- A B-tree used to efficiently store editable text data
-- and meta data.
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local node_split_threshold = 16
local node_merge_threshold = 16

local min = math.min
local max = math.max
local ceil = math.ceil
local insert = table.insert
local remove = table.remove

local BTree = {}
zu.BTree = BTree

function BTree:new()
    local o = {}

    o.root = { isLeaf = true, size = 0, height = 0, width = 0 }

    setmetatable( o, { __index = self } )
    return o
end

-------------------------------------------------------------------------------
-- Utility
-------------------------------------------------------------------------------
local function findLeftmostLeaf( node )
    while not node.isLeaf do
        node = node[1]
    end
    return node
end

local function findRightmostLeaf( node )
    while not node.isLeaf do
        node = node[#node]
    end
    return node
end

-------------------------------------------------------------------------------
-- Lookup
-------------------------------------------------------------------------------
local function lookupByIndex( node, index )
    local size = node.size
    if index > size then
        return
    end

    if node.isLeaf then
        return node, index
    end

    for i=1,#node,1 do
        local n = node[i]
        size = n.size
        if size >= index then
            return lookupByIndex( n, index )
        end

        index = index - size
    end
end

local function lookupYPositionByIndex( node, index )
    local size = node.size
    if index > size then
        return
    end

    local yp = 0

    if node.isLeaf then
        for i=1,index,1 do
            local e = node[i]
            yp = yp + e.height
        end
        return node, index, yp
    end

    for i=1,#node,1 do
        local n = node[i]
        size = n.size
        if size >= index then
            local fn, fi, fyp = lookupYPositionByIndex( n, index )
            return fn, fi, fyp + yp
        end

        yp = yp + n.height
        index = index - size
    end
end

local function lookupByYPosition( node, position )
    local height = node.height
    if position < 0 or position > height then
        return
    end

    if node.isLeaf then
        for i=1,#node,1 do
            local e = node[i]
            height = e.height
            if height > position then
                return node, i, i, position
            end

            position = position - height
        end
    end

    local index = 0

    for i=1,#node,1 do
        local n = node[i]
        height = n.height
        if height > position then
            local leafIndex, globalIndex
            n, leafIndex, globalIndex, position = lookupByYPosition( n, position )
            return n, leafIndex, globalIndex + index, position
        end

        position = position - height
        index = index + n.size
    end
end

-- Lookup API. Return iterators that can be used with the iteration API.
function BTree:findByIndex( index, dest_it )
    local n, li = lookupByIndex( self.root, index )
    dest_it[1] = n
    dest_it[2] = li
end

function BTree:findByYPosition( position, dest_it )
    local n, li, gi, p = lookupByYPosition( self.root, position )
    dest_it[1] = n
    dest_it[2] = li
    return gi, p
end

function BTree:findYPositionByIndex( index, dest_it )
    local n, li, yp = lookupYPositionByIndex( self.root, index )

    if dest_it then
        dest_it[1] = n
        dest_it[2] = li
    end

    return yp
end

-------------------------------------------------------------------------------
-- Insertion
-------------------------------------------------------------------------------
local function getOrCreateParentNode( bt, node )
    if node.parent then
        return node.parent
    end

    local newRoot = { isLeaf = false, size = node.size, height = node.height, width = node.width }
    insert( newRoot, node )
    node.parent = newRoot
    bt.root = newRoot

    return newRoot
end

local function splitNodeIfNeeded( bt, node )
    local count = #node
    if count <= node_split_threshold then
        return
    end

    local splitSize = ceil( count / ceil( count / node_split_threshold ) )

    local isLeaf = node.isLeaf
    local size = node.size
    local height = node.height

    local parent = getOrCreateParentNode( bt, node )
    local index = splitSize + 1
    local nextLeaf = node.nextLeaf

    -- Look up at which index we are within our parent to find out where
    -- to insert the new nodes
    local insertionIndex
    for i=1,#parent,1 do
        local v = parent[i]
        if v == node then
            insertionIndex = i + 1
        end
    end

    local newNode = { isLeaf = isLeaf, parent = parent }
    if isLeaf then
        newNode.prevLeaf = node
    end

    local newSize = 0
    local newHeight = 0
    local newWidth = 0

    while count > splitSize do
        local n = node[index]
        node[index] = nil
        index = index + 1

        size = size - ( n.size or 1 )
        newSize = newSize + ( n.size or 1 )
        newHeight = newHeight + ( n.height or 0 )
        newWidth = max( newWidth, ( n.width or 0 ) )

        if isLeaf then
            n[1] = newNode
        else
            n.parent = newNode
        end

        insert( newNode, n )

        height = height - ( n.height or 0 )
        count = count - 1

        if #newNode >= splitSize then
            if isLeaf then
                newNode.prevLeaf.nextLeaf = newNode
                newNode.nextLeaf = nextLeaf
                if nextLeaf then
                    nextLeaf.prevLeaf = newNode
                end
            end

            newNode.size = newSize
            newNode.height = newHeight
            newNode.width = newWidth
            insert( parent, insertionIndex, newNode )
            insertionIndex = insertionIndex + 1

            local prev = newNode
            newNode = { isLeaf = isLeaf, parent = parent }
            if isLeaf then
                newNode.prevLeaf = prev
            end

            newSize = 0
            newHeight = 0
            newWidth = 0
        end
    end

    if newSize > 0 then
        if isLeaf then
            newNode.prevLeaf.nextLeaf = newNode
            newNode.nextLeaf = nextLeaf
            if nextLeaf then
                nextLeaf.prevLeaf = newNode
            end
        end

        newNode.size = newSize
        newNode.height = newHeight
        newNode.width = newWidth
        insert( parent, insertionIndex, newNode )
    end

    -- Recompute the max width of the node we've just split
    newWidth = 0
    for i=1,#node,1 do
        local child = node[i]
        newWidth = max( newWidth, ( child.width or 0 ) )
    end

    node.size = size
    node.height = height
    node.width = newWidth

    splitNodeIfNeeded( bt, parent )
end

-- Insertion API
-- Insert a range of entries in the tree at the provided index.
-- The entries are provided by a generator function which should return lua objects to store.
--
--  The object can have the following optional fields: width (width in pixels), height (height in pixels)
--  0 is assumed if they are missing.
function BTree:insert( index, generator )
    local val = generator()
    if val == nil then
        return
    end

    local size = 1
    local width = val.width or 0
    local height = val.height or 0

    local leaf, indexInLeaf = self.root, index

    if index == ( self.root.size + 1 ) then
        -- Special case for inserting right after the last index:
        -- We can't do a lookup for that. Instead,
        -- We find the rightmost children and set the index
        -- to 1 past its end.
        leaf = findRightmostLeaf( self.root )
        indexInLeaf = #leaf + 1
    elseif self.root.size ~= 0 then
        leaf, indexInLeaf = lookupByIndex( self.root, index )
    end

    if leaf == nil then
        return
    end

    val[1] = leaf
    insert( leaf, indexInLeaf, val )

    while( true ) do
        local v = generator()
        if v == nil then
            break
        end

        indexInLeaf = indexInLeaf + 1

        size = size + 1
        height = height + ( v.height or 0 )
        width = max( width, v.width or 0 )
        v[1] = leaf
        insert( leaf, indexInLeaf, v )
    end

    leaf.size = leaf.size + size
    leaf.height = leaf.height + height
    leaf.width = max( leaf.width, width )

    local parent = leaf.parent
    while parent do
        parent.size = parent.size + size
        parent.height = parent.height + height
        parent.width = max( leaf.width, parent.width )
        parent = parent.parent
    end

    splitNodeIfNeeded( self, leaf )
end

-------------------------------------------------------------------------------
-- Deletion
-------------------------------------------------------------------------------
local function collapseNodeIfNeeded( node )
    if node.size >= node_merge_threshold then
        return false
    end

    local curLeaf = node[1]
    local endLeaf = node[#node].nextLeaf

    node.prevLeaf = curLeaf.prevLeaf
    node.nextLeaf = endLeaf

    local i = 1

    while curLeaf ~= endLeaf do
        for j=1,#curLeaf,1 do
            local v = curLeaf[j]
            v[1] = node
            node[i] = v
            i = i + 1
        end
        curLeaf = curLeaf.nextLeaf
    end

    while i < #node do
        node[i] = nil
        i = i + 1
    end

    if node.prevLeaf then
        node.prevLeaf.nextLeaf = node
    end

    if node.nextLeaf then
        node.nextLeaf.prevLeaf = node
    end

    node.isLeaf = true
    return true
end

local function erase( node, index, count )
    local size = node.size
    if index > size then
        return
    end

    local height = node.height
    count = min( count, node.size - index + 1 )

    if node.isLeaf then
        for _=1,count,1 do
            height = height - ( node[index].height or 0 )
            remove( node, index )
        end

        node.size = size - count
        node.height = height

        local width = 0
        for i=1,#node,1 do
            local n = node[i]
            width = max( width, ( n.width or 0 ) )
        end
        node.width = width
        return node, node
    end

    -- Find the first child node that overlaps the deletion range
    local current, s
    local curIndex = 0
    for i=1,#node,1 do
        local n = node[i]
        s = n.size
        if s >= index then
            current = n
            curIndex = i
            break
        end

        index = index - s
    end

    if current == nil then
        return
    end

    -- Keep track of the leaves at the edges of the deletion range
    -- to fix up their chaining
    local rangeLeftmostLeaf
    local rangeRightmostLeaf

    local leftmostLeaf
    local rightmostLeaf

    -- If that node only partially overlaps the range,
    -- recursively delete the overlapping part
    if index > 1 then
        s = current.size
        height = height - current.height

        local lm
        lm, rangeLeftmostLeaf = erase( current, index, count )

        -- If we deleted the end of our first child block, then
        -- its leftmost leaf is also ours and we won't need to look it up later
        if curIndex == 1 then
            leftmostLeaf = lm
        end

        local delta = ( s - current.size )
        count = count - delta
        index = index - delta
        size = size - delta
        height = height + current.height
        curIndex = curIndex + 1
    elseif curIndex > 1 then
        -- If we didn't get rangeLeftmostLeaf from recursing into erase,
        -- we need to pick the rightmost leaf of our last child before the deletion range,
        -- if any
        rangeLeftmostLeaf = findRightmostLeaf( node[curIndex - 1] )
    end

    local atFirstNode = curIndex == 1
    local lastDeletedNode

    -- The deletion range may have been entirely comprised in the first node,
    -- in which case we're done (but we still need to update our width)
    if count > 0 then
        -- Find the last child node that overlaps the deletion range,
        -- completely delete any intervening node
        s = node.size
        for _=curIndex, #node, 1 do
            current = node[curIndex]
            if current.size > count then
                break
            end

            -- If we are completely deleting our first node, we need to retrieve
            -- its leftmost leaf's previous leaf
            if atFirstNode then
                rangeLeftmostLeaf = findLeftmostLeaf( current ).prevLeaf
            end

            lastDeletedNode = current
            height = height - current.height
            size = size - current.size
            count = count - current.size
            remove( node, curIndex )
        end

        -- Check count again because the deletion may have ended
        -- right at the end of one of our child node
        if count > 0 then
            size = size - current.size
            height = height - current.height

            -- Recursively delete the last part of the range
            local rm
            rangeRightmostLeaf, rm = erase( current, 1, count )

            -- If we deleted the end of our last child block, then
            -- its rightmost leaf is also ours and we won't need to look it up later
            if curIndex == #node then
                rightmostLeaf = rm
            end

            size = size + current.size
            height = height + current.height
        elseif not rangeRightmostLeaf and lastDeletedNode then
            -- If we didn't get rangeRightmostLeaf from recursing into erase,
            -- we need to pick the rightmost leafd of the last node we deleted
            rangeRightmostLeaf = findRightmostLeaf( lastDeletedNode ).nextLeaf
        end

        -- Fix up the leaves linked list
        if rangeLeftmostLeaf then
            rangeLeftmostLeaf.nextLeaf = rangeRightmostLeaf
        end

        if rangeRightmostLeaf then
            rangeRightmostLeaf.prevLeaf = rangeLeftmostLeaf
        end
    end

    node.size = size
    node.height = height

    -- Update our width
    local width = 0
    for i=1,#node,1 do
        width = max( width, ( node[i].width or 0 ) )
    end

    node.width = width

    if collapseNodeIfNeeded( node ) then
        -- We became a leaf, return ourself
        return node, node
    end

    -- If we didn't get the opportunity to get our leftmost leaf from
    -- recursing in erase, find it now
    if not leftmostLeaf then
        leftmostLeaf = findLeftmostLeaf( node )
    end

    -- Likewise for our rightmost leaf
    if not rightmostLeaf then
        rightmostLeaf = findRightmostLeaf( node )
    end

    return leftmostLeaf, rightmostLeaf
end

-- Deletion API
function BTree:clear()
    self.root = { isLeaf = true, size = 0, height = 0, width = 0 }
end

function BTree:erase( index, count )
    -- Special case for the deletion of the entire tree
    if index == 1 and count >= self.root.size then
        self:clear()
    else
        erase( self.root, index, count )
    end
end

-------------------------------------------------------------------------------
-- Update
-------------------------------------------------------------------------------
function BTree:updateSize( item, width, height )
    local node = item[1]
    local oldWidth = item.width
    local heightDelta = height - item.height

    item.width = width
    item.height = height

    while node do
        node.height = node.height + heightDelta

        if width > node.width then
            node.width = width
        elseif oldWidth == node.width and width ~= node.width then
            local w = 0
            for i=1,#node,1 do
                w = max( w, node[i].width )
            end
            node.width = w
        end

        node = node.parent
    end
end

-------------------------------------------------------------------------------
-- Iteration
-------------------------------------------------------------------------------
function BTree:itFirst( it )
    if self:size() == 0 then
        it[1] = nil
        return
    end
    it[1] = findLeftmostLeaf( self.root )
    it[2] = 1
end

function BTree:itLast( it )
    if self:size() == 0 then
        it[1] = nil
        return
    end
    it[1] = findRightmostLeaf( self.root )
    it[2] = #it[1]
end

function BTree.itIsEnd( it )
    return it[1] == nil
end

function BTree.itGet( it )
    return it[1][it[2]]
end

function BTree.itNext( it )
    local n = it[1]
    local i = it[2]

    if i == #n then
        it[1] = n.nextLeaf
        it[2] = 1
        return it[1] ~= nil
    end

    it[2] = i + 1
    return true
end

function BTree.itPrev( it )
    local n = it[1]
    local i = it[2]

    if i == 1 then
        it[1] = n.prevLeaf
        if it[1] == nil then
            return false
        end
        it[2] = #it[1]
        return true
    end

    it[2] = i - 1
    return true
end

function BTree.itCopy( it )
    return { it[1], it[2] }
end

-------------------------------------------------------------------------------
-- Misc accessors
-------------------------------------------------------------------------------
function BTree:width()
    return self.root.width
end

function BTree:height()
    return self.root.height
end

function BTree:size()
    return self.root.size
end
