--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor

local remove = table.remove
local sort = table.sort
local concat = table.concat

local itGet = zu.BTree.itGet

local EditorController = {}
zu.EditorController = EditorController

function EditorController:new( model, readOnly )
    local o = {}

    o.scrollUpdateEvent = {}
    o.model = model
    o.cursors = {}
    o.expectingPaste = false
    o.undoBuffer = zu.EditorUndoBuffer:new( model )
    o.clickTimeStamp = 0
    o.haveFocus = false
    o.mouseUpEvent = {}
    o.mouseDownEvent = {}
    o.gainFocusEvent = {}
    o.loseFocusEvent = {}
    o.attachedPopups = {}
    setmetatable( o, { __index = self } )

    o.readOnly = readOnly
    o:createCursor( 1, 1, nil, true )

    return o
end

function EditorController:attachPopup( popup )
    self.attachedPopups[popup] = true
end

function EditorController:detachPopup( popup )
    self.attachedPopups[popup] = nil
end

function EditorController:isMouseOver()
    if self.frame:IsMouseOver() then
        return true
    end

    for k, _ in pairs( self.attachedPopups ) do
        if k:isMouseOver() then
            return true
        end
    end

    return false
end

function EditorController:focus()
    if self.frozen then
        return
    end

    self.view:setFocus( true )
    self.frame:EnableKeyboard( true )
    self.helperEditBox:SetFocus()

    if not self.haveFocus then
        self.haveFocus = true
        zu.Raise( self.gainFocusEvent )
    end
end

function EditorController:unfocus()
    self.view:setFocus( false )
    self.frame:EnableKeyboard( false )
    self.helperEditBox:ClearFocus()

    if self.haveFocus then
        self.haveFocus = false
        zu.Raise( self.loseFocusEvent )
    end
end

local function updateFocus( self )
    if not self.frozen and self.haveFocus then
        self:focus()
    else
        self:unfocus()
    end
end

function EditorController:freeze()
    self.frozen = true
    updateFocus( self )
end

function EditorController:unfreeze()
    self.frozen = false
    updateFocus( self )
end

function EditorController:setView( view )
    self.view = view
    self.frame = view.frame
    self.frame:EnableMouse( true )
    self.frame:EnableKeyboard( false )
    self.frame:EnableMouseWheel( true )

    self.frame:RegisterEvent( "GLOBAL_MOUSE_DOWN" )
    self.frame:RegisterEvent( "GLOBAL_MOUSE_UP" )

    self.frame:SetScript( "OnEvent", function( _, e )
        if e == "GLOBAL_MOUSE_DOWN" then
            if self:isMouseOver() then
                self:focus()
            else
                self:unfocus()
            end
        elseif self:isMouseOver() then
            zu.Raise( self.mouseUpEvent )
        end
    end )

    self.frame:SetScript( "OnMouseWheel", function( _, direction ) self:onMouseWheel( direction ) end )

    self.frame:SetScript( "OnMouseDown", function() self:onMouseDown() end )
    self.frame:SetScript( "OnMouseUp", function() self:onMouseUp() end )

    local function onKeyUp()
        -- Cancel the key repetition timer/ticker, if any.
        if self.repeater then
            self.repeater:Cancel()
            self.repeater = nil
            self.repeatingFunc = nil
        end
    end
    self.frame:SetScript( "OnKeyUp", onKeyUp )

    self.frame:SetScript( "OnKeyDown", function( f, key )
        if self[key] then
            f:SetPropagateKeyboardInput( false )
            self[key]( self )
        else
            f:SetPropagateKeyboardInput( true )
        end
    end )

    -- We still need a native editBox for input handling (since key events only gives us key presses
    -- and we want characters), and as a way to communicate with the clipboard (since user addon
    -- code can't directly read/write it).
    self.helperEditBox = CreateFrame( "EditBox", nil, self.frame )
    self.helperEditBox:SetMultiLine( true )
    self.helperEditBox:SetFrameLevel( self.frame:GetFrameLevel() - 1 )
    self.helperEditBox:SetScript( "OnChar", function( _, c ) self:onChar( c ) end )
    self.helperEditBox:SetScript( "OnTextChanged",
        function( f )
            f:HighlightText()
            self:onEditBoxTextChanged()
        end )

    self.helperEditBox:SetScript( "OnKeyUp", onKeyUp )
    self.helperEditBox:SetAutoFocus( false )
end

local function startKeyRepetition( self, func )
    if self.repeatingFunc == func then
        return
    end

    self.repeatingFunc = func

    if self.repeater then
        self.repeater:Cancel()
    end

    local function repeatKey()
        self.repeatingFunc( self )
    end

    local function initialRepeatKey()
        self.repeatingFunc( self )
        self.repeater = C_Timer.NewTicker( 0.025, repeatKey )
    end

    self.repeater = C_Timer.NewTimer( 0.5, initialRepeatKey )
end

-- TODO we don't really have a proper "primary cursor", it's currently
-- whichever cursor is first in the list (which can change everytime we sort them
-- to deal with cursor movement)
function EditorController:getPrimaryCursorPosition()
    return self.cursors[1][1], self.cursors[1][2]
end

function EditorController:createCursor( l, c, line, silent )
    return zu.EditorCursor:new( self.model, self, l, c, line, silent )
end

function EditorController:restoreCursor( startLine, startCol, endLine, endCol )
    if not endLine then
        self:createCursor( startLine, startCol )
    else
        local cursor = self:createCursor( endLine, endCol )
        cursor:moveTo( startLine, startCol, true )
    end
end

function EditorController:removeCursor( cursor )
    local cursors = self.cursors
    for i=1,#cursors,1 do
        if cursors[i] == cursor then
            remove( cursors, i )
            return
        end
    end
end

local function clearCursors( self )
    local cursors = self.cursors

    for _=1,#cursors,1 do
        cursors[1]:destroy()
    end
end

function EditorController:setPrimaryCursorPosition( fl, fc )
    self.cursors[1]:moveTo( fl, fc )
end

function EditorController:setPrimaryCursorExtent( fl, fc, ll, lc )
    self.cursors[1]:moveTo( fl, fc )
    self.cursors[1]:moveTo( ll, lc, true )
end

function EditorController:setText( text )
    self.model:replaceTextBuffer( text )
    wipe( self.cursors )
    self:createCursor( 1, 1, nil, true )
    self:scrollToCursors()
end

function EditorController:onMouseDown()
    self:focus()

    local model = self.model
    local x, y = self.view:getModelRelativeMousePosition()
    local l, c, line = model:coordsToLineCol( x, y )

    if l == nil then
        return
    end

    local time = GetTime()

    if l ~= self.clickLine or c ~= self.clickCol then
        self.clickTimeStamp = 0
    end

    if ( time - self.clickTimeStamp ) < 0.333 then
        self.clickTimeStamp = 0
        self:onDoubleClick()
        return
    end

    self.clickTimeStamp = time
    self.clickLine = l
    self.clickCol = c

    local ctrlDown = IsLeftControlKeyDown() or IsRightControlKeyDown()
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()
    local cursors = self.cursors

    if not ctrlDown then
        -- Destroy any other currently existing cursor,
        -- move the first one. Keep/extend selection if shift is held.
        for i=1,#cursors,1 do
            if i > 1 then
                cursors[2]:destroy( 2 )
            end
        end

        cursors[1]:moveTo( l, c, shiftDown )
        self.currentDraggingCursor = cursors[1]
    else
        -- If there is already a cursor there,
        -- remove it instead, unless it's the only cursor.
        local cursor = zu.EditorCursor.FindInLine( line, c )
        if cursor then
            if #cursors > 1 then
                cursor:destroy()
                self.model:queueLineUpdate( line )
                model:flushPendingEvents()
                self.view:resetBlinking()
            end

            return
        else
            self.currentDraggingCursor = self:createCursor( l, c, line )
        end
    end

    self.view.onUpdate = function() self:onUpdate() end
    model:flushPendingEvents()
    self.view:resetBlinking()
    zu.Raise( self.mouseDownEvent )
end

function EditorController:onMouseUp()
    self.currentDraggingCursor = nil
    self.view.onUpdate = nil
end

function EditorController:onUpdate()
    if not self.currentDraggingCursor then
        return
    end

    local model = self.model
    local x, y = self.view:getModelRelativeMousePosition()

    x = min( x, model.btree:width() )
    y = min( y, model.btree:height() - 1 )

    local l, c = model:coordsToLineCol( x, y )
    self.currentDraggingCursor:moveTo( l, c, true )
    model:flushPendingEvents()
    self.view:resetBlinking()
end

function EditorController:onDoubleClick()
    if self.doubleClickSelectsAll then
        self:selectAll()
        return
    end

    local model = self.model
    local x, y = self.view:getModelRelativeMousePosition()
    local l, c, line = model:coordsToLineCol( x, y )

    if l == nil then
        return
    end

    local cursor = self:createCursor( l, c, line )
    cursor:moveToPreviousWordFrontier( false, true )
    cursor:moveToNextWordFrontier( true, true )

    model:flushPendingEvents()
    self.view:resetBlinking()
end

function EditorController:scrollToCursors()
    local cursors = self.cursors
    local view = self.view
    local model = view.model

    local minLine = cursors[1][1] - 1
    local minCol = cursors[1][2] - 1
    local maxLine, maxCol = minLine, minCol

    for i=2,#cursors,1 do
        local c = cursors[i]
        minLine = min( minLine, c[1] - 1 )
        maxLine = max( maxLine, c[1] - 1 )
        minCol = min( minCol, c[2] - 1 )
        maxCol = max( maxCol, c[2] - 1 )
    end

    local charWidth = model.charWidth
    local lineHeight = model.lineHeight

    local cwidth = ( maxCol - minCol ) * charWidth
    local cheight = ( maxLine - minLine ) * lineHeight

    if cwidth > view.width or cheight > view.height then
        -- The set of cursors don't fit in the view, don't bother scrolling at all
        model:flushPendingEvents()
        self.view:resetBlinking()
        return
    end

    local cx = minCol * charWidth
    local cy = minLine * lineHeight
    local cmaxx = cwidth + cx + charWidth * 2
    local cmaxy = cheight + cy + lineHeight * 2

    local vx = view.scrollXPosition
    local vmaxx = view.scrollXPosition + view.width
    local vy = view.scrollYPosition
    local vmaxy = view.scrollYPosition + view.height

    local scrollChanged = false

    if cmaxx > vmaxx then
        -- Cursors are right or partly to the right of the view, scroll right
        -- to bring them into view
        view:setScrollXPosition( cmaxx - view.width )
        scrollChanged = true
    elseif cx < vx then
        -- Cursors are left or partly to the left of the view, scroll lzft
        -- to bring them into view
        view:setScrollXPosition( cx )
        scrollChanged = true
    end

    if cmaxy > vmaxy then
        -- Cursors are below or partly below the view, scroll down
        -- to bring them into view
        view:setScrollYPosition( cmaxy - view.height )
        scrollChanged = true
    elseif cy < vy then
        -- Cursors are above or partly above the view, scroll up
        -- to bring them into view
        view:setScrollYPosition( cy )
        scrollChanged = true
    end

    if scrollChanged then
        zu.Raise( self.scrollUpdateEvent )
    end

    model:flushPendingEvents()
    self.view:resetBlinking()
end

local it = {}

function EditorController:selectAll()
    clearCursors( self )

    local model = self.model
    model.btree:itLast( it )
    local line = itGet( it )

    self:restoreCursor( model.btree:size(), strlenutf8( line[2] ) + 1, 1, 1 )
    self.model:flushPendingEvents()
    self.view:resetBlinking()
end

-- Mouse wheel handler
function EditorController:onMouseWheel( direction )
    local vy = self.view.scrollYPosition

    -- The scrolling amount is hardcoded for now.
    vy = vy - direction * 2 * self.model.lineHeight * self.view.lineHeightScale

    vy = max( vy, 0 )
    vy = min( vy, self.view.scrollableHeight )

    if vy ~= self.view.scrollYPosition then
        self.view:setScrollYPosition( vy )
        zu.Raise( self.scrollUpdateEvent )
    end
end

-- Key press handlers
function EditorController:LEFT()
    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()
    local ctrlDown = IsLeftControlKeyDown() or IsRightControlKeyDown()

    -- Sort the cursors first. We need to update them from first to last,
    -- but we need to sort them in the reverse order since previously
    -- updated cursors may get deleted, thus invalidating the current
    -- iteration index.
    sort( cursors, zu.EditorCursor.LastToFirstOrdering )
    for i=#cursors,1,-1 do
        local cursor = cursors[i]
        if not shiftDown and cursor[3] then
            cursor:collapseSelection( true )
        elseif ctrlDown then
            cursor:moveToPreviousWordFrontier( shiftDown )
        else
            cursor:moveLeft( shiftDown )
        end
    end

    self:scrollToCursors()
    startKeyRepetition( self, self.LEFT )
end

function EditorController:RIGHT()
    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()
    local ctrlDown = IsLeftControlKeyDown() or IsRightControlKeyDown()

    -- Sort the cursors first. We need to update them from last to first,
    -- but we need to sort them in the reverse order since previously
    -- updated cursors may get deleted, thus invalidating the current
    -- iteration index.
    sort( cursors, zu.EditorCursor.FirstToLastOrdering )

    for i=#cursors,1,-1 do
        local cursor = cursors[i]
        if not shiftDown and cursor[3] then
            cursor:collapseSelection( false )
        elseif ctrlDown then
            cursor:moveToNextWordFrontier( shiftDown )
        else
            cursor:moveRight( shiftDown )
        end
    end

    self:scrollToCursors()
    startKeyRepetition( self, self.RIGHT )
end

function EditorController:UP()
    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()

    -- Sort the cursors first. We need to update them from first to last,
    -- but we need to sort them in the reverse order since previously
    -- updated cursors may get deleted, thus invalidating the current
    -- iteration index.
    sort( cursors, zu.EditorCursor.LastToFirstOrdering )

    for i=#cursors,1,-1 do
        cursors[i]:moveUp( shiftDown )
    end

    self:scrollToCursors()
    startKeyRepetition( self, self.UP )
end

function EditorController:DOWN()
    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()

    -- Sort the cursors first. We need to update them from last to first,
    -- but we need to sort them in the reverse order since previously
    -- updated cursors may get deleted, thus invalidating the current
    -- iteration index.
    sort( cursors, zu.EditorCursor.FirstToLastOrdering )

    for i=#cursors,1,-1 do
        cursors[i]:moveDown( shiftDown )
    end

    self:scrollToCursors()
    startKeyRepetition( self, self.DOWN )
end

-- Misc navigation and selection keys
function EditorController:PAGEUP()
    local view = self.view
    local model = view.model

    local numLines = floor( view.height / model.lineHeight ) - 2

    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()

    sort( cursors, zu.EditorCursor.LastToFirstOrdering )

    for i=#cursors,1,-1 do
        local cursor = cursors[i]
        local destLine = max( cursor[1] - numLines, 1 )
        cursors[i]:moveTo( destLine, cursor[2], shiftDown )
    end

    self:scrollToCursors()
    startKeyRepetition( self, self.PAGEUP )
end

function EditorController:PAGEDOWN()
    local view = self.view
    local model = view.model

    local numLines = floor( view.height / model.lineHeight ) - 2

    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()

    sort( cursors, zu.EditorCursor.LastToFirstOrdering )

    for i=#cursors,1,-1 do
        local cursor = cursors[i]
        local destLine = min( cursor[1] + numLines, model.btree:size() )
        cursors[i]:moveTo( destLine, cursor[2], shiftDown )
    end

    self:scrollToCursors()
    startKeyRepetition( self, self.PAGEDOWN )
end

function EditorController:HOME()
    local cursors = self.cursors
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()

    sort( cursors, zu.EditorCursor.LastToFirstOrdering )
    for i=#cursors,1,-1 do
        local cursor = cursors[i]
        cursor:moveTo( cursor[1], 1, shiftDown )
    end

    self:scrollToCursors()
end

function EditorController:END()
    local cursors = self.cursors
    local model = self.model
    local shiftDown = IsLeftShiftKeyDown() or IsRightShiftKeyDown()

    sort( cursors, zu.EditorCursor.LastToFirstOrdering )
    for i=#cursors,1,-1 do
        local cursor = cursors[i]

        model.btree:findByIndex( cursor[1], it )
        local line = itGet( it )

        cursor:moveTo( cursor[1], strlenutf8( line[2] ) + 1, shiftDown )
    end

    self:scrollToCursors()
end

function EditorController:A()
    if IsLeftControlKeyDown() or IsRightControlKeyDown() then
        self:selectAll()
        return
    end

    self.frame:SetPropagateKeyboardInput( true )
end

local singleLineTextArray = {}
local removedText = {}

function EditorController:replaceTextAtCursors( textToInsert, moveToEnd, oneLinePerCursor, preStep, wholeLineIfNoSelection, reason, indenter )
    if self.readOnly then
        return
    end

    wipe( removedText )

    local cursors = self.cursors
    sort( cursors, zu.EditorCursor.FirstToLastOrdering )

    self.undoBuffer:snapshotCursors( cursors )

    local text = oneLinePerCursor and singleLineTextArray or textToInsert
    local needFinalNewLine = false

    -- If we are in wholeLineIfNoSelection mode, we need to remove any additional
    -- cursors on lines that have a cursor without selections.
    if wholeLineIfNoSelection then
        local lineToDeleteCursors = 0
        local curIndex = 1
        while curIndex <= #cursors do
            local cursor = cursors[curIndex]
            local curStartLine = min( cursor[1], cursor[3] or cursor[1] )

            if lineToDeleteCursors == curStartLine then
                cursor:destroy()
            else
                curIndex = curIndex + 1

                if not cursor[3] then
                    lineToDeleteCursors = curStartLine
                end
            end
        end
    end

    local undoStep = {}

    for i=1,#cursors,1 do
        local cursor = cursors[i]

        if preStep and not preStep( cursor ) then
            return
        end

        if oneLinePerCursor then
            singleLineTextArray[1] = textToInsert[i]
        end

        -- If we are in "replace the entire line if there is no selection" mode
        -- and the cursor has no selection, remember that we need an explicit final newline at the end.
        needFinalNewLine = wholeLineIfNoSelection and cursor[3] == nil

        local remText, lcc, ccc, l, c = cursor:replaceText( text, moveToEnd, wholeLineIfNoSelection, reason, indenter )
        local lastLine = cursor[1] + #text - 1

        local undoOperation = { l, c, #text, strlenutf8( text[#text] ), remText }
        undoStep[#undoStep+1] = undoOperation

        -- Update the the position of all following cursors
        for j=i+1,#cursors,1 do
            local otherCur = cursors[j]

            otherCur[1] = otherCur[1] + lcc
            if otherCur[3] then
                otherCur[3] = otherCur[3] + lcc
            end

            if otherCur[1] == lastLine then
                otherCur[2] = otherCur[2] + ccc
            end

            if otherCur[3] and otherCur[3] == lastLine then
                otherCur[4] = otherCur[4] + ccc
            end
        end

        for j=1,#remText,1 do
            -- If we took the entire line, remove the ending newline. We'll put it after
            -- the last consecutive whole line that we have cut (this is to be consistent with ctrl+c
            -- behavior)
            if needFinalNewLine and j == #remText then
                break
            end

            removedText[#removedText + 1] = remText[j]
        end
    end

    if needFinalNewLine then
        removedText[#removedText + 1] = ""
    end

    -- Merge cursors if needed
    for i=#cursors,1,-1 do
        local cursor = self.cursors[i]
        cursor:moveTo( cursor[1], cursor[2] )
    end

    self.undoBuffer:pushUndoStep( undoStep )
    self.undoBuffer:snapshotCursors( cursors )

    self:scrollToCursors()
    return removedText
end

local emptyTextArray = { "" }

local function extendSelTo1CharPrestep( cursor )
    if not cursor[3] then
        cursor[3] = cursor[1]
        cursor[4] = cursor[2] + 1
    end
    return true
end

local function moveLeftPrestep( cursor )
    if cursor[3] then
        return true
    end

    if cursor[1] == 1 and cursor[2] == 1 then
        return false
    end

    cursor:moveLeft( false )
    return extendSelTo1CharPrestep( cursor )
end

function EditorController:DELETE()
    if self.readOnly then
        return
    end

    self:replaceTextAtCursors( emptyTextArray, false, false, extendSelTo1CharPrestep, false, "delete" )
    startKeyRepetition( self, self.DELETE )
end

function EditorController:BACKSPACE()
    if self.readOnly then
        return
    end

    self:replaceTextAtCursors( emptyTextArray, false, false, moveLeftPrestep, false, "backspace" )
    startKeyRepetition( self, self.BACKSPACE )
end

local charTextArray = {}
function EditorController:onChar( c )
    -- Check if ctrl is down but not alt at the same time:
    -- it turns out that the "alt gr" key on some keyboards (like french ones) is not
    -- just "the right alt key but it has a funny name for some reason" like I believed
    -- my entire life but actually a key that does ctrl+alt (what the fuck)
    -- So we have to only consider the case where ctrl is down but not alt here,
    -- otherwise alt gr character combinations don't work.
    local isCtrlDown = not IsLeftAltKeyDown() and not IsRightAltKeyDown()
        and ( IsLeftControlKeyDown() or IsRightControlKeyDown() )

    if self.expectingPaste or isCtrlDown or self.readOnly then
        return
    end

    wipe( charTextArray )

    if c == "\n" then
        charTextArray[1] = ""
        charTextArray[2] = ""
        self:replaceTextAtCursors( charTextArray, true, false, nil, false, "enter", self.indenter )
    else
        charTextArray[1] = c
        self:replaceTextAtCursors( charTextArray, true, false, nil, false, "character", self.indenter )
    end

    self.model:flushPendingEvents()
end

-- Cut: just use replaceTextAtCursors to replace the selections with nothing
-- and return the removed text, since it already does that.
function EditorController:X()
    if self.readOnly then
        return
    end

    if IsLeftControlKeyDown() or IsRightControlKeyDown() then
        local cutText = self:replaceTextAtCursors( emptyTextArray, false, false, moveLeftPrestep, true, "cut" )
        if #cutText > 0 then
            self.helperEditBox:SetText( concat( cutText, "\n" ) )
            self.helperEditBox:HighlightText()
            wipe( cutText )
        end
    end

    self.frame:SetPropagateKeyboardInput( true )
end

-- Copy: call extractText on every cursor.
local copyTextArray = {}
function EditorController:C()
    if IsLeftControlKeyDown() or IsRightControlKeyDown() then
        local model = self.model
        local cursors = self.cursors
        sort( cursors, zu.EditorCursor.FirstToLastOrdering )

        local needFinalNewLine = false

        for i=1,#cursors,1 do
            local cursor = cursors[i]
            local firstLine, firstCol = cursor[1], cursor[2]
            local lastLine, lastCol = cursor[3] or firstLine, cursor[4] or firstCol

            if firstLine > lastLine then
                firstLine, lastLine = lastLine, firstLine
                firstCol, lastCol = lastCol, firstCol
            elseif firstLine == lastLine and firstCol > lastCol then
                firstCol, lastCol = lastCol, firstCol
            end

            -- If we grabbed the entire line because the cursor had no selection,
            -- we'll need to add a newline at the end. Except if it's followed by something else,
            -- in which case the newline will be implicit.
            needFinalNewLine = firstLine == lastLine and firstCol == lastCol
            model:extractText( copyTextArray, firstLine, firstCol, lastLine, lastCol )
        end

        if needFinalNewLine then
            copyTextArray[#copyTextArray+1] = ""
        end

        if #copyTextArray > 0 then
            self.helperEditBox:SetText( concat( copyTextArray, "\n" ) )
            self.helperEditBox:HighlightText()
            wipe( copyTextArray )
        end
    end

    self.frame:SetPropagateKeyboardInput( true )
end

-- Paste
function EditorController:V()
    if self.readOnly then
        return
    end

    if IsLeftControlKeyDown() or IsRightControlKeyDown() then
        self.expectingPaste = true
        self.helperEditBox:HighlightText()
    end

    self.frame:SetPropagateKeyboardInput( true )
end

local pasteTextArray = {}
function EditorController:pasteText( text, reason )
    -- Turn the incoming text into an array of strings.
    local lineGen = zu.LineGenerator( text )

    while(true) do
        local line = lineGen()
        if line == nil then
            break
        end
        pasteTextArray[#pasteTextArray+1] = line
    end

    -- If the incoming text has as many lines as we have cursors, paste one line at each cursor.
    -- Otherwise, paste the entire thing at each cursor. (this matches vscode's behavior)
    self:replaceTextAtCursors( pasteTextArray, true,
        #pasteTextArray == #self.cursors,
        nil, false, reason or "paste", self.indenter )

    wipe( pasteTextArray )
end

function EditorController:paste()
    -- Turn the incoming pasted text into an array of strings.
    self:pasteText( self.helperEditBox:GetText(), "clipboardPaste" )
    startKeyRepetition( self, self.paste )
end

function EditorController:onEditBoxTextChanged()
    if not self.expectingPaste then
        return
    end

    self.expectingPaste = false
    self:paste()
end

-- Undo/Redo
function EditorController:Z()
    if self.readOnly then
        return
    end

    if IsLeftControlKeyDown() or IsRightControlKeyDown() then
        local undoBuffer = self.undoBuffer
        if not undoBuffer:canUndo() then
            return
        end

        clearCursors( self )
        local newCursors = undoBuffer:undo()

        -- Restore cursors
        for i=1,#newCursors,1 do
            local newCur = newCursors[i]
            self:restoreCursor( newCur[1], newCur[2], newCur[3], newCur[4] )
        end

        self:scrollToCursors()
        startKeyRepetition( self, self.Z )
    else
        self.frame:SetPropagateKeyboardInput( true )
    end
end

function EditorController:Y()
    if self.readOnly then
        return
    end

    if IsLeftControlKeyDown() or IsRightControlKeyDown() then
        local undoBuffer = self.undoBuffer
        if not undoBuffer:canRedo() then
            return
        end

        clearCursors( self )
        local newCursors = undoBuffer:redo()

        -- Restore cursors
        for i=1,#newCursors,1 do
            local newCur = newCursors[i]
            self:restoreCursor( newCur[1], newCur[2], newCur[3], newCur[4] )
        end

        self:scrollToCursors()
        startKeyRepetition( self, self.Y )
    else
        self.frame:SetPropagateKeyboardInput( true )
    end
end

-- Indention
local tabTextArray = {}

function EditorController:TAB()
    if self.readOnly then
        return
    end

    -- If one or more cursor have a selection, increase or decrease indention on every line.
    -- Otherwise, insert the right number of spaces to reach the next indention level at each
    -- cursor.
    -- If shift is pressed, we are always in indention mode.
    local indention = self.indention
    local indentMode = false
    local indentRight = true

    if IsLeftShiftKeyDown() or IsRightShiftKeyDown() then
        indentMode = true
        indentRight = false
    end

    local cursors = self.cursors
    if not indentMode then
        for i=1,#cursors,1 do
            if cursors[i][3] ~= nil then
                indentMode = true
                break
            end
        end
    end

    if indentMode then
        sort( cursors, zu.EditorCursor.LastToFirstOrdering )

        self.undoBuffer:snapshotCursors( cursors )
        local undoStep = {}

        for i=#cursors,1,-1 do
            local cursor = cursors[i]
            cursor:indentLines( indention, indentRight, undoStep )
        end

        self.model:queueBTreeUpdate()
        self.model:flushPendingEvents()
        self.view:resetBlinking()

        self.undoBuffer:pushUndoStep( undoStep )
        self.undoBuffer:snapshotCursors( cursors )
    else
        for i=1,#cursors,1 do
            local cursor = cursors[i]

            local col = cursor[2]
            if cursor[3] and cursor[1] > cursor[3] then
                col = cursor[4]
            end

            local numSpacesToAdd = indention - ( ( col - 1 ) % indention )
            local indent = ""
            for _=1,numSpacesToAdd,1 do
                indent = indent .. " "
            end

            tabTextArray[#tabTextArray+1] = indent
        end

        self:replaceTextAtCursors( tabTextArray, true, true, nil, false, "tab" )
        wipe( tabTextArray )
    end

    startKeyRepetition( self, self.TAB )
end
