--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local remove = table.remove
local min = math.min
local max = math.max
local abs = math.abs

local itIsEnd = zu.BTree.itIsEnd
local itNext = zu.BTree.itNext
local itPrev = zu.BTree.itPrev
local itGet = zu.BTree.itGet

local EditorCursor = {}
zu.EditorCursor = EditorCursor

local it = {}
local it2 = {}

function EditorCursor:new( model, controller, l, c, line, silent )
    local o = { l, c, model = model, controller = controller }

    if line == nil then
        model.btree:findByIndex( l, it )

        if itIsEnd( it ) then
            return
        end

        line = itGet( it )
    end

    local curInfos = zu.CreateTable( "cursorInfos" )
    curInfos[1] = c
    curInfos[2] = nil
    curInfos[3] = nil
    line[3][o] = curInfos

    if not silent then
        model:queueLineUpdate( line )
    end

    setmetatable( o, { __index = self } )

    controller.cursors[#controller.cursors+1] = o
    return o
end

function EditorCursor:destroy( index )
    local currentLine = self[1]
    local updateRangeEnd = self[3] or currentLine

    if currentLine > updateRangeEnd then
        currentLine, updateRangeEnd = updateRangeEnd, currentLine
    end

    self.model.btree:findByIndex( currentLine, it2 )

    while currentLine <= updateRangeEnd and not itIsEnd( it2 ) do
        local line = itGet( it2 )

        zu.ReleaseTable( "cursorInfos", line[3][self] )
        line[3][self] = nil

        self.model:queueLineUpdate( line )
        currentLine = currentLine + 1
        itNext( it2 )
    end

    local cursors = self.controller.cursors
    if index then
        remove( cursors, index )
        return
    end

    for i=1,#cursors,1 do
        if cursors[i] == self then
            remove( cursors, i )
            return
        end
    end
end

function EditorCursor.FindInLine( line, col )
    for cur,curInfos in pairs( line[3] ) do
        if curInfos[1] == col then
            return cur
        end

        if curInfos[2] and col >= curInfos[2]
            and col <= ( curInfos[3] or strlenutf8( line[2] ) + 2 ) then
            return cur
        end
    end
end

-- Merge the other selection to self.
-- We assume that the start of self lies inside of other.
-- Other's start becomes our start (the cursor's position),
-- and we keep our end. We then unbind other from all the line it is bound
-- to, and bind them to self instead.
-- This is meant to be called by MoveTo, which will take care of fixing
-- up the cursorInfos on the line where the merging occurs.
local function merge( self, other )
    self[1] = other[1]
    self[2] = other[2]

    local currentLine = other[1]
    local updateRangeEnd = other[3] or currentLine

    if currentLine > updateRangeEnd then
        currentLine, updateRangeEnd = updateRangeEnd, currentLine
    end

    self.model.btree:findByIndex( currentLine, it2 )

    while currentLine <= updateRangeEnd and not itIsEnd( it2 ) do
        local line = itGet( it2 )

        -- If the line is already bound to self, just delete
        -- the cursorInfos of other. Otherwise, transfer the cursorInfos.
        if line[3][self] then
            zu.ReleaseTable( "cursorInfos", line[2][other] )
        else
            line[3][self] = line[3][other]
        end

        line[3][other] = nil

        self.model:queueLineUpdate( line )
        currentLine = currentLine + 1
        itNext( it2 )
    end

    local cursors = self.controller.cursors
    for i=1,#cursors,1 do
        if cursors[i] == other then
            remove( cursors, i )
            break
        end
    end

    return max( self[1], self[3] or self[1] )
end

-- Move the cursor to a new position. If keepSelection is true, the selection will be
-- extended/shrinked to the new position, otherwise it will be cleared.
function EditorCursor:moveTo( l, c, keepSelection )
    local startLine = self[1]
    local endLine = self[3] or startLine
    local endCol = self[4] or self[2]

    local newStartLine = l
    local newEndLine = endLine
    local newEndCol = endCol

    if not keepSelection then
        newEndLine = l
        newEndCol = c
    end

    self[1] = l
    self[2] = c
    if newEndLine ~= l or newEndCol ~= c then
        self[3] = newEndLine
        self[4] = newEndCol
    else
        self[3] = nil
        self[4] = nil
    end

    -- Normalize the start and end positions to be in top->bottom, left->right order
    if startLine > endLine then
        startLine, endLine = endLine, startLine
    end

    local forward = true
    if newStartLine > newEndLine then
        newStartLine, newEndLine = newEndLine, newStartLine
        forward = false
    elseif newStartLine == newEndLine and c > newEndCol then
        forward = false
    end

    -- We can have four cases here:
    --  * newStartLine matches startLine (moving the end of the selection),
    --    which means we need to update from the min of (newEndLine, endLine)
    --    to the max of same
    --  * newEndLine matches endLine (moving the start of the selection),
    --    which means we need to update from the min of (newStartLine, StartLine)
    --    to the max of same
    --  * both the old and new start/end match (just moving the cursor horizontally),
    --    which means we need to update only the line where the cursor is
    --  * neither the starts or the ends match (moving the selection from
    --    one side of the cursor to the other), which means we need to update
    --    from the min of the start lines to the max of the end lines
    local currentLine, updateRangeEnd

    if startLine == newStartLine then
        if endLine == newEndLine then
            currentLine = l
            updateRangeEnd = l
        else
            currentLine = min( endLine, newEndLine )
            updateRangeEnd = max( endLine, newEndLine )
        end
    elseif endLine == newEndLine then
        currentLine = min( startLine, newStartLine )
        updateRangeEnd = max( startLine, newStartLine )
    else
        currentLine = min( startLine, newStartLine )
        updateRangeEnd = max( endLine, newEndLine )
    end

    self.model.btree:findByIndex( currentLine, it )

    while currentLine <= updateRangeEnd and not itIsEnd( it ) do
        local line = itGet( it )
        local lineLen = strlenutf8( line[2] ) + 1

        if currentLine >= newStartLine and currentLine <= newEndLine then
            if currentLine == self[1] then
                self[2] = min( self[2], lineLen )
            end

            local isStartLine = currentLine == self[1]
            local isEndLine = currentLine == self[3]

            local minSpannedCol, maxSpannedCol

            if forward then
                minSpannedCol = isStartLine and self[2] or 1
                maxSpannedCol = self[4] and ( isEndLine and self[4] or lineLen ) or minSpannedCol
            else
                maxSpannedCol = isStartLine and self[2] or lineLen
                minSpannedCol = self[4] and ( isEndLine and self[4] or 1 ) or maxSpannedCol
            end

            minSpannedCol = min( minSpannedCol, lineLen )
            maxSpannedCol = min( maxSpannedCol, lineLen )

            -- Check for any overlapping cursor/selection.
            -- If another selection begins within the current selection,
            -- destroy it.
            -- If our start position is within another selection, merge them.
            for k,v in pairs( line[3] ) do
                if k ~= self then
                    local lastSelCol = v[3] or lineLen + 1

                    -- If we just swallowed another selection's cursor, delete that selection.
                    if v[1] and v[1] >= minSpannedCol and v[1] <= maxSpannedCol then
                        k:destroy()
                    elseif v[2] and minSpannedCol <= lastSelCol and maxSpannedCol >= v[2] then
                        -- Check if we swallowed another selection's cursor for the case where
                        -- that cursor is not on the current line. This can happen when a selection
                        -- is extended upward at once accross an existing selection. In this case
                        -- merging would be incorrect, we want to destroy the existing selection instead.
                        if k[1] ~= currentLine and k[1] >= newStartLine and k[1] <= newEndLine then
                            k:destroy()
                        else
                            newEndLine = merge( self, k )

                            isStartLine = currentLine == self[1]
                            isEndLine = currentLine == self[3]

                            -- Update the start/end cols so that we set the correct
                            -- cursorInfos according to our newly merged selection.
                            -- However, don't update the line variables used for the line
                            -- iteration: we don't want to iterate through the lines
                            -- of the selection we merged since merge() already took
                            -- care of it.
                            if forward then
                                minSpannedCol = isStartLine and self[2] or 1
                                maxSpannedCol = self[4] and ( isEndLine and self[4] or lineLen ) or minSpannedCol
                            else
                                maxSpannedCol = isStartLine and self[2] or strlenutf8( line[2] ) + 1
                                minSpannedCol = self[4] and ( isEndLine and self[4] or 1 ) or maxSpannedCol
                            end
                        end
                    end
                end
            end

            if self[3] == nil then
                minSpannedCol = nil
                maxSpannedCol = nil
            end

            local curInfos = zu.CreateTable( "cursorInfos" )
            if currentLine == self[1] then
                curInfos[1] = self[2]
            else
                curInfos[1] = nil
            end

            curInfos[2] = minSpannedCol
            curInfos[3] = currentLine == newEndLine and maxSpannedCol or nil

            line[3][self] = curInfos
        else
            zu.ReleaseTable( "cursorInfos", line[2][self] )
            line[3][self] = nil
        end

        self.model:queueLineUpdate( line )
        currentLine = currentLine + 1
        itNext( it )
    end
end

function EditorCursor:collapseSelection( toLeft )
    if not self[3] then
        return
    end

    local firstLine, firstCol = self[1], self[2]
    local lastLine, lastCol = self[3], self[4]

    if firstLine > lastLine then
        firstLine, lastLine = lastLine, firstLine
        firstCol, lastCol = lastCol, firstCol
    elseif firstCol > lastCol then
        firstCol, lastCol = lastCol, firstCol
    end

    if toLeft then
        self:moveTo( firstLine, firstCol, false )
    else
        self:moveTo( lastLine, lastCol, false )
    end
end

function EditorCursor.FirstToLastOrdering( a, b )
    if a[1] < b[1] then
        return true
    end

    if a[1] > b[1] then
        return false
    end

    return a[2] < b[2]
end

function EditorCursor.LastToFirstOrdering( a, b )
    if a[1] > b[1] then
        return true
    end

    if a[1] < b[1] then
        return false
    end

    return a[2] > b[2]
end

function EditorCursor:moveLeft( keepSelection )
    if self[1] == 1 and self[2] == 1 then
        return
    end

    if self[2] > 1 then
        self:moveTo( self[1], self[2] - 1, keepSelection )
    else
        self.model.btree:findByIndex( self[1], it )
        if itIsEnd( it ) then
            return
        end

        itPrev( it )
        local line = itGet( it )
        self:moveTo( self[1] - 1, strlenutf8( line[2] ) + 1, keepSelection )
    end
end

function EditorCursor:moveRight( keepSelection )
    self.model.btree:findByIndex( self[1], it )
    if itIsEnd( it ) then
        return
    end

    local line = itGet( it )
    local maxCol = strlenutf8( line[2] ) + 1

    if self[1] == self.model.btree:size() and self[2] == maxCol then
        return
    end

    if self[2] < maxCol then
        self:moveTo( self[1], self[2] + 1, keepSelection )
    else
        self:moveTo( self[1] + 1, 1, keepSelection )
    end
end

function EditorCursor:moveUp( keepSelection )
    if self[1] == 1 then
        if not keepSelection then
            self:moveTo( self[1], self[2], false )
        end
        return
    end

    self:moveTo( self[1] - 1, self[2], keepSelection )
end

function EditorCursor:moveDown( keepSelection )
    if self[1] == self.model.btree:size() then
        if not keepSelection then
            self:moveTo( self[1], self[2], false )
        end
        return
    end

    self:moveTo( self[1] + 1, self[2], keepSelection )
end

local spaceAndPunc =
{
    [9] = true,     -- tab
    [32] = true,    -- space
    [34] = true,    -- "
    [39] = true,    -- '
    [40] = true,    -- (
    [41] = true,    -- )
    [44] = true,    -- ,
    [46] = true,    -- .
    [47] = true,    -- /
    [58] = true,    -- :
    [59] = true,    -- ;
    [91] = true,    -- [
    [93] = true,    -- ]
    [123] = true,   -- {
    [125] = true    -- }
}

function EditorCursor:moveToPreviousWordFrontier( keepSelection, sameLineOnly )
    if self[2] == 1 then
        if sameLineOnly then
            return
        end

        if self[1] == 1 then
            return
        end

        self:moveLeft( keepSelection )
    end

    self.model.btree:findByIndex( self[1], it )
    local line = itGet( it )
    local text = line[2]

    local i = zu.UTF8IndexToByteIndex( text, self[2] - 1 )
    local isSpaceOrPunc = spaceAndPunc[text:byte( i )]

    while i >= 1 do
        local c = text:byte( i )
        local isSoP = spaceAndPunc[c]

        if isSoP ~= isSpaceOrPunc then
            break
        end

        i = i - 1
    end

    self:moveTo( self[1], zu.ByteIndexToUTF8Index( text, i ) + 1, keepSelection )
end

function EditorCursor:moveToNextWordFrontier( keepSelection, sameLineOnly )
    self.model.btree:findByIndex( self[1], it )
    local line = itGet( it )
    local text = line[2]

    if self[2] == ( strlenutf8( text ) + 1 ) then
        if sameLineOnly then
            return
        end

        if self[1] == self.model.btree:size() then
            return
        end

        self:moveRight( keepSelection )

        self.model.btree:findByIndex( self[1], it )
        line = itGet( it )
        text = line[2]
    end

    local i = zu.UTF8IndexToByteIndex( text, self[2] )
    local isSpaceOrPunc = spaceAndPunc[text:byte( i )]

    while i <= #text do
        local c = text:byte( i )
        local isSoP = spaceAndPunc[c]

        if isSoP ~= isSpaceOrPunc then
            break
        end

        i = i + 1
    end

    self:moveTo( self[1], zu.ByteIndexToUTF8Index( text, i ), keepSelection )
end

function EditorCursor:replaceText( textArray, moveToEnd, wholeLineIfNoSelection, reason, indenter )
    local firstLine = self[1]
    local lastLine = self[3] or self[1]
    local firstCol = self[2]
    local lastCol = self[4] or self[2]

    if firstLine > lastLine then
        firstLine, lastLine = lastLine, firstLine
        firstCol, lastCol = lastCol, firstCol
    elseif firstLine == lastLine and firstCol > lastCol then
        firstCol, lastCol = lastCol, firstCol
    end

    self[3] = nil
    self[4] = nil
    local cline
    self.model.btree:findByIndex( firstLine, it )
    cline = itGet( it )

    local newCursorLine = 1
    local newCursorCol = 1

    if moveToEnd then
        newCursorLine = #textArray
        newCursorCol = strlenutf8( textArray[#textArray] ) + 1
    end

    if indenter then
        local prevLineState

        if itPrev( it ) then
            prevLineState = itGet( it )[5]
        end

        local headLen = zu.UTF8IndexToByteIndex( cline[2], firstCol - 1 )
        local nfc, ncl, ncc = indenter( textArray, cline[2]:sub( 1, headLen ), prevLineState, cline[5] )
        newCursorLine = ncl or newCursorLine
        newCursorCol = ncc or newCursorCol
        firstCol = nfc or firstCol
    end

    if wholeLineIfNoSelection and firstLine == lastLine and firstCol == lastCol then
        lastLine = lastLine + 1
        firstCol = 1
        lastCol = 1
    end

    local removedText, lineCountChange, columnCountChange =
        self.model:replaceText( textArray, firstLine, firstCol, lastLine, lastCol, reason )

    lastLine = firstLine + newCursorLine - 1
    lastCol = newCursorCol

    if lastLine == firstLine then
        lastCol = lastCol + firstCol - 1
    end

    self.model.btree:findByIndex( lastLine, it )
    cline = itGet( it )
    self[1] = lastLine
    self[2] = lastCol

    -- Update the cursorinfo on the cursor's line
    local curInfos = cline[3][self] or zu.CreateTable( "cursorInfos" )
    cline[3][self] = curInfos
    curInfos[1] = firstCol
    curInfos[2] = nil
    curInfos[3] = nil

    return removedText, lineCountChange, columnCountChange, firstLine, firstCol, lastLine, lastCol
end

local function countLeadingSpaces( str )
    for i=1,#str,1 do
        if str:byte( i ) ~= 32 then
            return i - 1
        end
    end

    return #str
end

function EditorCursor:indentLines( indention, indentRight, undoStep )
    local model = self.model
    local firstLine = self[1]
    local lastLine = self[3] or firstLine
    local firstCol = self[2]
    local lastCol = self[4] or firstCol

    if firstLine > lastLine then
        firstLine, lastLine = lastLine, firstLine
        firstCol, lastCol = lastCol, firstCol
    end

    self.model.btree:findByIndex( firstLine, it )

    for l=firstLine,lastLine,1 do
        local line = itGet( it )
        local spaces = countLeadingSpaces( line[2] )
        local numSpacesToAdd = 0

        if indentRight or spaces > 0 then
            numSpacesToAdd = -( spaces % indention )

            if indentRight then
                numSpacesToAdd = numSpacesToAdd + indention
            end

            if numSpacesToAdd == 0 then
                numSpacesToAdd = -4
            end

            local indent = ""
            for _=1,abs( numSpacesToAdd ),1 do
                indent = indent .. " "
            end

            local undoOperation

            if numSpacesToAdd > 0 then
                line[2] = indent .. line[2]
                undoOperation = { l, 1, 1, numSpacesToAdd, { "" } }
            else

                line[2] = line[2]:sub( -numSpacesToAdd + 1 )
                undoOperation = { l, 1, 1, 0, { indent } }
            end

            undoStep[#undoStep+1] = undoOperation

            if l == firstLine then
                firstCol = max( firstCol + numSpacesToAdd, 1 )
                self[2] = max( self[2] + numSpacesToAdd, 1 )
            end

            if l == lastLine then
                lastCol = max( lastCol + numSpacesToAdd, 1 )

                if self[4] then
                    self[4] = max( self[4] + numSpacesToAdd, 1 )
                end
            end
        end

        -- Destroy every other cursor on this line and
        -- update our own curInfos
        for k,v in pairs( line[3] ) do
            if k ~= self then
                k:destroy()
            elseif l == firstLine or l == lastLine then
                if v[1] then
                    v[1] = self[2]
                end

                if v[2] and l == firstLine then
                    v[2] = firstCol
                end

                if v[3] and l == lastLine then
                    v[3] = lastCol
                end
            end
        end

        model:queueLineUpdate( line, true )
        model:queueLineRecolor( it, l )
        model.btree:updateSize( line, strlenutf8( line[2] ) * model.charWidth, line.height )
        model:queueTextChangedEvent()

        itNext( it )
    end
end
