--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The text editor widget.
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local EditorClass = {}
zu.EditorClass = EditorClass
setmetatable( EditorClass, { __index = zu.CompositeClass } )

-- Default values, can be overriden by the template or inside individual instances
EditorClass.text = ""
EditorClass.indention = 4

local defaultFont = CreateFont( "zuDefaultEditorFont" )
defaultFont:SetFont( "Interface\\AddOns\\Smack_UI\\zlui\\media\\Hack-Regular.ttf", 16, "" )
EditorClass.font = defaultFont

function EditorClass:new( tpl )
    local o = {}
    o.textChangedEvent = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function zu.Editor( s )
    return EditorClass:new( s )
end

local function updateSliders( self )
    if not self.height then
        return
    end

    local selfModelHeight = self.height / self.view.lineHeightScale
    local totalW, totalH = self.model:getSize()
    totalW = totalW + self.model.charWidth * 5
    totalH = totalH + ( selfModelHeight - ( self.model.lineHeight * 3 ) )

    totalW = totalW * self.view.charWidthScale
    totalH = totalH * self.view.lineHeightScale

    self.view.scrollableWidth = totalW
    self.view.scrollableHeight = totalH

    self.hSlider:setTotalAndVisible( totalW, self.view.width, self.view.width )
    self.vSlider:setTotalAndVisible( totalH, self.view.height, self.view.height )

    self.hSlider:setValue( self.view.scrollXPosition )
    self.vSlider:setValue( self.view.scrollYPosition )
end

local spacerTpl = {}

function EditorClass:create( parentFrame )
    if self.model then
        return
    end

    self.model = zu.EditorModel:new()
    self.model.colorizer = self.colorizer
    self.font:SetTextColor( self:getSkin():getEditorTextColor() )
    self.model:setTextBuffer( self.text )

    self.model.textChangedEvent = self.textChangedEvent

    self.view = zu.EditorView {}
    self.view.model = self.model
    self.view.font = self.font

    self.controller = zu.EditorController:new( self.model )
    self.controller.indention = self.indention
    self.controller.indenter = function( textArray, currentLineHead, prevState, currentState )
        if self.indenter then
            return self.indenter( textArray, currentLineHead, prevState, currentState, self.indention )
        end
    end

    self.controller.doubleClickSelectsAll = self.doubleClickSelectsAll

    self.hSlider = zu.HorizontalSlider
    {
        background = false,
        fade = true,
        autoFadeOut = true,
        alpha = 0.4,
        onValueUpdated = function( v )
            self.view:setScrollXPosition( v )
        end
    }

    self.vSlider = zu.VerticalSlider
    {
        background = false,
        fade = true,
        alpha = 0.4,
        onValueUpdated = function( v )
            self.view:setScrollYPosition( v )
        end
    }

    self.content = zu.OverlayLayout
    {
        zu.HorizontalLayout
        {
            zu.VerticalLayout
            {
                zu.Spacer( spacerTpl ),
                self.hSlider
            },
            self.vSlider
        },

        self.view
    }

    self.content.parent = self
    self.content:create( parentFrame )

    self.controller:setView( self.view )

    zu.AddListener( self.controller.scrollUpdateEvent, self, updateSliders )
    zu.AddListener( self.model.btreeUpdatedEvent, self, updateSliders )

    if self.onMouseUp then
        zu.AddListener( self.controller.mouseUpEvent, self, self.onMouseUp )
    end

    if self.onMouseDown then
        zu.AddListener( self.controller.mouseDownEvent, self, self.onMouseDown )
    end

    if self.onGainFocus then
        zu.AddListener( self.controller.gainFocusEvent, self, self.onGainFocus )
    end
    if self.onLoseFocus then
        zu.AddListener( self.controller.loseFocusEvent, self, self.onLoseFocus )
    end

    if self.onViewScrolled then
        zu.AddListener( self.view.scrolledEvent, self, self.onViewScrolled )
    end

    self.view.mouseHoverUpdate = self.mouseHoverUpdate
end

function EditorClass:show()
    zu.CompositeClass.show( self )

    self.controller:unfreeze()

    local slidersShown = false
    self.sliderFader = C_Timer.NewTicker( 0.1,
        function()
            if self.frozen then
                return
            end

            if self.view.frame:IsMouseOver() then
                if not slidersShown then
                    self.vSlider:fadeIn()
                    self.hSlider:fadeIn()
                    slidersShown = true
                end
            else
                if slidersShown then
                    self.vSlider:fadeOut()
                    self.hSlider:fadeOut()
                    slidersShown = false
                end
            end
        end )
end

function EditorClass:hide()
    self.controller:freeze()

    if self.sliderFader then
        self.sliderFader:Cancel()
        self.sliderFader = nil
    end
    zu.CompositeClass.hide( self )
end

function EditorClass:freeze()
    zu.CompositeClass.freeze( self )
    self.controller:freeze()
end

function EditorClass:unfreeze()
    zu.CompositeClass.unfreeze( self )
    self.controller:unfreeze()
end


function EditorClass:release()
    if self.sliderFader then
        self.sliderFader:Cancel()
        self.sliderFader = nil
    end
    zu.CompositeClass.release( self )
end

function EditorClass:layout( x, y, width, height, minLevel, maxLevel )
    zu.CompositeClass.layout( self, x, y, width, height, minLevel, maxLevel )
    updateSliders( self )
end

function EditorClass:isEmpty()
    return self.model:isEmpty()
end

function EditorClass:setText( text )
    self.controller:setText( text )
end

function EditorClass:getText( trim_trailing_whitespace, timeslicer )
    return self.model:getText( trim_trailing_whitespace, timeslicer )
end

function EditorClass:pasteText( text, reason )
    self.controller:pasteText( text, reason )
end

function EditorClass:lineColToCoords( l, c )
    return self.view:lineColToCoords( l, c )
end

function EditorClass:getPrimaryCursorPosition()
    return self.controller:getPrimaryCursorPosition()
end

function EditorClass:setPrimaryCursorPosition( fl, fc )
    self.controller:setPrimaryCursorPosition( fl, fc )
end

function EditorClass:setPrimaryCursorExtent( fl, fc, ll, lc )
    self.controller:setPrimaryCursorExtent( fl, fc, ll, lc )
end

function EditorClass:focus()
    self.controller:focus()
end

function EditorClass:haveFocus()
    return not self.frozen and self.controller.haveFocus
end

function EditorClass:attachPopup( popup )
    self.controller:attachPopup( popup )
end

function EditorClass:detachPopup( popup )
    self.controller:detachPopup( popup )
end
