--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local EditorLineClass = {}
zu.EditorLineClass = EditorLineClass
setmetatable( EditorLineClass, { __index = zu.Widget } )

function EditorLineClass:new( tpl )
    local o = {}
    o.cursorFrames = {}
    o.selectionFrames = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function EditorLineClass:release( view )
    zu.Widget.release( self )
    zu.RemoveListener( view.cursorsVisiblityEvent, self )
    zu.RemoveListener( self.line.updateEvent, self )

    local cursorFrames = self.cursorFrames
    for i=1,#cursorFrames,1 do
        zu.ReleaseFrame( "cursor", cursorFrames[i] )
        cursorFrames[i] = nil
    end

    local selectionFrames = self.selectionFrames
    for i=1,#selectionFrames,1 do
        zu.ReleaseFrame( "selection", selectionFrames[i] )
        selectionFrames[i] = nil
    end

    local decs = self.line.decs
    if not decs then
        return
    end

    for k,_ in pairs( decs ) do
        k:detach( self )
    end
end

local function updateCursorsVisibility( self, show )
    local cursorFrames = self.cursorFrames

    if show then
        for i=1,#cursorFrames,1 do
            cursorFrames[i]:Show()
        end
    else
        for i=1,#cursorFrames,1 do
            cursorFrames[i]:Hide()
        end
    end
end

-- (re)create frames to represent all the cursors and selections
-- present on this line, if any. Also update the text if necessary.
local function update( self, textChanged )
    if textChanged then
        self.text:SetText( self.line[4] or self.line[2] )
        self.lineLength = strlenutf8( self.line[2] )
    end

    local lineLength = self.lineLength

    local cursorFrames = self.cursorFrames
    for i=1,#cursorFrames,1 do
        zu.ReleaseFrame( "cursor", cursorFrames[i] )
        cursorFrames[i] = nil
    end

    local selectionFrames = self.selectionFrames
    for i=1,#selectionFrames,1 do
        zu.ReleaseFrame( "selection", selectionFrames[i] )
        selectionFrames[i] = nil
    end

    local charWidth = self.view.charWidth
    local skin = self:getSkin()

    for _,curInfos in pairs( self.line[3] ) do
        if curInfos[1] then
            local curFrame = zu.CreateFrame( "cursor", "Frame", self.frame )

            if not curFrame.inited then
                local tex = curFrame:CreateTexture( nil, "BACKGROUND" )
                tex:SetAllPoints( curFrame )
                tex:SetColorTexture( skin:getEditorCursorColor() )
                curFrame.inited = true
            end

            local xpos = ( curInfos[1] - 1 ) * charWidth
            curFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", xpos, 0 )
            curFrame:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMLEFT", xpos + zu.Scale( self.cursorThickness ), 0 )

            if self.view.showCursors then
                curFrame:Show()
            end

            cursorFrames[#cursorFrames+1] = curFrame
        end

        if curInfos[2] then
            local selFrame = zu.CreateFrame( "selection", "Frame", self.frame )

            if not selFrame.inited then
                local tex = selFrame:CreateTexture( nil, "BACKGROUND" )
                tex:SetAllPoints( selFrame )
                tex:SetColorTexture( skin:getEditorSelectionColor() )
                selFrame.inited = true
            end

            selFrame:SetFrameLevel( self.frame:GetFrameLevel() - 1 )
            local xStart = ( curInfos[2] - 1 ) * charWidth

            local endCol = curInfos[3] or ( lineLength + 2 )
            local xEnd = ( endCol - 1 ) * charWidth
            selFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", xStart, 0 )
            selFrame:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMLEFT", xEnd, 0 )
            selFrame:Show()

            selectionFrames[#selectionFrames+1] = selFrame
        end
    end

    local decs = self.line.decs
    if not decs then
        return
    end

    for k,_ in pairs( decs ) do
        k:attach( self )
    end
end

function EditorLineClass:create( parentFrame, view, lineNum )
    self.lineNum = lineNum

    if self.frame then
        self.frame:ClearAllPoints()
        self.frame:SetParent( parentFrame )
    else
        self.frame = CreateFrame( "Frame", nil, parentFrame )
        self.text = self.frame:CreateFontString( nil, "ARTWORK" )
        self.text:SetFontObject( view.font )
        self.text:SetJustifyH( "LEFT" )
        self.text:SetJustifyV( "CENTER" )
        self.text:SetAllPoints( self.frame )
   end

  zu.AddListener( self.line.updateEvent, self, update )
  zu.AddListener( view.cursorsVisiblityEvent, self, updateCursorsVisibility )
  self.view = view

  self.text:Show()
  update( self, true )
end

function EditorLineClass:layout( x, y, width, height )
    self.text:SetTextScale( zu.ContentScaleFactor() )

    local topInset = zu.Scale( self.line.topInset or 0 )
    local bottomInset = zu.Scale( self.line.bottomInset or 0 )

    self.x = x
    self.y = y
    self.frame:SetSize( width + 64, height - topInset - bottomInset )
    self.frame:SetPoint( "TOPLEFT", self.frame:GetParent(), "TOPLEFT", x, -y - topInset )

    self:update()
end

function EditorLineClass:update()
    update( self )

    local decs = self.line.decs
    if not decs then
        return
    end

    for k,_ in pairs( decs ) do
        k:update( self )
    end
end

function EditorLineClass:decorationHitTest( x, y )
    local decs = self.line.decs
    if not decs then
        return
    end

    for k,_ in pairs( decs ) do
        if k:hitTest( self, x, y ) then
            return k
        end
    end
end
