--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The editor's data model.
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local itIsEnd = zu.BTree.itIsEnd
local itPrev = zu.BTree.itPrev
local itNext = zu.BTree.itNext
local itGet = zu.BTree.itGet
local itCopy = zu.BTree.itCopy

local min = math.min
local floor = math.floor

local concat = table.concat

local yield = coroutine.yield
local wrap = coroutine.wrap

local EditorModel = {}

-- We want the model to remain display independent and since wow insists
-- on injecting some fucking hidden scale factor everywhere all the time depending
-- on the wow window's size we have absolutely no reliable way of knowing the size
-- of the actual font in a resolution independent way,
-- so we have to hardcode a "normalized" character size in the
-- model and let the view rescale this according to the current wow font size
-- after being fucked up by wow's "magic" scaling.
EditorModel.charWidth = 10
EditorModel.charHeight = 20
EditorModel.lineHeight = EditorModel.charHeight + 2

zu.EditorModel = EditorModel

local it = {}
local it2 = {}

-- Misc utiliy functions
function zu.LineGenerator( text )
    local lineArray = { strsplit( "\n\r", text ) }

    local function gen()
        for i=1,#lineArray,1 do
            yield( lineArray[i] )
        end
    end

    return wrap( gen )
end

-- Take a string and the index of an utf-8 codepoint within that string,
-- and return a byte index suitable for use with string.sub()
function zu.UTF8IndexToByteIndex( str, utf8Index )
    local result = utf8Index
    local remaining = utf8Index

    for i=1,#str,1 do
        local c = str:byte( i )

        if c >= 128 and c < 192 then
            result = result + 1
        else
            remaining = remaining - 1

            if remaining == 0 then
                break
            end
        end
    end

    return result
end

-- Take a string and the index of a byte within that string,
-- and return an utf-8 codepoint index. We assume that i points to
-- the first byte of a code point.
function zu.ByteIndexToUTF8Index( str, byteIndex )
    local result = byteIndex
    local remaining = byteIndex

    for i=1,#str,1 do
        local c = str:byte( i )

        if c >= 128 and c < 192 then
            result = result - 1
        end

        remaining = remaining - 1

        if remaining == 0 then
            break
        end
    end

    return result
end

function EditorModel:new()
    local o = {}

    o.btree = zu.BTree:new()
    o.btreeUpdatedEvent = {}
    o.pendingEvents = {}
    o.pendingLineColoring = { {}, {} }
    o.decorations = {}

    setmetatable( o, { __index = self } )
    return o
end

-- Replace the entire text buffer.
function EditorModel:setTextBuffer( text )
    self.btree:clear()

    local lineGen = zu.LineGenerator( text )
    local lineHeight = self.lineHeight
    local charWidth = self.charWidth

    local colorizerState

    local function gen()
        local line = lineGen()
        if line == nil then
            return
        end

        local lineData = { nil, line, {}, -- 1: parent btree node, 2: text, 3: cursors/selections, 4: colored text (optional), 5: colorizer state
            height = lineHeight,
            width = strlenutf8( line ) * charWidth,
            updateEvent = {} }

        if self.colorizer then
            lineData[4], colorizerState = self.colorizer( line, colorizerState )
            lineData[5] = colorizerState
        end

        return lineData
    end

    self.btree:insert( 1, gen )
end

-- Same as above, but to replace the text buffer after initializing
-- the editor.
function EditorModel:replaceTextBuffer( text )
    self:setTextBuffer( text )
    self:queueBTreeUpdate()
end

function EditorModel:getSize()
    return self.btree:width(), self.btree:height()
end

function EditorModel:YCoordToLine( y )
    y = min( y, self.btree:height() - 1 )

    local l, ryp = self.btree:findByYPosition( y, it )

    if itIsEnd( it ) then
        return
    end

    local line = itGet( it )
    return l, line, ryp
end

function EditorModel:coordsToLineCol( x, y )
    y = min( y, self.btree:height() - 1 )

    local l = self.btree:findByYPosition( y, it )

    if itIsEnd( it ) then
        return
    end

    local line = itGet( it )
    local col = floor( x / self.charWidth ) + 1
    col = min( col, strlenutf8( line[2] ) + 1 )

    return l, col, line
end

function EditorModel:lineColToCoords( l, c )
    local ypos = self.btree:findYPositionByIndex( l )
    if not ypos then
        return
    end

    return ( c - 1 ) * self.charWidth, ypos
end

function EditorModel:queueLineUpdate( line, textChanged )
    if #line.updateEvent then
        self.pendingEvents[line.updateEvent] = self.pendingEvents[line.updateEvent] or textChanged or false
    end
end

function EditorModel:queueLineRecolor( lineit, lineNum )
    if self.colorizer then
        local line = itGet( lineit )
        self.pendingLineColoring[1][lineNum] = itCopy( lineit )
        self.pendingLineColoring[2][line] = lineNum
    end
end

function EditorModel:queueBTreeUpdate()
    self.pendingBTreeUpdate = true
end

function EditorModel:queueTextChangedEvent( reason )
    self.pendingTextChangedEvent = true
    self.pendingTextChangeReason = reason
end

function EditorModel:flushPendingEvents()
   if self.colorizer then
        for _,lineit in pairs( self.pendingLineColoring[1] ) do
            local state

            local prevLineIt = itCopy( lineit )
            itPrev( prevLineIt )
            if not itIsEnd( prevLineIt ) then
                local prevline = itGet( prevLineIt )
                state = prevline and prevline[5] or nil
            end

            local line = itGet( lineit )

            -- Recolor the line and all following lines as long as the colorizer
            -- starting state for those lines has changed.
            -- Remove any encountered line from the pending recoloring list,
            -- to avoid doing them multiple times.
            repeat
                line[4], state = self.colorizer( line[2], state )

                local oldState = line[5]
                line[5] = state

                self:queueLineUpdate( line, true )

                local index = self.pendingLineColoring[2][line]
                if index then
                    self.pendingLineColoring[1][index] = nil
                end

                itNext( lineit )
                if itIsEnd( lineit ) then
                    break
                end

                line = itGet( lineit )
            until oldState == state
        end

        wipe( self.pendingLineColoring[1] )
        wipe( self.pendingLineColoring[2] )
    end

    if self.pendingBTreeUpdate then
        self.pendingBTreeUpdate = false
        zu.Raise( self.btreeUpdatedEvent )
    end

    for k, textChanged in pairs( self.pendingEvents ) do
        zu.Raise( k, textChanged )
        self.pendingEvents[k] = nil
    end

    if self.pendingTextChangedEvent then
        self.pendingTextChangedEvent = false
        zu.Raise( self.textChangedEvent, self.pendingTextChangeReason )
    end
end

function EditorModel:replaceText( newTextArray, firstLine, firstCol, lastLine, lastCol, reason )
    local extractedText = {}

    self.btree:findByIndex( firstLine, it )
    local fline = itGet( it )

    if lastLine == firstLine and lastLine < self.btree:size() and lastCol > ( strlenutf8( fline[2] ) + 1 ) then
        -- If we are stopping past the end of the first line it means we are deleting its
        -- newline character, so act as if the selection extended to the first column of the next line
        lastLine = lastLine + 1
        lastCol = 1
    end

    self.btree:findByIndex( lastLine, it2 )
    local lline = itGet( it2 )

    local firstColByteIndex = zu.UTF8IndexToByteIndex( fline[2], firstCol )
    local lastColByteIndex = zu.UTF8IndexToByteIndex( lline[2], lastCol )

    local tail = lline[2]:sub( lastColByteIndex )

    -- Grab the sections of texts removed on the first and the last line.
    local lastRemovedLine
    if firstLine == lastLine then
        extractedText[1] = fline[2]:sub( firstColByteIndex, lastColByteIndex - 1 )
    else
        extractedText[1] = fline[2]:sub( firstColByteIndex )
        lastRemovedLine = lline[2]:sub( 1, lastColByteIndex - 1 )
    end

    -- First line: keep the part preceding the insertion range,
    -- append the first line to insert. As a special case, if the replacement text
    -- is only one line long, also append the tail part of the last line,
    -- then delete all the other lines and exit.
    fline[2] = fline[2]:sub( 1, firstColByteIndex - 1 ) .. newTextArray[1]

    for k,v in pairs( fline[3] ) do
        if v[1] and v[1] >= firstCol or v[2] and v[2] >= firstCol then
            zu.ReleaseTable( "cursorInfos", v )
            fline[3][k] = nil
        end
    end

    local columnCountChange
    if #newTextArray == 1 then
        -- Special case when the inserted text is only one line long:
        -- the only line we want to keep is the first one, onto which we need
        -- to append the tail of the last line to remove.
        columnCountChange = strlenutf8( fline[2] ) - lastCol + 1
        fline[2] = fline[2] .. tail
    else
        columnCountChange = strlenutf8( newTextArray[#newTextArray] ) - lastCol + 1
    end

    -- Update the first line's width in the btree
    self.btree:updateSize( fline, strlenutf8( fline[2] ) * self.charWidth, fline.height )
    self:queueLineUpdate( fline, true )

    -- Since we may insert many new lines and we don't want to recolor them right now
    -- but we can't get their iterator to queue them up for later either,
    -- clear the coloring state for the first line and every subsequent line.
    -- This way, recoloring the first line will trigger a recoloring of all affected
    -- and newly inserted lines in cascade.
    fline[5] = nil

    -- Intermediate lines: while we are still inside of the interval to be replaced
    -- and we still have more lines to replace it with, just update the line's content.
    -- After that, if we have run out of existing lines to update, insert the rest as
    -- new lines. Otherwise, remove the excedent existing lines.
    local i = 2
    local currentLineToRemove = firstLine + 1
    local lastLineToRemove = lastLine

    itNext( it )
    -- Update the lines that can be with the new text
    while i <= #newTextArray and currentLineToRemove <= lastLineToRemove do
        local line = itGet( it )

        if i == #newTextArray then
            line[2] = newTextArray[i] .. tail
        else
            extractedText[#extractedText+1] = line[2]
            line[2] = newTextArray[i]
        end

        -- Clear all the cursorInfos on that line
        for k,v in pairs( line[3] ) do
            zu.ReleaseTable( "cursorInfos", v )
            line[3][k] = nil
        end

        self.btree:updateSize( line, strlenutf8( line[2] ) * self.charWidth, line.height )
        self:queueLineUpdate( line, true )

        line[5] = nil

        itNext( it )
        i = i + 1
        currentLineToRemove = currentLineToRemove + 1
    end

    local lineCountChange = 0

    if i <= #newTextArray then
        -- We have more lines in the incoming text but we already updated all the existing lines,
        -- we need to insert additional lines
        local lineHeight = self.lineHeight
        local charWidth = self.charWidth

        lineCountChange = #newTextArray - i + 1

        local function gen()
            for j=i,#newTextArray,1 do
                local line = newTextArray[j]

                if j == #newTextArray then
                    line = line .. tail
                end

                yield { nil, line, {}, -- 1: parent btree node, 2: text, 3: cursors/selections, 4: colored text (optional), 5: colorizer state
                    height = lineHeight,
                    width = strlenutf8( line ) * charWidth,
                    updateEvent = {} }
            end
        end

        self.btree:insert( currentLineToRemove, wrap( gen ) )
    elseif currentLineToRemove <= lastLineToRemove then
        -- We finished the text to insert but there are still more existing lines to delete:
        -- save them in the removed text array and remove them from the btree
        local firstLineToRemove = currentLineToRemove
        local numLinesToRemove = lastLineToRemove - firstLineToRemove + 1

        lineCountChange = -numLinesToRemove

        while currentLineToRemove <= lastLineToRemove do
            local line = itGet( it )

            if currentLineToRemove ~= lastLineToRemove then
                extractedText[#extractedText+1] = line[2]
            end

            for _,v in pairs( line[3] ) do
                zu.ReleaseTable( "cursorInfos", v )
            end

            -- TODO detach decorations from these lines


            itNext( it )
            currentLineToRemove = currentLineToRemove + 1
        end

        self.btree:erase( firstLineToRemove, numLinesToRemove )
    end

    extractedText[#extractedText+1] = lastRemovedLine
    self:queueBTreeUpdate()
    self:queueTextChangedEvent( reason )

    -- Look up the first line again to queue it for recoloring:
    -- we can't use the iterator we got at the beginning of the function
    -- as insertions/removals from the btree may have invalidated it
    self.btree:findByIndex( firstLine, it )
    self:queueLineRecolor( it, firstLine )

    -- Return the removed text (as an array of strings), the delta of the number of
    -- lines (how many were added/deleted), and the delta of the number of column
    -- on the last line
    return extractedText, lineCountChange, columnCountChange
end

local etit = {}
function EditorModel:extractText( textArray, firstLine, firstCol, lastLine, lastCol )
    self.btree:findByIndex( firstLine, etit )
    local fline = itGet( etit )

    if firstLine == lastLine then
        -- If the span is of length 0, extract the entire line instead.
        -- (this is the same as vscode's ctrl+c behavior)
        if firstCol == lastCol then
            textArray[#textArray+1] = fline[2]
            return
        end

        local firstColByteIndex = zu.UTF8IndexToByteIndex( fline[2], firstCol )
        local lastColByteIndex = zu.UTF8IndexToByteIndex( fline[2], lastCol )
        textArray[#textArray+1] = fline[2]:sub( firstColByteIndex, lastColByteIndex - 1 )
        return
    end

    local firstColByteIndex = zu.UTF8IndexToByteIndex( fline[2], firstCol )
    textArray[#textArray+1] = fline[2]:sub( firstColByteIndex )
    firstLine = firstLine + 1
    itNext( etit )

    while firstLine <= lastLine do
        local line = itGet( etit )

        if firstLine == lastLine then
            local lastColByteIndex = zu.UTF8IndexToByteIndex( line[2], lastCol )
            textArray[#textArray+1] = line[2]:sub( 1, lastColByteIndex - 1 )
        else
            textArray[#textArray+1] = line[2]
        end

        itNext( etit )
        firstLine = firstLine + 1
    end
end

function EditorModel:getText( trim_trailing_whitespace, timeslicer )
    local textArray = {}
    local it_local = {}

    self.btree:itFirst( it_local )

    while not itIsEnd( it_local ) do
        local line = itGet( it_local )[2]

        if trim_trailing_whitespace then
            local i = #line
            repeat
                local c = line:byte(i)
                i = i - 1
            until c ~= 32 and c ~= 9
            line = line:sub( 1, i + 1 )
        end

        textArray[#textArray+1] = line
        itNext( it_local )

        if timeslicer then
            timeslicer()
        end
    end

    if trim_trailing_whitespace then
        for i=#textArray,1,-1 do
            if #textArray[i] > 0 then
                break
            end
            textArray[i] = nil

            if timeslicer then
                timeslicer()
            end
        end
        textArray[#textArray+1] = ""
    end

    return concat( textArray, "\n" )
end

function EditorModel:isEmpty()
    if self.btree:size() ~= 1 then
        return
    end

    self.btree:itFirst( it )
    local line = itGet( it )
    return line[2] == ""
end

function EditorModel:addDecoration( decoration )
    local decLines = {}
    self.decorations[decoration] = decLines

    local topInset = decoration.topInset or 0
    local bottomInset = decoration.bottomInset or 0

    local firstLine = decoration.firstLine
    self.btree:findByIndex( firstLine, it )

    for _ = firstLine, decoration.lastLine, 1 do
        if itIsEnd( it ) then
            break
        end

        local line = itGet( it )
        decLines[line] = true

        local decInfos = line.decs or {}

        decInfos[decoration] = true
        line.decs = decInfos
        local newTopInset = max( topInset, line.topInset or 0 )
        local newBottomInset = max( bottomInset, line.bottomInset or 0 )

        if newTopInset ~= line.topInset or bottomInset ~= line.bottomInset then
            line.topInset = newTopInset
            line.bottomInset = newBottomInset
            self.btree:updateSize( line, line.width, self.lineHeight + newTopInset + newBottomInset )
            self:queueBTreeUpdate()
        end

        self:queueLineUpdate( line )
        itNext( it )
    end
end

function EditorModel:removeDecoration( decoration )
    local decLines = self.decorations[decoration]

    if not decLines then
        return
    end

    self.decorations[decoration] = nil

    for line,_ in pairs( decLines ) do
        local decInfos = line.decs

        if decInfos then
            decInfos[decoration] = nil
        end

        -- Recompute insets
        local topInset = 0
        local bottomInset = 0
        for k, _ in pairs( decInfos ) do
            topInset = max( topInset, k.topInset or 0 )
            bottomInset = max( bottomInset, k.bottomInset or 0 )
        end

        if topInset ~= line.topInset or bottomInset ~= line.bottomInset then
            line.topInset = topInset
            line.bottomInset = bottomInset
            self.btree:updateSize( line, line.width, self.lineHeight + topInset + bottomInset )
            self:queueBTreeUpdate()
        end

        self:queueLineUpdate( line )
        itNext( it )
    end
end

function EditorModel:clearDecorations()
    for k, decLines in pairs( self.decorations ) do
        for line,_ in pairs( decLines ) do
            line.decs = nil

            -- Remove insets
            if line.topInset or line.bottomInset then
                line.topInset = nil
                line.bottomInset = nil
                self.btree:updateSize( line, line.width, self.lineHeight )
                self:queueBTreeUpdate()
            end

            self:queueLineUpdate( line )
        end

        zu.ReleaseWidget( k )
    end

    wipe( self.decorations )
    self:flushPendingEvents()
end

function EditorModel:updateDecorations()
    self:flushPendingEvents()
end
