--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local floor = math.floor

local EditorUndoBuffer = {}
zu.EditorUndoBuffer = EditorUndoBuffer

function EditorUndoBuffer:new( model )
    local o = {}

    o.currentUndoStep = 0
    o.model = model

    setmetatable( o, { __index = self } )
    return o
end

function EditorUndoBuffer:snapshotCursors( cursors )
    -- If the current undo step is not the last undo step in the buffer,
    -- clear all subsequent ones.
    if #self > 0 then
        for i=(self.currentUndoStep + 1) * 2,#self,1 do
            self[i] = nil
        end
    end

    local snapshot

    -- If the last entry in the undo buffer is a cursor snapshot (odd index), overwrite it.
    -- Otherwise, create a new one.
    if #self % 2 == 1 then
        snapshot = self[#self]
        wipe( snapshot )
    else
        snapshot = {}
        self[#self+1] = snapshot
    end

    for i=1,#cursors,1 do
        local cursor = cursors[i]
        local cs = {}
        cs[1] = cursor[1]
        cs[2] = cursor[2]
        cs[3] = cursor[3]
        cs[4] = cursor[4]

        snapshot[#snapshot+1] = cs
    end
end

function EditorUndoBuffer:pushUndoStep( undoStep )
    self.currentUndoStep = self.currentUndoStep + 1
    self[self.currentUndoStep * 2] = undoStep
end

function EditorUndoBuffer:canUndo()
    return self.currentUndoStep > 0
end

function EditorUndoBuffer:canRedo()
    return self.currentUndoStep < floor( #self / 2 )
end

function EditorUndoBuffer:undo()
    local undoStep = self[self.currentUndoStep * 2]
    self.currentUndoStep = self.currentUndoStep - 1

    local model = self.model

    -- Perform the text replacements from last to first: the insertion positions
    -- are recorded after the previous cursors have been processed, so we don't want to invalidate
    -- them by undoing the previous cursors first.
    for i=#undoStep,1,-1 do
        local undoOperation = undoStep[i]
        local lastLine = undoOperation[3] + undoOperation[1] - 1
        local lastCol = undoOperation[4]
        if undoOperation[3] == 1 then
            lastCol = lastCol + undoOperation[2]
        else
            lastCol = lastCol + 1
        end

        local undoText = undoOperation[5]

        -- Swap the old text with the text that it was replaced with, so that we can redo the operation later
        -- if needed.
        undoOperation[5] = model:replaceText( undoText, undoOperation[1], undoOperation[2], lastLine, lastCol, "undo" )

        undoOperation[3] = #undoText
        undoOperation[4] = strlenutf8( undoText[#undoText] )
    end

    return self[self.currentUndoStep * 2 + 1]
end

function EditorUndoBuffer:redo()
    self.currentUndoStep = self.currentUndoStep + 1
    local undoStep = self[self.currentUndoStep * 2]

    local model = self.model

    -- Perform the text replacements from first to last, for the same reason we do them last to first
    -- in undo().
    for i=1,#undoStep,1 do
        local undoOperation = undoStep[i]
        local lastLine = undoOperation[3] + undoOperation[1] - 1
        local lastCol = undoOperation[4]
        if undoOperation[3] == 1 then
            lastCol = lastCol + undoOperation[2]
        else
            lastCol = lastCol + 1
        end

        local undoText = undoOperation[5]

        -- Swap the old text with the text that it was replaced with, so that we can redo the operation later
        -- if needed.
        undoOperation[5] = model:replaceText( undoText, undoOperation[1], undoOperation[2], lastLine, lastCol, "redo" )

        undoOperation[3] = #undoText
        undoOperation[4] = strlenutf8( undoText[#undoText] )
    end

    return self[self.currentUndoStep * 2 + 1]
end
