--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- The core display view of the editor.
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local itIsEnd = zu.BTree.itIsEnd
local itNext = zu.BTree.itNext
local itGet = zu.BTree.itGet

local min = math.min
local max = math.max
local floor = math.floor

local EditorViewClass = {}
zu.EditorViewClass = EditorViewClass
setmetatable( EditorViewClass, { __index = zu.Widget } )

EditorViewClass.levelRange = 1

local it = {}

EditorViewClass.cursorThickness = 1
EditorViewClass.canScroll = true

function EditorViewClass:new( tpl )
    local o = {}

    o.scrollXPosition = 0
    o.scrollYPosition = 0
    o.scrollableHeight = 0
    o.scrollableWidth = 0
    o.lineWidgets = {}
    o.spareLineWidgets = {}
    o.width = 0
    o.height = 0
    o.showCursors = true
    o.focus = false
    o.cursorsVisiblityEvent = {}
    o.scrolledEvent = {}

    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function zu.EditorView( s )
    return EditorViewClass:new( s )
end

-- (re)create the necessary widgets to display the current view and lay them out.
local function update( self )
    -- Recompute the scaling factors to convert pixel positions to the "normalized" pixel size used
    -- by the model, since our actually displayed font size can change anytime wow's window is
    -- resized and we have zero control over it because wow sucks ass
    if not self.fontString then
        self.fontString = self.frame:CreateFontString()
        self.fontString:SetJustifyH( "LEFT" )
        self.fontString:SetJustifyV( "CENTER" )
        self.fontString:Hide()
        self.fontString:SetFontObject( self.font )
        self.fontString:SetText( "W" )
    end

    self.fontString:SetTextScale( zu.ContentScaleFactor() )

    self.charWidth = floor( self.fontString:GetUnboundedStringWidth() )
    self.charWidthScale = self.charWidth / self.model.charWidth

    local lh = floor( self.fontString:GetLineHeight() + 0.5 )
    local spc = floor( self.fontString:GetSpacing() + 0.5 )

    self.lineHeightScale = ( lh + spc ) / self.model.lineHeight

    self.currentHoveredDecoration = nil

    self.scrollXPosition = min( self.scrollXPosition, self.scrollableWidth - self.width )
    self.scrollXPosition = max( self.scrollXPosition, 0 )
    self.scrollYPosition = min( self.scrollYPosition, self.scrollableHeight - self.height )
    self.scrollYPosition = max( self.scrollYPosition, 0 )

    -- The direct access to the btree inside of the model is ugly
    -- but lua is a garbage tier language where every useful pattern such
    -- as encapsulation has a runtime cost so I'm wary of doing it in
    -- performance critical code
    local lineNum, offset = self.model.btree:findByYPosition( self.scrollYPosition / self.lineHeightScale, it )
    offset = offset or 0
    offset = zu.Scale( offset * self.lineHeightScale )

    local charWidth = self.model.charWidth
    local cursorThickness = self.cursorThickness

    -- findByYPosition returns the relative position within the found element:
    -- This is how much vertical scrolling we need for pixel precision scrolling.
    self.YScrollOffset = offset
    self.contentFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", -self.scrollXPosition, offset )

    -- All the current line widgets are now considered spare widgets:
    -- if we encounter their line again during the update, we'll reuse
    -- them. Any that hasn't been reused after the update is released.
    local spareWidgets = self.lineWidgets

    -- Initialize the new widget table with the old spare table, which
    -- should be empty.
    local widgets = self.spareLineWidgets

    local y = 0
    local maxY = self.height + offset
    local width = max( self.model.btree:width(), charWidth )
    width = floor( max( width, self.width ) * self.charWidthScale )

    local minLevel = self.minLevel
    local maxLevel = self.maxLevel

    while not itIsEnd( it ) do
        if y > maxY then
            break
        end

        local line = itGet( it )
        local lh = floor( line.height * self.lineHeightScale )

        local w = spareWidgets[line]
        if w then
            w.lineNum = lineNum
            widgets[line] = w
            spareWidgets[line] = nil
        else
            -- create new widget
            w = zu.CreateWidget( zu.EditorLineClass )
            widgets[line] = w

            w.parent = self
            w.line = line
            w.charWidth = charWidth
            w.cursorThickness = cursorThickness

            w:create( self.contentFrame, self, lineNum )
            w:show()
        end

        w:layout( 0, y, width, lh, minLevel, maxLevel )

        y = y + lh
        lineNum = lineNum + 1
        itNext( it )
    end

    self.contentFrame:SetSize( width, zu.Scale( y ) )

    for _,w in pairs( spareWidgets ) do
        zu.ReleaseWidget( w, self )
    end
    wipe( spareWidgets )

    self.lineWidgets = widgets
    self.spareLineWidgets = spareWidgets
end

function EditorViewClass:getModelRelativeMousePosition()
    local mouseX, mouseY = GetCursorPosition()
    local scale = self.frame:GetEffectiveScale()

    local x, y, _, h = self.frame:GetScaledRect()

    if not x or not y or not h then
        return
    end

    x = floor( ( mouseX - x ) / scale )
    y = floor( ( h + y - mouseY ) / scale )

    x = max( x, 0 )
    y = max( y, 0 )
    x = min( x, self.width )
    y = min( y, self.height )

    x = x + self.scrollXPosition
    y = y + self.scrollYPosition

    return x / self.charWidthScale, floor( y / self.lineHeightScale + 0.5 )
end

local function onUpdate( self )
    if self.onUpdate then
        if self.currentHoveredDecoration then
            self.currentHoveredDecoration:hovering()
            self.currentHoveredDecoration = nil
        end
        self.onUpdate()
        return
    end

    local x, y
    local l, line

    if self.mouseHoverUpdate then
        x, y = self:getModelRelativeMousePosition()
        if not x or not y then
            return
        end
        local relativeYPosition
        l, line, relativeYPosition = self.model:YCoordToLine( y )
        self.mouseHoverUpdate( self.hover, l, line, relativeYPosition, x, y )
    end

    if not self.hover then
        return
    end

    if not self.mouseHoverUpdate then
        x, y = self:getModelRelativeMousePosition()
        if not x or not y then
            return
        end
        l, line = self.model:YCoordToLine( y )
    end

    local lw = self.lineWidgets[line]
    if not lw then
        return
    end

    local dec = lw:decorationHitTest( x - lw.x, y - lw.y )

    local currentHoveredDecoration = self.currentHoveredDecoration
    if currentHoveredDecoration and dec ~= currentHoveredDecoration then
        currentHoveredDecoration:hovering()
    end

    self.currentHoveredDecoration = dec
    if dec then
        dec:hovering( lw, x - lw.x, y - lw.y )
    end
end

local function onEnter( self )
    self.hover = true
end

local function onLeave( self )
    self.hover = false
    if self.currentHoveredDecoration then
        self.currentHoveredDecoration:hovering()
        self.currentHoveredDecoration = nil
    end
end

function EditorViewClass:create( parentFrame )
    if not self.decorator then
        self.decorator = self:getSkin().EditorViewDecorator:new( self )
    end

    if self.frame then
        self.frame:SetParent( parentFrame )
    else
        self.frame = CreateFrame( "Frame", nil, parentFrame )
        self.contentFrame = CreateFrame( "Frame", nil, self.frame )

        self.frame:SetClipsChildren( true )

        self.frame:SetScript( "OnEnter", function() onEnter( self ) end )
        self.frame:SetScript( "OnLeave", function() onLeave( self ) end )

        self.decorator:apply()
    end

    update( self )
    zu.AddListener( self.model.btreeUpdatedEvent, self, update )

    self.frame:SetScript( "OnUpdate", function() onUpdate( self ) end )
end

function EditorViewClass:release()
    zu.Widget.release( self )
    self.frame:SetScript( "OnUpdate", nil )
end

function EditorViewClass:calcMinMax()
    update( self )
    self.minWidth = zu.Scale( self.model.charWidth * 10 )
    self.preferredWidth = zu.Scale( self.model.charWidth * 60 )
    self.minHeight = zu.Scale( self.model.lineHeight * 10 )
    self.preferredHeight = zu.Scale( self.model.lineHeight * 20 )
end

function EditorViewClass:show()
    zu.Widget.show( self )
    self:resetBlinking()
end

function EditorViewClass:hide()
    zu.Widget.hide( self )

    if self.blinker then
        self.blinker:Cancel()
        self.blinker = nil
    end
end

function EditorViewClass:setLevel( level )
    self.frame:SetFrameLevel( level )
end

function EditorViewClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )
    self.frame:SetSize( width, height )
    self.frame:SetPoint( "TOPLEFT", self.frame:GetParent(), "TOPLEFT", x, -y )
    update( self )
end

function EditorViewClass:setScrollXPosition( x )
    if not self.canScroll then
        return
    end

    self.scrollXPosition = x
    self.contentFrame:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", -x, self.YScrollOffset )
    zu.Raise( self.scrolledEvent )
end

function EditorViewClass:setScrollYPosition( y )
    if not self.canScroll then
        return
    end

    self.scrollYPosition = y
    update( self )
    zu.Raise( self.scrolledEvent )
end

function EditorViewClass:resetBlinking()
    if self.readOnly then
        self.showCursors = false
        zu.Raise( self.cursorsVisiblityEvent, false )
        return
    end

    if self.blinker then
        self.blinker:Cancel()
    end

    zu.Raise( self.cursorsVisiblityEvent, self.focus )
    self.showCursors = false

    local function blink()
        zu.Raise( self.cursorsVisiblityEvent, self.showCursors and self.focus )
        self.showCursors = not self.showCursors
    end

    self.blinker = C_Timer.NewTicker( 0.5, blink )
end

function EditorViewClass:setFocus( focus )
    if self.focus == focus then
        return
    end

    self.focus = focus
    if focus == true then
        self:resetBlinking()
    else
        zu.Raise( self.cursorsVisiblityEvent, false )
    end
end

function EditorViewClass:lineColToCoords( l, c )
    local x, y = self.model:lineColToCoords( l, c )

    x = x * self.charWidthScale
    y = y * self.lineHeightScale

    x = x - self.scrollXPosition
    y = y - self.scrollYPosition

    if x < 0 or y < 0 then
        return
    end

    if x > self.width or y > self.height then
        return
    end

    return x, y
end
