--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- Flow layout: lays out children horizontally,
-- and create additional rows if everything doesn't
-- fit in the given width.
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor
local sqrt = math.sqrt

local FlowLayoutClass = {}
zu.FlowLayoutClass = FlowLayoutClass
setmetatable( FlowLayoutClass, { __index = zu.GroupClass } )

-- Default values, can be overriden by the template or inside individual instances
FlowLayoutClass.numElements = 0

function FlowLayoutClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    o.tpl = tpl
    return o
end

function zu.FlowLayout( tpl )
    return FlowLayoutClass:new( tpl )
end

function FlowLayoutClass:create( parentFrame )
    if self.tpl then
        self.numElements = #self.tpl
    end

    self.hspacing = self.hspacing or self:getSkin().spacing
    self.vspacing = self.vspacing or self:getSkin().spacing

    for i=1, self.numElements, 1 do
        local v = self[i]
        v.parent = self
        v:create( parentFrame )
    end
end

local Row = {}
setmetatable( Row, { __index = zu.Widget } )

function Row:new()
    local o = {}

    o.cols = zu.Spreader:new(
        -- The cols directly contain widgets so we need
        -- to redirect all those accessors since each widget
        -- class may have overloaded them
        function( s, width ) s:setWidth( width ) end,
        function( s ) return s:getPreferredWidth() end,
        function( s ) return s:getMinWidth() end,
        function( s ) return s:getMaxWidth() end )

    o.cols.numItems = 0

    o.height = 0
    o.preferredHeight = 0
    o.minHeight = 0
    o.maxdHeight = 0

    setmetatable( o, { __index = self } )
    return o
end

function FlowLayoutClass:calcMinMax()
    if self.tpl then
        self.numElements = #self.tpl
    end

    -- Our min and max policy is as follows:

    -- Min width: we use the largest min width of our children,
    -- to ensure that everything fits when we are compressed horizontally
    -- into a single column

    -- Min height: we use the largest min height of our children,
    -- to ensure that everything fits when we are expanded horizontally
    -- into a single row

    -- Max width: the sum of the max width of all our children, unless
    -- we have at least one with unlimited width, in which case we also
    -- have an unlimited width

    -- preferred width: the square root of the sum of the preferred areas
    -- of all the widgets, to try and have a square-ish default shape
    -- preferred height: same as preferred width
    local minWidth = zu.Scale( self.hspacing * ( self.numElements - 1 ) )
    local minHeight = 0

    local maxWidth = minWidth
    local maxHeight = minHeight

    local levelRange = 0
    local totalArea = 0

    for i=1,self.numElements,1 do
        local v = self[i]
        v:calcMinMax()
        minWidth = max( minWidth, v.minWidth )
        minHeight = max( minHeight, v.minHeight )

        if maxWidth and v.maxWidth then
            maxWidth = maxWidth + v.maxWidth
        else
            maxWidth = nil
        end

        if maxHeight and v.maxHeight then
            maxHeight = maxHeight + v.maxHeight
        else
            maxHeight = nil
        end

        totalArea = totalArea + v.preferredWidth * v.preferredHeight
        levelRange = max( levelRange, v.levelRange )
    end

    if maxWidth then
        maxWidth = maxWidth + zu.Scale( self.hspacing )
    end

    self.minWidth = minWidth
    self.minHeight = minHeight
    self.maxWidth = maxWidth
    self.maxHeight = minHeight
    self.levelRange = levelRange

    if totalArea then
        self.preferredWidth = sqrt( totalArea )
    else
        self.preferredWidth = 0
    end
    self.preferredHeight = self.preferredWidth

    self.rows = zu.Spreader:new(
        Row.setHeight,
        Row.getPreferredHeight,
        Row.getMinHeight,
        Row.getMaxHeight )

    self.rows.numItems = 0
end

local function newRow( self, numRows )
    local rows = self.rows

    numRows = numRows + 1
    local r = rows[numRows]

    if r then
        r.height = 0
        r.preferredHeight = 0
        r.minHeight = 0
        r.maxHeight = 0
        r.cols.numItems = 0
    else
        rows[numRows] = Row:new()
    end

    return numRows
end

local function completeRow( self, numRows, width )
    local row = self.rows[numRows]
    local cols = row.cols

    row.preferredHeight = cols.numItems > 0 and row.preferredHeight / cols.numItems or 0
    cols:precompute()
    cols:computeItemsSizes( width - zu.Scale( self.hspacing * ( cols.numItems - 1 ) ) )

    -- Invoke heightForWidth for all the items in the row,
    -- update our minHeight if its smaller than any of the
    -- item's hfw.
    for i=1,cols.numItems,1 do
        local v = cols[i]
        v.hfw = v:heightForWidth( v.width )
        if v.hfw then
            row.minHeight = max( row.minHeight, v.hfw )
        end
    end
end

function FlowLayoutClass:heightForWidth( width )
    -- At this point we know the width that we'll have,
    -- so we can decide how many row we need and which widget
    -- will fit in each.
    --
    -- We then spread them out along their X axes, so we can
    -- invoke their own heightForWidth().
    local rows = self.rows
    local remainingWidth = nil
    local numRows = 0

    numRows = newRow( self, numRows )
    remainingWidth = width

    for i=1,self.numElements,1 do
        local v = self[i]
        if v.minWidth > remainingWidth then
            -- Row full, create a new one
            completeRow( self, numRows, width )
            numRows = newRow( self, numRows )
            remainingWidth = width
        end

        local row = rows[numRows]
        row.cols.numItems = row.cols.numItems + 1
        row.cols[row.cols.numItems] = v

        -- Update the row's preferred/min/max height
        row.preferredHeight = row.preferredHeight + v.preferredHeight
        row.minHeight = max( row.minHeight, v.minHeight )
        if row.maxHeight and v.maxHeight then
            row.maxHeight = min( row.maxHeight, v.maxHeight )
        else
            row.maxHeight = nil
        end

        remainingWidth = remainingWidth - ( v.minWidth + zu.Scale( self.hspacing ) )
    end

    completeRow( self, numRows, width )

    rows.numItems = numRows
    rows:precompute()

    local hfw = 0
    for i=1,rows.numItems,1 do
        hfw = hfw + rows[i].minHeight
    end

    hfw = hfw + zu.Scale( self.vspacing * ( numRows - 1 ) )
    self.hfw = hfw

    return hfw
end

function FlowLayoutClass:setLevel( level )
    zu.GroupClass.setLevel( self, level )
    for i=1,self.numElements,1 do
        local w = self[i]
        w:setLevel( level )
    end
end

function FlowLayoutClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )

    -- We actually did part of the work in heightForWidth.
    -- All we have to do now is to spread the rows vertically
    -- and recurse into the layout() method of each widget.
    self.rows:computeItemsSizes( height - zu.Scale( self.vspacing * ( self.rows.numItems - 1 ) ) )

    local cellY = y
    local spacingForRow = zu.Scale( self.hspacing )

    for i=1,self.rows.numItems,1 do
        local row = self.rows[i]
        local cellX = x
        local cellHeight = row.height
        local cols = row.cols

        local remainingSpace = width - row.cols.totalUsedSize

        for j=1,cols.numItems,1 do
            if self.justify == "justified" and cols.numItems > 1 then
                if i == self.rows.numItems then
                    spacingForRow = zu.Scale( self.hspacing )
                else
                    local numRemainingItems = cols.numItems - j
                    if numRemainingItems > 0 then
                        spacingForRow = floor( remainingSpace / numRemainingItems + 0.5 )
                        remainingSpace = remainingSpace - spacingForRow
                    end
                end
            end

            local widget = cols[j]
            local cellWidth = widget.width
            widget.width = nil

            -- Clamp the widget's size to its max size,
            -- Since we may have ended up with more than that,
            -- and we want to center it within that size.
            local w = cellWidth
            if not widget.horizontalFill and widget.maxWidth then
                w = min( w, widget.maxWidth )
            end

            local h = cellHeight
            if not widget.verticalFill and widget.maxHeight then
                h = min( h, widget.maxHeight )
            end

            -- Make sure to give a height of at least the hfw that was requested
            -- by that widget.
            if widget.hfw then
                h = max( h, widget.hfw )
            end

            -- Center the widget within its cell
            local wx = cellX + ( cellWidth - w ) / 2
            local wy = cellY + ( cellHeight - h ) / 2
            wx = floor( wx + 0.5 )
            wy = floor( wy + 0.5 )

            widget:layout( wx, wy, w, h )

            cellX = cellX + cellWidth + spacingForRow
        end

        cellY = cellY + row.height + zu.Scale( self.vspacing )
    end
end
