--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor

local GridLayoutClass = {}
zu.GridLayoutClass = GridLayoutClass
setmetatable( GridLayoutClass, { __index = zu.GroupClass } )

-- Default values, can be overriden by the template or inside individual instances
GridLayoutClass.numColumns = 0
GridLayoutClass.hjustify = "center"
GridLayoutClass.vjustify = "center"

function GridLayoutClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    o.tpl = tpl
    return o
end

function zu.GridLayout( tpl )
    return GridLayoutClass:new( tpl )
end

function GridLayoutClass:reuse( tpl )
    self.tpl = tpl
    zu.GroupClass.reuse( self, tpl )
end

-- HorizontalLayout is a convenience wrapper
-- that creates a grid with a single row.
function zu.HorizontalLayout( gl )
    local grid = GridLayoutClass:new( gl )
    grid.numColumns = 0
    return grid
end

-- VerticalLayout is a convenience wrapper
-- that creates a grid with one column.
function zu.VerticalLayout( gl )
    local grid = GridLayoutClass:new( gl )
    grid.numColumns = 1
    return grid
end

function GridLayoutClass:create( parentFrame )
    self.numElements = #self.tpl
    self.spacing = self.spacing or self:getSkin().spacing

    for i=1, self.numElements, 1 do
        local v = self[i]
        v.parent = self
        v:create( parentFrame )
    end
end

local Row = {}
setmetatable( Row, { __index = zu.Widget } )

function Row:new()
    local o = {}

    o.preferredHeight = 0
    o.minHeight = 0
    o.maxHeight = 0

    setmetatable( o, { __index = self } )
    return o
end

function Row:getMinHeight()
    if self.hfw then
        return max( self.minHeight, self.hfw )
    end

    return self.minHeight
end

function Row:getMaxHeight()
    if not self.maxHeight then
        return
    end

    if self.hfw then
        return max( self.maxHeight, self.hfw )
    end

    return self.maxHeight
end

local Col = {}
setmetatable( Col, { __index = zu.Widget } )

function Col:new()
    local o = {}

    o.preferredWidth = 0
    o.minWidth = 0
    o.maxWidth = 0

    setmetatable( o, { __index = self } )
    return o
end

function GridLayoutClass:calcMinMax()
    self.numElements = #self.tpl

    -- If the number of columns is 0, it means we want
    -- to put everything in a single row
    local numCols = self.numColumns
    if numCols == 0 then
        numCols = self.numElements
    end

    local numRows = self.numElements / numCols
    if ( self.numElements - 1 ) % numCols then
        numRows = numRows + 1
    end

    self.rows = zu.Spreader:new(
        Row.setHeight,
        Row.getPreferredHeight,
        Row.getMinHeight,
        Row.getMaxHeight )

    self.cols = zu.Spreader:new(
        Col.setWidth,
        Col.getPreferredWidth,
        Col.getMinWidth,
        Col.getMaxWidth )

    local rows = self.rows
    local cols = self.cols

    local irow = 1
    local icol = 1

    local numNonEmptyRows = 0
    local numNonEmptyCols = 0

    local levelRange = 0

    do
        local row = rows[1] or Row:new()
        local col = cols[1] or Col:new()

        for i=1,self.numElements,1 do
            local v = self[i]
            v:calcMinMax()

            rows[irow] = row
            cols[icol] = col

            row.preferredHeight = max( row.preferredHeight, v.preferredHeight )
            col.preferredWidth = max( col.preferredWidth, v.preferredWidth )

            row.minHeight = max( row.minHeight, v.minHeight )
            col.minWidth = max( col.minWidth, v.minWidth )

            levelRange = max( levelRange, v.levelRange )

            if row.maxHeight and v.maxHeight then
                row.maxHeight = row.maxHeight > 0 and min( row.maxHeight, v.maxHeight )
                    or v.maxHeight
            else
                row.maxHeight = nil
            end

            if col.maxWidth and v.maxWidth then
                if self.greedyMaxWidth then
                    col.maxWidth = col.maxWidth > 0 and max( col.maxWidth, v.maxWidth )
                    or v.maxWidth
                else
                    col.maxWidth = col.maxWidth > 0 and min( col.maxWidth, v.maxWidth )
                    or v.maxWidth
                end
            else
                col.maxWidth = nil
            end

            icol = icol + 1
            if icol > numCols then
                icol = 1
                irow = irow + 1
                row = rows[irow] or Row:new()
            end

            col = cols[icol] or Col:new()
        end
    end

    for i=1, #rows, 1 do
        local row = rows[i]
        local preferredHeight = row.preferredHeight

        -- Make sure that max sizes are greater or equal to min
        if row.maxHeight and row.maxHeight < row.minHeight then
            row.maxHeight = row.minHeight
        end

        -- Clamp preferred sizes.
        preferredHeight = max( preferredHeight, row.minHeight )

        if row.maxHeight then
            preferredHeight = min( preferredHeight, row.maxHeight )
        end

        row.preferredHeight = preferredHeight

        if row.maxHeight ~= 0 then
            numNonEmptyRows = numNonEmptyRows + 1
        end
    end

    for i=1, #cols, 1 do
        local col = cols[i]
        local preferredWidth = col.preferredWidth

        -- Make sure that max sizes are greater or equal to min
        if col.maxWidth and col.maxWidth < col.minWidth then
            col.maxWidth = col.minWidth
        end

        -- Clamp preferred sizes.
        preferredWidth = max( preferredWidth, col.minWidth )

        if col.maxWidth then
            preferredWidth = min( preferredWidth, col.maxWidth )
        end

        col.preferredWidth = preferredWidth

        if col.maxWidth ~= 0 then
            numNonEmptyCols = numNonEmptyCols + 1
        end
    end

    self.rows:precompute()
    self.cols:precompute()

    self.numNonEmptyRows = numNonEmptyRows
    self.numNonEmptyCols = numNonEmptyCols

    local verTotalSpacing = self.spacing * ( numNonEmptyRows - 1 )
    local horTotalSpacing = self.spacing * ( numNonEmptyCols - 1 )

    self.preferredHeight = self.rows.preferredSize + verTotalSpacing
    self.minHeight = self.rows.minSize + verTotalSpacing
    self.maxHeight = self.rows.maxSize
    if self.maxHeight then
        self.maxHeight = self.maxHeight + verTotalSpacing
    end

    self.preferredWidth = self.cols.preferredSize + horTotalSpacing
    self.minWidth = self.cols.minSize + horTotalSpacing
    self.maxWidth = self.cols.maxSize
    if self.maxWidth then
        self.maxWidth = self.maxWidth + horTotalSpacing
    end

    self.levelRange = levelRange
end

function GridLayoutClass:heightForWidth( width )
    -- Compute the columns widths for the requested width
    self.cols:computeItemsSizes( width - self.spacing * ( self.numNonEmptyCols - 1 ) )

    local rows = self.rows
    for i=1, #rows, 1 do
        rows[i].hfw = nil
    end

    local numCols = self.numColumns
    if numCols == 0 then
        numCols = self.numElements
    end

    local irow = 1
    local icol = 1
    local row = self.rows[1]
    local col = self.cols[1]

    -- If our content hfw behavior changed, it is possible that we had a hfw constraint before
    -- and don't have one anymore now, so clear it before recomputing it
    self.hfw = nil

    local recomputeRows = false

    for i=1,self.numElements,1 do
        local v = self[i]
        local hfw = v:heightForWidth( col.width )

        if hfw then
            if row.hfw then
                row.hfw = max( row.hfw, hfw )
            else
                row.hfw = hfw
            end

            row.minSize = row.hfw
            row.maxSize = row.hfw
            row.preferredSize = row.hfw
            recomputeRows = true
        end

        icol = icol + 1
        if icol > numCols then
            icol = 1
            irow = irow + 1
            row = self.rows[irow]
        end

        col = self.cols[icol]
    end

    if recomputeRows then
        self.rows:precompute()
        self.hfw = self.rows.minSize + self.spacing * ( self.numNonEmptyRows - 1 )
    end

    return self.hfw
end

function GridLayoutClass:setLevel( level )
    zu.GroupClass.setLevel( self, level )
    for i=1,self.numElements,1 do
        local w = self[i]
        w:setLevel( level )
    end
end

function GridLayoutClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )

    self.rows:computeItemsSizes( height - self.spacing * ( self.numNonEmptyRows - 1 ) )

    local cellX = x
    local cellY = y

    local numCols = self.numColumns
    if numCols == 0 then
        numCols = self.numElements
    end

    local irow = 1
    local icol = 1
    local row = self.rows[1]
    local col = self.cols[1]

    for i=1,self.numElements,1 do
        local widget = self[i]
        local cellWidth = col.width
        local cellHeight = row.height

        -- Clamp the widget's size to its max size,
        -- Since we may have ended up with more than that,
        -- and we want to center it within that size.
        local w = cellWidth
        if not widget.horizontalFill and widget.maxWidth then
            w = min( w, widget.maxWidth )
        end

        local h = cellHeight
        if not widget.verticalFill and widget.maxHeight then
            h = min( h, widget.maxHeight )
        end

        -- Make sure to give a height of at least the hfw that was requested
        -- by that .
        if widget.hfw then
            h = max( h, widget.hfw )
        end

        local wx = cellX
        local wy = cellY

        -- Center the widget within its cell
        if self.hjustify == "center" then
            wx = wx + ( cellWidth - w ) / 2
            wx = floor( wx + 0.5 )
        end

        if self.vjustify == "center" then
            wy = wy + ( cellHeight - h ) / 2
            wy = floor( wy + 0.5 )
        end

        widget:layout( wx, wy, w, h )

        cellX = cellX + col.width + ( col.width > 0 and self.spacing or 0 )

        icol = icol + 1
        if icol > numCols then
            cellX = x
            cellY = cellY + row.height +(  row.height > 0 and self.spacing or 0 )

            icol = 1
            irow = irow + 1
            row = self.rows[irow]
        end

        col = self.cols[icol]
    end
end
