--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- A layout widget that overlays multiple
-- widgets on top of one another.
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor

local OverlayLayoutClass = {}
zu.OverlayLayoutClass = OverlayLayoutClass
setmetatable( OverlayLayoutClass, { __index = zu.GroupClass } )

function OverlayLayoutClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    o.numElements = #tpl
    return o
end

function zu.OverlayLayout( tpl )
    return OverlayLayoutClass:new( tpl )
end

function OverlayLayoutClass:create( parentFrame )
    for i=1, self.numElements, 1 do
        local v = self[i]
        v.parent = self
        v:create( parentFrame )
    end
end

function OverlayLayoutClass:calcMinMax()
    local minWidth = 0
    local minHeight = 0
    local maxWidth
    local maxHeight
    local preferredWidth = 0
    local preferredHeight = 0
    local levelRange = 0

    for i=1, self.numElements, 1 do
        local w = self[i]
        w:calcMinMax()
        minWidth = max( minWidth, w.minWidth )
        minHeight = max( minHeight, w.minHeight )

        if w.maxWidth then
            maxWidth = maxWidth and min( maxWidth, w.maxWidth ) or w.maxWidth
        end

        if w.maxHeight then
            maxHeight = maxHeight and min( maxHeight, w.maxHeight ) or w.maxHeight
        end

        preferredWidth = max( preferredWidth, w.preferredWidth )
        preferredHeight = max( preferredHeight, w.preferredHeight )
        levelRange = levelRange + w.levelRange
    end

    self.minWidth = minWidth
    self.minHeight = minHeight
    self.maxWidth = maxWidth
    self.maxHeight = maxHeight
    self.preferredWidth = preferredWidth
    self.preferredHeight = preferredHeight
    self.levelRange = levelRange
end

function OverlayLayoutClass:heightForWidth( width )
    local hfw = nil

    for i=1, self.numElements, 1 do
        local w = self[i]
        local whfw = w:heightForWidth( width )

        if not hfw then
            hfw = whfw
        elseif whfw then
            hfw = max( hfw, whfw )
        end
    end

    self.hfw = hfw
    return hfw
end

function OverlayLayoutClass:setLevel( level )
    zu.GroupClass.setLevel( self, level )

    level = level + self.levelRange

    for i=1, self.numElements, 1 do
        local widget = self[i]
        level = level - widget.levelRange
        widget:setLevel( level )
    end
end

function OverlayLayoutClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )

    for i=1, self.numElements, 1 do
        local widget = self[i]

        local w = max( width, widget.minWidth )
        if not widget.horizontalFill then
            w = min( w, widget.maxWidth or w )
        end

        local h = max( height, widget.minHeight )
        if not widget.verticalFill then
            h = min( h, widget.maxHeight or h )
            h = max( h, widget.hfw or h )
        end

        local wx = x + floor( ( width - w ) / 2 + 0.5 )
        local wy = y + floor( ( height - h ) / 2 + 0.5 )

        widget:layout( wx, wy, w, h )
    end
end
