--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- A layout widget that displays only one of its children,
-- switchable to another at any time.
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max

local PagerLayoutClass = {}
zu.PagerLayoutClass = PagerLayoutClass
setmetatable( PagerLayoutClass, { __index = zu.GroupClass } )

PagerLayoutClass.currentlyDisplayedIndex = 1

function PagerLayoutClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    o.numElements = #tpl
    return o
end

function zu.Pager( tpl )
    return PagerLayoutClass:new( tpl )
end

function PagerLayoutClass:create( parentFrame )
    for i=1, self.numElements, 1 do
        local v = self[i]
        v.parent = self
        v:create( parentFrame )
    end
end

function PagerLayoutClass:show()
    self[self.currentlyDisplayedIndex]:show()
end

function PagerLayoutClass:hide()
    self[self.currentlyDisplayedIndex]:hide()
end

function PagerLayoutClass:calcMinMax()
    local minWidth = 0
    local minHeight = 0
    local maxWidth
    local maxHeight
    local preferredWidth = 0
    local preferredHeight = 0
    local levelRange = 0

    for i=1, self.numElements, 1 do
        local w = self[i]
        w:calcMinMax()
        minWidth = max( minWidth, w.minWidth )
        minHeight = max( minHeight, w.minHeight )

        if w.maxWidth then
            maxWidth = maxWidth and min( maxWidth, w.maxWidth ) or w.maxWidth
        end

        if w.maxHeight then
            maxHeight = maxHeight and min( maxHeight, w.maxHeight ) or w.maxHeight
        end

        preferredWidth = max( preferredWidth, w.preferredWidth )
        preferredHeight = max( preferredHeight, w.preferredHeight )

        levelRange = max( levelRange, w.levelRange )
    end

    if maxWidth then
        preferredWidth = min( preferredWidth, maxWidth )
    end

    if maxHeight then
        preferredHeight = min( preferredHeight, maxHeight )
    end

    self.minWidth = minWidth
    self.minHeight = minHeight
    self.maxWidth = maxWidth
    self.maxHeight = maxHeight
    self.preferredWidth = preferredWidth
    self.preferredHeight = preferredHeight
    self.levelRange = levelRange
end

function PagerLayoutClass:heightForWidth( width )
    local currentWidget = self[self.currentlyDisplayedIndex]
    local hfw = currentWidget:heightForWidth( width )
    self.hfw = hfw
    return hfw
end

function PagerLayoutClass:setLevel( level )
    zu.GroupClass.setLevel( self, level )
    local currentWidget = self[self.currentlyDisplayedIndex]
    currentWidget:setLevel(level )
end

function PagerLayoutClass:layout( x, y, width, height, level )
    local currentWidget = self[self.currentlyDisplayedIndex]
    currentWidget:layout( x, y, width, height, level )
    zu.GroupClass.layout( self, x, y, width, height, level )
end

function PagerLayoutClass:switch( newWidgetIndex )
    if self.currentlyDisplayedIndex == newWidgetIndex then
        return
    end

    local currentWidget = self[self.currentlyDisplayedIndex]
    currentWidget:hide()

    self.currentlyDisplayedIndex = newWidgetIndex

    currentWidget = self[newWidgetIndex]
    currentWidget:show()

    self:relayout()
    self:setLevel( self.level )
end
