--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor

local ResizeGroupClass = {}
zu.ResizeGroupClass = ResizeGroupClass
setmetatable( ResizeGroupClass, { __index = zu.GroupClass } )

ResizeGroupClass.vertical = false

function ResizeGroupClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    o.numElements = #tpl
    o.sashes = {}
    return o
end

function zu.HorResizeGroup( tpl )
    return ResizeGroupClass:new( tpl )
end

function zu.VerResizeGroup( tpl )
    local rg = ResizeGroupClass:new( tpl )
    rg.vertical = true
    return rg
end

local function GetCurrentHeight( self )
    if not self.majorSize then
        return self:getPreferredHeight()
    end

    local size = max( self.minHeight, self.majorSize )
    local maxSize = self.maxHeight
    if maxSize then
        size = min( self.maxHeight, size )
    end

    return size
end

local function GetCurrentWidth( self )
    if not self.majorSize then
        return self:getPreferredWidth()
    end

    local size = max( self.minWidth, self.majorSize )
    local maxSize = self.maxWidth
    if maxSize then
        size = min( self.maxWidth, size )
    end

    return size
end

local function setWidgetMajorSize( self, s )
    if self.resizeable then
        self.majorSize = s
    end
end

local function wrapWidget( self, w )
    local wrapped = { wrapped = w }

    if self.vertical then
        wrapped.getCurrentSize = GetCurrentHeight
    else
        wrapped.getCurrentSize = GetCurrentWidth
    end

    setmetatable( wrapped, { __index = w } )
    return wrapped
end

-- Remove the given amount of pixels from widget's preferred sizes (the preferred size defined by the wrappers,
-- not their actual preferred sizes which are only used as default sizes), starting at the provided index,
-- in the given direction.
-- This will only move to the next widget if the current one can't be shrunk further, until the total amount
-- of pixels requested have been removed or we reach the last widget.
local function ShrinkWidgets( self, index, direction, amount )
    local getMinFunc = self.widgets.getMinFunc
    local numWidgets = self.numElements

    while index > 0 and index <= numWidgets and amount > 0 do
        local w = self.widgets[index]
        local oldSize = w:getCurrentSize()
        if oldSize then
            local newSize = max( oldSize - amount, getMinFunc( w ) )

            if w.resizeable then
                w.majorSize = newSize
            end
            amount = amount - oldSize + newSize
        end

        index = index + direction
    end

    return amount
end

-- Add the given amount of pixels to widget's preferred sizes (the preferred size defined by the wrappers,
-- not their actual preferred sizes which are only used as default sizes), starting at the provided index,
-- in the given direction.
-- This will only move to the next widget if the current one can't be expanded further, until the total amount
-- of pixels requested have been added or we reach the last widget.
local function ExpandWidgets( self, index, direction, amount )
    local getMaxFunc = self.widgets.getMaxFunc
    local numWidgets = self.numElements

    while index > 0 and index <= numWidgets and amount > 0 do
        local w = self.widgets[index]
        local oldSize = w:getCurrentSize()
        local newSize = oldSize + amount

        local maxSize = getMaxFunc( w )
        if maxSize then
            newSize = min( newSize, maxSize )
        end

        if w.resizeable then
            w.majorSize = newSize
        end
        amount = amount - newSize + oldSize

        index = index + direction
    end

    return amount
end

local function SashOnEnter( sash )
    if sash.onHover then
        sash:onHover( true )
    end
end

local function SashOnLeave( sash )
    if not sash.dragging and sash.onHover then
        sash:onHover( false )
    end
end

local function SashOnUpdate( sash )
    if sash.owner.frozen then
        return
    end

    -- Compute the new wanted position for this sash
    local newX, newY = sash.owner:getRelativeMousePosition()
    newX = newX - sash.mOffsetX
    newY = newY - sash.mOffsetY

    local delta

    if sash.owner.vertical then
        delta = newY - sash.y
    else
        delta = newX - sash.x
    end

    if delta > 0 then
        local remain = ShrinkWidgets( sash.owner, sash.index + 1, 1, delta )
        ExpandWidgets( sash.owner, sash.index, -1, delta - remain )
        zu.Widget.relayout( sash.owner )
    elseif delta < 0 then
        local remain = ShrinkWidgets( sash.owner, sash.index, -1, -delta )
        ExpandWidgets( sash.owner, sash.index + 1, 1, -delta - remain )
        zu.Widget.relayout( sash.owner )
    end
end

local function SashOnMouseDown( sash )
    sash.dragging = true
    sash.owner:getWindow():freeze()
    local x, y = sash.owner:getRelativeMousePosition()
    sash.mOffsetX = x - sash.x
    sash.mOffsetY = y - sash.y
    sash:SetScript( "OnUpdate", SashOnUpdate )
end

local function SashOnMouseUp( sash )
    sash.dragging = false
    sash:SetScript( "OnUpdate", nil )
    sash.owner:getWindow():unfreeze()

    if sash.onHover and not sash:IsMouseOver() then
        sash:onHover( false )
    end
end

function ResizeGroupClass:create( parentFrame )
    self.sashSize = self:getSkin().SashDecorator.SashSize()
    self.spacing = 2

    self.parentFrame = parentFrame

    self.settings = {}
    if self.name then
        self.settings = self:getWindow().settings[self.name] or {}
        self:getWindow().settings[self.name] = self.settings
    end

    if self.vertical then
        self.widgets = zu.Spreader:new(
            setWidgetMajorSize,
            GetCurrentHeight,
            zu.Widget.getMinHeight,
            zu.Widget.getMaxHeight )
    else
        self.widgets = zu.Spreader:new(
            setWidgetMajorSize,
            GetCurrentWidth,
            zu.Widget.getMinWidth,
            zu.Widget.getMaxWidth )
    end

    for i=1, self.numElements, 1 do
        local v = self[i]
        v.parent = self
        v:create( parentFrame )

        local w = wrapWidget( self, v )
        self.widgets[i] = w

        w.majorSize = self.settings[i]
    end
end

function ResizeGroupClass:calcMinMax()
    local getMajorMinSize
    local getMajorMaxSize
    local getMajorPreferredSize

    local getMinorMinSize
    local getMinorMaxSize
    local getMinorPreferredSize

    if self.vertical then
        getMajorMinSize = zu.Widget.getMinHeight
        getMajorMaxSize = zu.Widget.getMaxHeight
        getMajorPreferredSize = zu.Widget.getPreferredHeight

        getMinorMinSize = zu.Widget.getMinWidth
        getMinorMaxSize = zu.Widget.getMaxWidth
        getMinorPreferredSize = zu.Widget.getPreferredWidth
    else
        getMajorMinSize = zu.Widget.getMinWidth
        getMajorMaxSize = zu.Widget.getMaxWidth
        getMajorPreferredSize = zu.Widget.getPreferredWidth

        getMinorMinSize = zu.Widget.getMinHeight
        getMinorMaxSize = zu.Widget.getMaxHeight
        getMinorPreferredSize = zu.Widget.getPreferredHeight
    end

    local minMajorSize = 0
    local maxMajorSize = 0
    local preferredMajorSize = 0

    local minMinorSize = 0
    local maxMinorSize = -1
    local preferredMinorSize = 0

    local numResizeableWidgets = 0

    local levelRange = 0

    for i=1, self.numElements, 1 do
        local w = self.widgets[i]
        w.needSash = false
        w.wrapped:calcMinMax()

        local wMinSize = getMajorMinSize( w )
        local wMaxSize = getMajorMaxSize( w )

        levelRange = max( levelRange, w.levelRange )

        minMajorSize = minMajorSize + getMajorMinSize( w )

        w.resizeable = false

        if wMinSize > 0 and wMinSize ~= wMaxSize then
            numResizeableWidgets = numResizeableWidgets + 1
            w.resizeable = true
        end

        if maxMajorSize then
            maxMajorSize = wMaxSize and ( maxMajorSize + wMaxSize ) or nil
        end

        preferredMajorSize = preferredMajorSize + getMajorPreferredSize( w )

        minMinorSize = max( minMinorSize, getMinorMinSize( w ) )

        if maxMinorSize then
            wMaxSize = getMinorMaxSize( w )

            if maxMinorSize < 0 then
                maxMinorSize = wMaxSize
            else
                maxMinorSize = wMaxSize and min( maxMinorSize, wMaxSize ) or nil
            end
        end

        preferredMinorSize = max( preferredMinorSize, getMinorPreferredSize( w ) )
    end

    -- Determine where to insert sashes, given that the first sash can only
    -- appear after the first resizeable widget, and the last one before the last
    -- resizeable one
    local numSashes = 0
    local remainingResizeable = numResizeableWidgets
    for i=1, self.numElements, 1 do
        local w = self.widgets[i]

        if w.resizeable then
            remainingResizeable = remainingResizeable - 1
        end

        -- If there are no more resizeable widgets after the current one, we're done
        if remainingResizeable == 0 then
            break
        end

        -- If the current widget is resizeable, it needs to be followed by a sash
        if w.resizeable then
            numSashes = numSashes + 1
            w.needSash = true
        else
            -- If the current widget is not resizeable but was preceded by at least one resizeable one
            -- and the next one is resizeable, it needs a sash too
            if numSashes > 0 and i < self.numElements and self.widgets[i+1].resizeable then
                numSashes = numSashes + 1
                w.needSash = true
            end
        end
    end

    local spacing = zu.Scale( self.spacing )
    local totalSpacing = spacing * ( self.numElements - 1 ) + ( zu.Scale( self.sashSize ) + spacing ) * numSashes
    minMajorSize = minMajorSize + totalSpacing
    maxMajorSize = maxMajorSize and maxMajorSize + totalSpacing or nil
    preferredMajorSize = preferredMajorSize + totalSpacing

    if self.vertical then
        self.minHeight = minMajorSize
        self.maxHeight = maxMajorSize
        self.preferredHeight = preferredMajorSize

        self.minWidth = minMinorSize
        self.maxWidth = maxMinorSize
        self.preferredWidth = preferredMinorSize
    else
        self.minWidth = minMajorSize
        self.maxWidth = maxMajorSize
        self.preferredWidth = preferredMajorSize

        self.minHeight = minMinorSize
        self.maxHeight = maxMinorSize
        self.preferredHeight = preferredMinorSize
    end

    self.totalSpacing = totalSpacing
    self.levelRange = levelRange
end

local function horizHFW( self, width )
    self.widgets:precompute()

    -- Compute the widths for the requested width
    self.widgets:computeItemsSizes( width - self.totalSpacing )

    local hfw

    for i=1, self.numElements, 1 do
        local w = self.widgets[i]
        local whfw = w.wrapped:heightForWidth( w.majorSize )

        if whfw then
            hfw = hfw and max( hfw, whfw ) or whfw
        end
    end

    self.hfw = hfw
    return hfw
end

local function vertHFW( self, width )
    local haveHFW = false

    for i=1, self.numElements, 1 do
        local w = self.widgets[i].wrapped
        local whfw = w:heightForWidth( width )

        if whfw then
            haveHFW = true

            w.minHeight = whfw
            w.maxHeight = whfw
            w.preferredHeight = whfw
        end
    end

    self.widgets:precompute()

    self.hfw = haveHFW and ( self.widgets.minSize + self.totalSpacing ) or nil
    return self.hfw
end

function ResizeGroupClass:heightForWidth( width )
    -- If we are recomputing hfw as part of an external relayouting request,
    -- we need to fix up the widget sizes in case a previously non resizeable widget becomes
    -- resizeable again: the total prefered widget size must match the total pixel size,
    -- or that uncollapsed widget won't get its original size back.
    if self.relayoutRequested then
        self.relayoutRequested = false

        local spacing = zu.Scale( self.spacing )
        local sashSize = zu.Scale( self.sashSize )

        for i=1, self.numElements, 1 do
            local w = self.widgets[i]

            local wsize = self.vertical and w.height or w.width
            if wsize ~= nil then
                local newsize = w:getCurrentSize()
                if newsize > wsize then
                    local delta = newsize - wsize

                    -- Also take into account sash addition/removal
                    -- to fix up the other widget's sizes
                    if w.needSash ~= w.haveSash then
                        local size = sashSize + spacing
                        delta = delta + ( w.needSash and size or -size )
                    end

                    delta = ShrinkWidgets( self, i + 1, 1, delta )
                    ShrinkWidgets( self, i - 1, -1, delta )
                end
            end
        end
    end

    if self.vertical then
        return vertHFW( self, width )
    else
        return horizHFW( self, width )
    end
end

function ResizeGroupClass:relayout()
    self.relayoutRequested = true
    zu.Widget.relayout( self )
end

function ResizeGroupClass:setLevel( level )
    zu.GroupClass.setLevel( self, level )

    for i=1,self.numElements,1 do
        local w = self[i]
        w:setLevel( level )

        if self.sashes[i] then
            self.sashes[i]:SetFrameLevel( level )
        end
    end
end

function ResizeGroupClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )

    local spacing = zu.Scale( self.spacing )
    local sashSize = zu.Scale( self.sashSize )

    local majorSize = self.vertical and height or width
    self.widgets:computeItemsSizes( majorSize - self.totalSpacing )

    local cellX = x
    local cellY = y

    for i=1, self.numElements, 1 do
        local widget = self.widgets[i]
        local cellWidth
        local cellHeight

        if self.vertical then
            cellWidth = width
            cellHeight = widget:getCurrentSize()
        else
            cellWidth = widget:getCurrentSize()
            cellHeight = height
        end

        -- Clamp the widget's size to its max size,
        -- Since we may have ended up with more than that,
        -- and we want to center it within that size.
        local w = cellWidth
        if widget.maxWidth then
            w = min( w, widget.maxWidth )
        end

        local h = cellHeight
        if widget.maxHeight then
            h = min( h, widget.maxHeight )
        end

        -- Make sure to give a height of at least the hfw that was requested.
        if widget.hfw then
            h = max( h, widget.hfw )
        end

        -- Center the widget within its cell
        local wx = cellX + ( cellWidth - w ) / 2
        local wy = cellY + ( cellHeight - h ) / 2
        wx = floor( wx + 0.5 )
        wy = floor( wy + 0.5 )

        widget.wrapped:layout( wx, wy, w, h )

        -- If the widget is not currently resizeable, don't overwrite its
        -- size setting: it may be a currently collapsed widget and we don't
        -- want to lose its previously set size.
        if widget.resizeable then
            self.settings[i] = widget:getCurrentSize()
        end

        -- The widgets preferred size along the major axis are always their current size,
        -- since this is what we manipulate to resize them, it's easier to keep it this way.
        if self.vertical then
            widget.preferredWidth = w
        else
            widget.preferredHeight = h
        end

        if self.vertical then
            cellY = cellY + cellHeight + spacing
        else
            cellX = cellX + cellWidth + spacing
        end

        -- Create a sash if one is needed after that widget, as determined during calcMinMax
        if widget.needSash then
            local sash = self.sashes[i] or zu.CreateFrame( "sash", "Frame", self.parentFrame )
            sash:Show()
            self.sashes[i] = sash

            sash.owner = self
            sash.index = i

            sash.vertical = not self.vertical

            if not sash.inited then
                sash.decorator = self:getSkin().SashDecorator:new( sash )
                sash:SetScript( "OnEnter", SashOnEnter )
                sash:SetScript( "OnLeave", SashOnLeave )
                sash:SetScript( "OnMouseDown", SashOnMouseDown )
                sash:SetScript( "OnMouseUp", SashOnMouseUp )

                sash.inited = true
            end

            sash.decorator:layout()

            if self.vertical then
                sash.x = cellX
                sash.y = cellY

                sash:SetSize( cellWidth, sashSize )
                cellY = cellY + sashSize + spacing
            else
                sash.x = cellX
                sash.y = cellY

                sash:SetSize( sashSize, cellHeight )
                cellX = cellX + sashSize + spacing
            end

            sash:SetPoint( "TOPLEFT", self.parentFrame, "TOPLEFT", sash.x, -sash.y )
            if self.level then
                sash:SetFrameLevel( self.level )
            end
        else
            -- If a sash was previously created at this index, destroy it
            -- (sashes may appear or disappear because of collapsible widgets, which
            -- may switch between being resizeable or not depending on whether their
            -- content is displayed)
            if self.sashes[i] then
                zu.ReleaseFrame( "sash", self.sashes[i] )
                self.sashes[i] = nil
            end
        end

        widget.haveSash = widget.needSash
    end
end

function ResizeGroupClass:getRelativeMousePosition()
    local mouseX, mouseY = GetCursorPosition()
    local scale = self.parentFrame:GetEffectiveScale()

    local x, y, _, h = self.parentFrame:GetScaledRect()

    x = floor( ( mouseX - x ) / scale )
    y = floor( ( h + y - mouseY ) / scale )

    x = max( x, 0 )
    y = max( y, 0 )
    x = min( x, self.width )
    y = min( y, self.height )

    return x, y
end
