--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- This helper spreads item accross
-- one axis over the specified amount of
-- space, taking their respective min/max/preferred
-- sizes along that axis, and assigning them
-- a size along that axis.
--
-- This is used as a building block for
-- layout classes.
--
-- We assume that the provided space is within
-- the total min/max range of the item to lay out.
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local max = math.max
local floor = math.floor
local ceil = math.ceil

local sort = table.sort

local Spreader = {}
zu.Spreader = Spreader

function Spreader:new( setSizeFunc, getPreferredFunc, getMinFunc, getMaxFunc )
    local o = {}

    o.getPreferredFunc = getPreferredFunc
    o.getMinFunc = getMinFunc
    o.getMaxFunc = getMaxFunc
    o.setSizeFunc = setSizeFunc
    o.sizeReqs = {}
    o.itemStates = {}
    o.itemsByMinSize = {}
    o.itemsByMaxSize = {}

    setmetatable( o, { __index = self } )
    return o
end

-- We have to define these here because if we define them inline, lua allocates new copies
-- of these functions everytime we call table.sort, and since this is part of the layout
-- process called constantly during resizing, we can't afford leaking all that memory
-- for nothing.
-- What a fucking garbage language. Love to get runtime performance penalties for writing
-- idiomatic code.
local minSizeSortFunc = function( a, b ) return a[2] > b[2] end
local maxSizeSortFunc = function( a, b ) return a[2] < b[2] end

-- precompute should be called when item are
-- added/removed, or when their min/max/preferred sizes
-- change.
function Spreader:precompute()
    self.numItems = self.numItems or #self

    local prefFunc = self.getPreferredFunc
    local minFunc = self.getMinFunc
    local maxFunc = self.getMaxFunc

    local sizeReqs = self.sizeReqs

    local preferredSize = 0
    local minSize = 0
    local maxSize = 0

    -- Copy the size reqs into a fast indexed array,
    -- and compute our own min/max/preferred size.
    for i=1,self.numItems,1 do
        local v = self[i]

        if sizeReqs[i] then
            sizeReqs[i][1] = prefFunc(v)
            sizeReqs[i][2] = minFunc(v)
            sizeReqs[i][3] = maxFunc(v)
        else
            sizeReqs[i] = { prefFunc(v), minFunc(v), maxFunc(v) }
        end

        sizeReqs[i][2] = ceil( sizeReqs[i][2] )
        if sizeReqs[i][3] then
            sizeReqs[i][3] = max( sizeReqs[i][2], floor( sizeReqs[i][3] ) )
        end

        sizeReqs[i][1] = max( sizeReqs[i][2], floor( sizeReqs[i][1] + 0.5 ) )

        preferredSize = preferredSize + sizeReqs[i][1]
        minSize = minSize + sizeReqs[i][2]

        if maxSize and sizeReqs[i][3] then
            maxSize = maxSize + sizeReqs[i][3]
        else
            maxSize = nil
        end
    end

    self.preferredSize = preferredSize
    self.minSize = minSize
    self.maxSize = maxSize

    -- Build arrays storing at which availableFreeSize each item will hit
    -- its own min or max size.
    --
    -- Note: the size we assign to each item for a given
    -- available size is proportional
    -- to their preferred size.
    local itemsByMinSize = self.itemsByMinSize
    local itemsByMaxSize = self.itemsByMaxSize

    local minSizeItemCount = 0
    local maxSizeItemCount = 0

    for i=1,self.numItems,1 do
        local reqs = sizeReqs[i]

        local minAvailSize = reqs[1] > 0 and reqs[2] * preferredSize / reqs[1] or 0
        minSizeItemCount = minSizeItemCount + 1

        if itemsByMinSize[minSizeItemCount] then
            itemsByMinSize[minSizeItemCount][1] = i
            itemsByMinSize[minSizeItemCount][2] = minAvailSize
        else
            itemsByMinSize[minSizeItemCount] = { i, minAvailSize }
        end

        if reqs[3] then
            local maxAvailSize = reqs[1] > 0 and reqs[3] * preferredSize / reqs[1] or 0
            maxSizeItemCount = maxSizeItemCount + 1
            if itemsByMaxSize[maxSizeItemCount] then
                itemsByMaxSize[maxSizeItemCount][1] = i
                itemsByMaxSize[maxSizeItemCount][2] = maxAvailSize
            else
                itemsByMaxSize[maxSizeItemCount] = { i, maxAvailSize }
            end
        end
    end

    -- Remove extraneous data left over from previous state
    for i = minSizeItemCount + 1, #itemsByMinSize, 1 do
        itemsByMinSize[i] = nil
    end

    for i = maxSizeItemCount + 1, #itemsByMaxSize, 1 do
        itemsByMaxSize[i] = nil
    end

    -- Sort the arrays
   sort( itemsByMinSize, minSizeSortFunc )
   sort( itemsByMaxSize, maxSizeSortFunc )
end

-- Determine the sizes and the positions of the items
-- along the relevant axis for the given total available size.
--
-- Returns the total size (which may be less than the provided size
-- if all the items have hit their max size)
function Spreader:computeItemsSizes( availableSize )
    -- For each item, track whether we are going to set them to
    -- their free size (preferred size proportional to the total free size),
    -- min size or max size.
    local itemStateFree = 1
    local itemStateMin = 2
    local itemStateMax = 3

    local itemStates = self.itemStates
    local sizeReqs = self.sizeReqs

    for i=1,self.numItems,1 do
        itemStates[i] = itemStateFree
    end

    -- Remaining size available for items locked to
    -- their min or max size yet.
    local remainingFreeSize = availableSize

    -- The sum of the preferred size of the currently free items
    local totalPreferredSize = self.preferredSize

    local lastMinSizeItem = nil

    -- First pass: find items that have to be locked to their min
    -- size because their free size would be too small
    local itemsByMinSize = self.itemsByMinSize
    for sortedIndex=1, #itemsByMinSize, 1 do
        local v = itemsByMinSize[sortedIndex]
        local i = v[1]
        local reqs = sizeReqs[i]

        local freeSize = totalPreferredSize > 0 and reqs[1] * remainingFreeSize / totalPreferredSize or 0
        freeSize = floor( freeSize + 0.5 )

        -- Because itemsByMinSize is sorted by decreasing order of
        -- the availableFreeSize at which they hit their min,
        -- it means that if we encounter an item not hitting
        -- its min, none of the following items will
        if freeSize > reqs[2] then
            break
        end

        -- Set the item to min state, update things accordingly.
        itemStates[i] = itemStateMin
        remainingFreeSize = remainingFreeSize - reqs[2]
        totalPreferredSize = totalPreferredSize - reqs[1]
        lastMinSizeItem = sortedIndex
    end

    -- Second pass: find items that have to be locked to their max
    -- size because their free size would be too big
    local itemsByMaxSize = self.itemsByMaxSize
    for j=1, #itemsByMaxSize, 1 do
        local v = itemsByMaxSize[j]
        local i = v[1]

        -- Skip items already set to their min size.
        if itemStates[i] ~= itemStateMin then
            local reqs = sizeReqs[i]

            local freeSize = totalPreferredSize > 0 and reqs[1] * remainingFreeSize / totalPreferredSize or 0
            freeSize = floor( freeSize + 0.5 )

            -- Because itemsByMaxSize is sorted by increasing order of
            -- the availableFreeSize at which they hit their max,
            -- it means that if we encounter an item not hitting
            -- its max, none of the following items will
            if freeSize < reqs[3] then
                break
            end

            -- Set the item to max state, update things accordingly.
            itemStates[i] = itemStateMax
            remainingFreeSize = remainingFreeSize - reqs[3]
            totalPreferredSize = totalPreferredSize - reqs[1]
        end
    end

    -- Third pass: since the second pass may have left us with additional room,
    -- iterate backwards from the last item set to its min state and free
    -- any of them that may now receive a free size bigger than their min.
    --
    -- If the newly freed item hits its max instead, set it to its max state.
    --
    -- Note that it means that some items may end up actually getting a larger size than their
    -- max size (case of a free item that would end up with more size than its max after a subsequent item has
    -- been locked to its max) which is why the max size isn't a strictly enforced requirement.
    -- Correctly enforcing min size is more important.
    if lastMinSizeItem then
        for sortedI = lastMinSizeItem, 1, -1 do
            local i = self.itemsByMinSize[sortedI][1]
            local reqs = sizeReqs[i]
            local den = ( totalPreferredSize + reqs[1] )
            local wouldBeFreeSize = den > 0 and reqs[1] * ( remainingFreeSize + reqs[2] )
                / den or 0
            wouldBeFreeSize = floor( wouldBeFreeSize + 0.5 )

            if wouldBeFreeSize < reqs[2] then
                -- If we're not unlocking this item, we're not
                -- unlocking any of the previous ones in the sorted array
                break
            end

            remainingFreeSize = remainingFreeSize + reqs[2]
            totalPreferredSize = totalPreferredSize + reqs[1]

            -- Check if we have to lock it at max state instead
            local freeSize = totalPreferredSize > 0 and reqs[1] * remainingFreeSize / totalPreferredSize or 0
            freeSize = floor( freeSize + 0.5 )
            if reqs[3] and freeSize > reqs[3] then
                itemStates[i] = itemStateMax
                remainingFreeSize = remainingFreeSize - reqs[3]
                totalPreferredSize = totalPreferredSize - reqs[1]
            else
                itemStates[i] = itemStateFree
            end
        end
    end

    -- Final pass: assign each item its final size
    -- according to its state.
    local totalUsedSize = 0
    local setSizeFunc = self.setSizeFunc

    for i=1,self.numItems,1 do
        local reqs = sizeReqs[i]
        local size = nil

        if itemStates[i] == itemStateFree then
            size = totalPreferredSize > 0 and reqs[1] * remainingFreeSize / totalPreferredSize or 0
            size = floor( size + 0.5 )

            -- Subtract the rounded size we actually used
            -- from the total, to avoid accumulating rounding
            -- errors and ending up off by a pixel
            remainingFreeSize = remainingFreeSize - size
            totalPreferredSize = totalPreferredSize - reqs[1]
        elseif itemStates[i] == itemStateMin then
            size = reqs[2]
        else
            size = reqs[3]
        end

        setSizeFunc( self[i], size )
        totalUsedSize = totalUsedSize + size
    end

    self.totalUsedSize = totalUsedSize
end
