--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local floor = math.floor

local ListSelectionController = {}
zu.ListSelectionController = ListSelectionController

function ListSelectionController:new()
    local o = {}
    o.selectionChangedEvent = {}
    o.highlightChangedEvent = {}
    o.scrollUpdateEvent = {}
    o.updateEvent = {}
    o.selectedItemIndex = 0
    o.highlightedItemIndex = 0
    setmetatable( o, { __index = self } )
    return o
end

local function scrollToSelection( self )
    if self.selectedItemIndex == 0 then
        return
    end

    local elemYPos = ( self.selectedItemIndex - 1 ) * self.view.lineHeight
    local relElemYPos = elemYPos - self.view.scrollPosition

    if relElemYPos < 0 then
        self.view:setScrollPosition( elemYPos )
        zu.Raise( self.scrollUpdateEvent )
    elseif ( relElemYPos + self.view.lineHeight ) > self.view.height then
        self.view:setScrollPosition( elemYPos - self.view.height + self.view.lineHeight)
        zu.Raise( self.scrollUpdateEvent )
    end
end

local function scrollToHighlighted( self )
    if self.highlightedItemIndex == 0 then
        return
    end

    local elemYPos = ( self.highlightedItemIndex - 1 ) * self.view.lineHeight
    local relElemYPos = elemYPos - self.view.scrollPosition

    if relElemYPos < 0 then
        self.view:setScrollPosition( elemYPos )
        zu.Raise( self.scrollUpdateEvent )
    elseif ( relElemYPos + self.view.lineHeight ) > self.view.height then
        self.view:setScrollPosition( elemYPos - self.view.height + self.view.lineHeight)
        zu.Raise( self.scrollUpdateEvent )
    end
end

-- Update the current selected element index on model change event
-- To correctly track the selected element when sorting the model,
-- for instance
local function update( self )
    self:selectElement( self.model:getElementIndex( self.selectedItem ) )
    zu.Raise( self.updateEvent )
end

-- In case the model was reset (aka all content was replaced and whatever
-- existed before may not exist anymore), we want to preserve the currently
-- highlighted item by row number (for the sake of not breaking keyboard
-- navigation every time we update the list during completion)
local function modelReset( self )
    self:selectElement( 0 )

    self.highlightedItem = self.model:getElement( self.highlightedItemIndex )
    if not self.highlightedItem then
        self.highlightedItemIndex = 0
    end

    zu.Raise( self.highlightChangedEvent )
    zu.Raise( self.updateEvent )
end

function ListSelectionController:init( view, model )
    self.view = view
    self.model = model
    self.frame = view.frame

    self.frame:RegisterEvent( "GLOBAL_MOUSE_UP" )
    self.frame:SetScript( "OnEvent", function( _, e )
        self:onMouseUp()
    end )

    self.frame:SetScript( "OnMouseDown", function() self:onMouseDown() end )
    self.frame:SetScript( "OnMouseUp", function() self:onMouseUp() end )

    self.frame:SetScript( "OnEnter", function()
        self.frame:SetScript( "OnUpdate", function() self:hoverUpdate() end )
        self:hoverUpdate()
    end )

    self.frame:SetScript( "OnLeave", function()
        self.frame:SetScript( "OnUpdate", nil )
        if not self.stickyHiglight then
            self.highlightedItemIndex = 0
            self.highlightedItem = nil
            zu.Raise( self.highlightChangedEvent )
        end

        if self.mouseHoverUpdate then
            self:mouseHoverUpdate( false )
        end
    end )

    if self.enableKeyboardNavigation then
        local function onKeyUp()
            -- Cancel the key repetition timer/ticker, if any.
            if self.repeater then
                self.repeater:Cancel()
                self.repeater = nil
                self.repeatingFunc = nil
            end
        end
        self.frame:SetScript( "OnKeyUp", onKeyUp )

        self.frame:SetScript( "OnKeyDown", function( f, key )
            if self[key] then
                f:SetPropagateKeyboardInput( false )
                self[key]( self )
            else
                f:SetPropagateKeyboardInput( true )
            end
        end )

        self.frame:EnableKeyboard( true )
    end

    zu.AddListener( self.model.updateEvent, self, update )
    zu.AddListener( self.model.modelResetEvent, self, modelReset )
end

function ListSelectionController:release()
    zu.RemoveListener( self.model.updateEvent, self )
end

local function getRelativeMousePosition( self )
    local view = self.view

    local mouseX, mouseY = GetCursorPosition()
    local scale = self.frame:GetEffectiveScale()

    local x, y, _, h = self.frame:GetScaledRect()

    x = floor( ( mouseX - x ) / scale )
    y = floor( ( h + y - mouseY ) / scale )

    x = max( x, 0 )
    y = max( y, 0 )
    x = min( x, self.view.width )
    y = min( y, self.view.height )

    y = y + view.scrollPosition

    return x, y
end

function ListSelectionController:onMouseDown()
    local _, y = getRelativeMousePosition( self )
    local itemToSelect = floor( y / self.view.lineHeight ) + 1
    itemToSelect = min( itemToSelect, self.model:count() )
    self:selectElement( itemToSelect )
end

function ListSelectionController:onMouseUp()
    if self.mouseUp then
        self:mouseUp()
    end
end

function ListSelectionController:hoverUpdate()
    local _, mouseY = GetCursorPosition()
    if mouseY == self.lasty then
        return
    end
    self.lasty = mouseY

    local _, y = getRelativeMousePosition( self )
    local itemIndex = floor( y / self.view.lineHeight ) + 1
    itemIndex = min( itemIndex, self.model:count() )

    if self.mouseHoverUpdate and self:mouseHoverUpdate( true ) then
        itemIndex = 0
    end

    if self.frozen then
        itemIndex = 0
    end

    if itemIndex ~= self.highlightedItemIndex then
        self.highlightedItemIndex = itemIndex
        self.highlightedItem = self.model:getElement( self.highlightedItemIndex )
        zu.Raise( self.highlightChangedEvent )
    end
end

local function startKeyRepetition( self, func )
    if self.repeatingFunc == func then
        return
    end

    self.repeatingFunc = func

    if self.repeater then
        self.repeater:Cancel()
    end

    local function repeatKey()
        self.repeatingFunc( self )
    end

    local function initialRepeatKey()
        self.repeatingFunc( self )
        self.repeater = C_Timer.NewTicker( 0.025, repeatKey )
    end

    self.repeater = C_Timer.NewTimer( 0.5, initialRepeatKey )
end

function ListSelectionController:UP()
    local newHighlightItemIndex = self.highlightedItemIndex - 1
    local newHighlightItem = self.model:getElement( newHighlightItemIndex )
    if not newHighlightItem then
        return
    end
    self.highlightedItemIndex = newHighlightItemIndex
    self.highlightedItem = self.model:getElement( newHighlightItemIndex )
    scrollToHighlighted( self )
    zu.Raise( self.highlightChangedEvent )
    startKeyRepetition( self, self.UP )
end

function ListSelectionController:DOWN()
    local newHighlightItemIndex = self.highlightedItemIndex + 1
    local newHighlightItem = self.model:getElement( newHighlightItemIndex )
    if not newHighlightItem then
        return
    end
    self.highlightedItemIndex = newHighlightItemIndex
    self.highlightedItem = self.model:getElement( newHighlightItemIndex )
    scrollToHighlighted( self )
    zu.Raise( self.highlightChangedEvent )
    startKeyRepetition( self, self.DOWN )
end

function ListSelectionController:ENTER()
    if self.highlightedItemIndex > 0 then
        self:selectElement( self.highlightedItemIndex )
    else
        self.frame:SetPropagateKeyboardInput( true )
    end
end

function ListSelectionController:createElement( ... )
    local element = self.model:createElement( ... )
    self:selectElement( self.model:getElementIndex( element ), true )
end

function ListSelectionController:deleteElement( element )
    local selChanged = false

    local index = self.model:getElementIndex( element )
    local count = self.model:count()

    if index == self.selectedItemIndex then
        self.selectedItemIndex = index < count and min( self.selectedItemIndex + 1, count ) or ( count - 1 )
        self.selectedItem = self.model:getElement( self.selectedItemIndex )
        selChanged = true
    end

    self.model:deleteElement( element )
    self.selectedItemIndex = self.model:getElementIndex( self.selectedItem )

    if selChanged then
        zu.Raise( self.selectionChangedEvent, self.selectedItem, false )
    end
end

function ListSelectionController:selectElement( index, isNew )
    if isNew == nil then
        isNew = false
    end

    if index ~= self.selectedItemIndex then
        self.selectedItemIndex = index
        self.selectedItem = self.model:getElement( self.selectedItemIndex )
        zu.Raise( self.selectionChangedEvent, self.selectedItem, isNew )
    end

    scrollToSelection( self )
end

function ListSelectionController:highlightElement( index )
    if index ~= self.highlightedItemIndex then
        self.highlightedItemIndex = index
        self.highlightedItem = self.model:getElement( self.highlightedItemIndex )
        zu.Raise( self.highlightChangedEvent )
    end

    scrollToSelection( self )
end

function ListSelectionController:freeze()
    self.frozen = true
    self:hoverUpdate()
end

function ListSelectionController:unfreeze()
    self.frozen = false

    if self.view.frame:IsMouseOver() then
        self:hoverUpdate()
    end
end
