--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
-- A scrollable list of elements of fixed height.
-- This is just the list's content widget, it doesn't actually
-- provide a scrolling control. The List widget wraps this along
-- with a slider in an horizontal group.
--
-- The elements are provided by a data model, which
-- is a table that should contain the following functions:
--   model:count() - returns the number of elements in the list
--   model:getWidget( index ) - returns the widget representing the element i of the list
--   note: getWidget( nil ) should return a dummy widget instance that will be
--   used for layout constraints. All elements should have the same height that it has,
--   and the same min/max/preferred width.
--
-- All elements should have the same layout constraints as the first element
-- (height, min/max/preferred width) for performance reasons.
-- it allows to set the scrolling position arbitrarily anywhere in the list and
-- determine which elements are visible efficiently.
--
-- Generating widgets that can't fit in the provided constraints will lead
-- to undefined results (ie it will be a clusterfuck). Also avoid using
-- widget that can trade width for height, ie flow layouts.
--
-- zu.CreateWidget() and zu.ReleaseWidget() should be used to create the widgets
-- (this will recycle previously used widgets and should avoid allocating memory
-- just to scroll the list)
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local min = math.min
local max = math.max
local floor = math.floor
local ceil = math.ceil

local ListViewClass = {}
zu.ListViewClass = ListViewClass
setmetatable( ListViewClass, { __index = zu.Widget } )

ListViewClass.spacing = 1
ListViewClass.firstDisplayedIndex = 1
ListViewClass.numDisplayedItems = 0
ListViewClass.scrollPosition = 0
ListViewClass.minDisplayedItems = 3
ListViewClass.preferredDisplayedItems = 10

function ListViewClass:new( tpl )
    local o = {}
    o.updateEvent = {}
    o.displayedWidgets = {}
    o.spareWidgets = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function zu.ListView( s )
    return ListViewClass:new( s )
end

local function updateContent( self )
    if not self.height then
        return
    end

    local th = self:getTotalHeight()
    local itemHeight = self.itemHeight
    local lineHeight = self.lineHeight

    local scrollPosition = self.scrollPosition
    scrollPosition = min( scrollPosition, th - self.height )
    scrollPosition = max( scrollPosition, 0 )
    self.scrollPosition = scrollPosition

    local count = self.model:count()

    local first = floor( scrollPosition / lineHeight )
    local offset = scrollPosition - ( first * lineHeight )

    local numDisplayedItems = ceil( ( self.height + offset ) / lineHeight )
    numDisplayedItems = min( numDisplayedItems, count - first )

    -- All the current line widgets are now considered spare widgets:
    -- if we encounter their line again during the update, we'll reuse
    -- them. Any that hasn't been reused after the update is released.
    local spareWidgets = self.displayedWidgets

    -- Initialize the new widget table with the old spare table, which
    -- should be empty.
    local widgets = self.spareWidgets

    self.firstDisplayedIndex = first
    self.numDisplayedItems = numDisplayedItems
    local controller = self.controller

    local width = self.width
    local y = -offset

    for i=1, numDisplayedItems, 1 do
        local item = self.model:getElement( i + first )
        local w = spareWidgets[item]

        if w then
            widgets[item] = w
            spareWidgets[item] = nil
        else
            -- create new widget
            w = self.model:getWidget( i + first )
            w.parent = self

            w:create( self.frame )

            if w.setController then
                w:setController( controller )
            end

            w:show()
            widgets[item] = w
        end

        w:calcMinMax()

        w:heightForWidth( width )
        if self.level then
            w:setLevel( self.level + 1 )
        end

        w:layout( 0, y, width, itemHeight )
        y = y + lineHeight
    end

    for _,w in pairs( spareWidgets ) do
        zu.ReleaseWidget( w )
    end
    wipe( spareWidgets )

    self.displayedWidgets = widgets
    self.spareWidgets = spareWidgets
end

local function onModelUpdated( self )
    if self.minimizeHeight then
        self:relayout()
    else
        updateContent( self )
    end
    zu.Raise( self.updateEvent )
end

local function onModelReset( self )
    -- Release all the wigdets and recreate them all
    for _, w in pairs( self.displayedWidgets ) do
        zu.ReleaseWidget( w )
    end
    wipe( self.displayedWidgets )

    onModelUpdated( self )
end

function ListViewClass:release()
    local updateEvent = self.model.updateEvent
    if updateEvent then
        zu.RemoveListener( updateEvent, self )
    end

    local modelResetEvent = self.model.modelResetEvent
    if modelResetEvent then
        zu.RemoveListener( modelResetEvent, self )
    end
end

function ListViewClass:create( parentFrame )
    if not self.decorator then
        self.decorator = self:getSkin().ListViewDecorator:new( self )
    end

    if self.frame then
        self.frame:SetParent( parentFrame )
    else
        self.frame = CreateFrame( "Frame", nil, parentFrame )
        self.frame:SetClipsChildren( true )

        local updateEvent = self.model.updateEvent
        if updateEvent then
            zu.AddListener( updateEvent, self, onModelUpdated )
        end

        local modelResetEvent = self.model.modelResetEvent
        if modelResetEvent then
            zu.AddListener( modelResetEvent, self, onModelReset )
        end

        self.frame:EnableMouseWheel( true )
        self.frame:SetScript( "OnMouseWheel", function( _, direction ) self:onMouseWheel( direction ) end )

        self.lineHeight = 1;

        self.decorator:apply()
    end
end

function ListViewClass:show()
    for _,w in pairs( self.displayedWidgets ) do
        w:show()
    end
    zu.Widget.show( self )
end

function ListViewClass:hide()
    for _,w in pairs( self.displayedWidgets ) do
        w:hide()
    end
    zu.Widget.hide( self )
end

function ListViewClass:setController( controller )
    self.controller = controller
    zu.RemoveListener( self.model.updateEvent, self )
    zu.AddListener( self.controller.updateEvent, self, onModelUpdated )
end

function ListViewClass:calcMinMax()
    local spacing = zu.Scale( self.spacing )

    local itemHeight, itemMinWidth, itemPreferredWidth, itemMaxWidth, levelRange
        = self.model:getElementSize( self )

    self.itemHeight = itemHeight
    self.lineHeight = itemHeight + spacing

    self.minWidth = itemMinWidth
    self.maxWidth = itemMaxWidth
    self.preferredWidth = itemPreferredWidth
    self.levelRange = levelRange + 1

    self.minHeight = floor( ( self.itemHeight + spacing ) * self.minDisplayedItems - spacing + 0.5 )
    self.preferredHeight = floor( ( self.itemHeight + spacing ) * self.preferredDisplayedItems - spacing + 0.5 )

    if self.minimizeHeight then
        self.maxHeight = floor( ( self.itemHeight + spacing ) * max( self.model:count(), 1 ) - spacing + 0.5 )
        self.minHeight = min( self.minHeight, self.maxHeight )
        self.preferredHeight = min( self.preferredHeight, self.maxHeight )
    end
end

function ListViewClass:setLevel( level )
    zu.Widget.setLevel( self, level )

    self.frame:SetFrameLevel( level )
    level = level + 1

    for _,w in pairs( self.displayedWidgets ) do
        w:setLevel( level )
    end
end

function ListViewClass:layout( x, y, width, height )
    zu.Widget.layout( self, x, y, width, height )
    self.frame:SetSize( width, height )
    self.frame:SetPoint( "TOPLEFT", self.frame:GetParent(), "TOPLEFT", x, -y )
    updateContent( self )
end

function ListViewClass:getTotalHeight()
    local count = self.model:count()
    if count == 0 then
        return 0
    end

    return floor( count * self.lineHeight - self.spacing + 0.5 )
end

function ListViewClass:setScrollPosition( pos )
    self.scrollPosition = pos
    updateContent( self )
    zu.Raise( self.updateEvent )
end

function ListViewClass:onMouseWheel( direction )
    local vy = self.scrollPosition

    -- The scrolling amount is hardcoded for now.
    vy = vy - direction * 2 * self.lineHeight

    vy = max( vy, 0 )
    vy = min( vy, self:getTotalHeight() )

    if vy ~= self.scrollPosition then
        self:setScrollPosition( vy )
    end
end
