--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local RichTextClass = {}
zu.RichTextClass = RichTextClass
setmetatable( RichTextClass, { __index = zu.CompositeClass } )

-- Helper fontstring just to retrieve the width of the space character...
local fontStringHelper = UIParent:CreateFontString()
fontStringHelper:Hide()
fontStringHelper:SetScale( zu.ContentScaleFactor() )
fontStringHelper:SetWordWrap( false )
fontStringHelper:SetNonSpaceWrap( false )

function RichTextClass:new( tpl )
    local o = {}
    setmetatable( tpl, { __index = self } )
    setmetatable( o, { __index = tpl } )
    return o
end

function zu.RichText( tpl )
    return RichTextClass:new( tpl )
end

function RichTextClass:release()
    if self.content then
        zu.ReleaseWidget( self.content )
        self.content = nil
    end
end

-- Enforce a minimal width that isn't too small, as it
-- makes justified text look ugly.
function RichTextClass:calcMinMax()
    zu.CompositeClass.calcMinMax( self )
    self.minWidth = max( self.minWidth, zu.Scale( 200 ) )
end

local inlineTags
local tagHandlers

local function handleTag( self, p, tag )
    local th = tagHandlers[tag]
    if not th then
        return
    end

    return th( self, p )
end

local spcTpl = {}
local function marginWrap( self, widget )
    if not self.currentMargin then
        return widget
    end

    local leftMargin = zu.CreateWidget( zu.SpacerClass, spcTpl )
    local rightMargin = zu.CreateWidget( zu.SpacerClass, spcTpl )

    leftMargin.minH = 1
    leftMargin.maxH = 1
    leftMargin.minW = self.currentMargin
    leftMargin.maxW = self.currentMargin

    rightMargin.minH = 1
    rightMargin.maxH = 1
    rightMargin.minW = self.currentMargin
    rightMargin.maxW = self.currentMargin

    return zu.CreateWidget( zu.GridLayoutClass,
    {
        numColumns = 0, hjustify="left",
        leftMargin, widget, rightMargin
    } )
end

local icLabelTpl = {}

local function inlineCodeHandler( self, p )
    icLabelTpl.font = zu.EditorClass.font
    local label = zu.CreateWidget( zu.MinimizeClass,
    {
        content = zu.CreateWidget( zu.LabelClass, icLabelTpl )
    } )

    while true do
        local t = p( true )
        if not t then
            break
        end

        if t[1] == zu.RichTextToken.Text then
            local text = self.codeColorizer and self.codeColorizer( t[2] ) or t[2]
            label.content:setLabel( text )
        elseif t[1] == zu.RichTextToken.TagEnd then
            break
        end
    end

    return label
end

local tvTpl = {}
local function codeBlockHandler( self, p )
    local inlineTextView = zu.CreateWidget( zu.InlineTextViewClass, tvTpl )
    inlineTextView.colorizer = self.codeColorizer

    while true do
        local t = p( true )
        if not t then
            break
        end

        if t[1] == zu.RichTextToken.Text then
            inlineTextView:setText( t[2] )
        elseif t[1] == zu.RichTextToken.TagEnd then
            break
        end
    end

    return marginWrap( self, inlineTextView )
end

local labelTpl = {}
local function buildWidgetsForText( self, p )
    local root = zu.CreateWidget( zu.GridLayoutClass, { numColumns = 1, greedyMaxWidth=true, hjustify="left" } )
    local paragraph

    local function beginParagraphIfNeeded()
        if not paragraph then
            paragraph = zu.CreateWidget( zu.FlowLayoutClass,
            {
                hspacing = self.whiteSpaceSize,
                vspacing = 0,
                justify = "justified"
            } )
        end
    end

    local function flushParagraph()
        if paragraph then
            root.tpl[#root.tpl+1] = marginWrap( self, paragraph )
            paragraph = nil
        else
            -- if the paragraph was empty, don't insert an empty flow layout, it won't work properly.
            -- Instead, insert a spacer for the line's height (so we can skip lines)
            -- TODO: lineHeight is prolly already scaled so add a flag to spacer to tell it not to scale
            -- TODO: needs that to react to relayout too sigh
            local spacer = zu.CreateWidget( zu.SpacerClass, spcTpl )
            spacer.minH = self.lineHeight
            spacer.maxH = self.lineHeight
            spacer.minW = 1
            spacer.maxW = 1
            root.tpl[#root.tpl+1] = spacer
        end
    end

    while true do
        local t = p( false )
        if not t then
            break
        end

        if t[1] == zu.RichTextToken.Text then
            beginParagraphIfNeeded()

            local label = zu.CreateWidget( zu.MinimizeClass,
            {
                content = zu.CreateWidget( zu.LabelClass, labelTpl )
            } )
            label.content:setLabel( t[2] )

            paragraph.tpl[#paragraph.tpl+1] = label
        elseif t[1] == zu.RichTextToken.NewLine then
            flushParagraph()
        elseif t[1] == zu.RichTextToken.TagStart then
            if inlineTags[t[2]] then
                beginParagraphIfNeeded()
                paragraph.tpl[#paragraph.tpl+1] = handleTag( self, p, t[2] )
            else
                if paragraph then
                    flushParagraph()
                end
                root.tpl[#root.tpl+1] = handleTag( self, p, t[2] )
            end
        elseif t[1] == zu.RichTextToken.TagEnd then
            break
        end
    end

    if paragraph then
        flushParagraph()
    end
    return root
end

local function marginBlockHandler( self, p )
    self.currentMargin = zu.Scale( 16 )
    local result = buildWidgetsForText( self, p )
    self.currentMargin = nil
    return result
end

inlineTags =
{
    c=1
}

tagHandlers =
{
    c = inlineCodeHandler,
    code = codeBlockHandler,
    margin = marginBlockHandler
}

local function buildWidgets( self )
    fontStringHelper:SetFontObject( self:getSkin().fontNormal )
    fontStringHelper:SetText( " " )
    self.whiteSpaceSize = floor( fontStringHelper:GetUnboundedStringWidth() )
    self.lineHeight = floor( fontStringHelper:GetLineHeight() )
    return buildWidgetsForText( self, zu.RichTextParser( self.text ) )
end

function RichTextClass:create( parentFrame )
    if not self.content then
        self.parentFrame = parentFrame
        self.content = buildWidgets( self )
        self.content.parent = self
    end

    self.content:create( parentFrame )
end

function RichTextClass:setText( text )
    self.text = text

    if self.content then
        if self.content then
            zu.ReleaseWidget( self.content )
            self.content = nil
        end

        self.content = buildWidgets( self )
        self.content.parent = self
        self.content:create( self.parentFrame )

        self:relayout()
    end
end
