--
-- Zlodo's UI framework
-- Author: Zlodo (Taliriah-Silvermoon EU)
--
local _, addon = ...
addon.zu = addon.zu or {}
local zu = addon.zu

local floor = math.floor

local WindowClass = {}
WindowClass.strata = "DIALOG"
WindowClass.skin = zu.DefaultSkin

function zu.Window( w )
    setmetatable( w, { __index = WindowClass } )

    -- Window can only have one child.
    w.content = w[1]

    return w
end

local function calcMinMax( self )
    local decorator = self.decorator
    decorator:calcMinMax()

    self.preferredWidth = decorator.preferredWidth
    self.preferredHeight = decorator.preferredHeight

    self.minWidth = decorator.minWidth
    self.minHeight = decorator.minHeight

    if decorator.maxWidth then
        self.maxWidth = decorator.maxWidth
    else
        self.maxWidth = nil
    end

    if decorator.maxHeight then
        self.maxHeight = decorator.maxHeight
    else
        self.maxHeight = nil
    end
end

local function createDecorator( self, decClass )
    if self.decorator then
        zu.ReleaseWidget( self.decorator )
    end

    self.decorator = zu.CreateWidget( decClass, ( { client = self.content } ) )
    self.decorator.parent = self
    self.decorator:create( self )
end

local function resetDefaultSize( self )
    -- Clamp our default size to our min/max range
    -- so that we don't end up with controls that won't fit
    -- if someone gives a preferred size outside of its own min/max range
    local width = self.preferredWidth
    local height = self.preferredHeight

    if self.settings then
        width = self.settings.width or width
        height = self.settings.height or height
    end

    width = math.max( width, self.minWidth )
    height = math.max( height, self.minHeight )

    if self.maxWidth then
        width = math.min( width, self.maxWidth )
    end

    if self.maxHeight then
        height = math.min( height, self.maxHeight )
    end

    if self.hfw then
        height = max( height, self.hfw )
    end

    if self.frame then
        self.frame:SetSize( width, height )
    end
end

function WindowClass:resetDefaultSize()
    resetDefaultSize( self )
    local width = self.frame:GetWidth()
    local height = self.frame:GetHeight()
    self:layout( width, height )
end

local function bringToFront( self )
    if not self.frame then
        return
    end

    self.frame:Raise()

    self.decorator:setLevel( self.frame:GetFrameLevel() )
    self.frontFrame:SetFrameLevel( self.frame:GetFrameLevel() + self.decorator.levelRange + 1 )
end

local function onKeyDown( self, key )
    if key == "ESCAPE" then
        self.frontFrame:SetPropagateKeyboardInput( false )
        self:hide()
    else
        self.frontFrame:SetPropagateKeyboardInput( true )
    end
end

function WindowClass:create()
    self.skin:init()

    if self.name then
        self.settings = zu.Settings[self.name] or {}
        zu.Settings[self.name] = self.settings
    end

    self.frame = CreateFrame( "Frame" )

    local anchor = self.settings and self.settings.anchor or nil
    if anchor then
        self.frame:SetPoint( anchor, nil, anchor, self.settings.x, self.settings.y )
    else
        self.frame:SetPoint( "CENTER", nil, "CENTER", 0, 0 )
    end

    self.frame:SetMovable( true )
    self.frame:SetResizable( true )
    self.frame:SetFrameStrata( self.strata )
    self.frame:SetScale( zu.ScaleFactor() )

    local decClass = self.decoratorClass or self.skin.WindowDecorator
    createDecorator( self, decClass )

    calcMinMax( self )

    self.frame:SetScript( "OnSizeChanged", function( _, w, h ) self:sizeChanged( w, h ) end )
    self.frame:SetScript( "OnUpdate", function() self:onUpdate() end )
    self.frame:EnableMouse( true )

    -- Create an invisible frame that we'll use to block mouse events when freezing the window
    self.frontFrame = CreateFrame( "Frame", nil, self.frame )
    self.frontFrame:SetAllPoints( self.frame )
    self.frontFrame:SetFrameStrata( self.strata )

    self.frontFrame:SetScript( "OnKeyDown", function( _, key ) onKeyDown( self, key ) end )
    self.frontFrame:EnableKeyboard( true )

    self.frame:RegisterEvent( "ADDONS_UNLOADING" )
    self.frame:RegisterEvent( "DISPLAY_SIZE_CHANGED" )
    self.frame:RegisterEvent( "UI_SCALE_CHANGED" )

    self.frame:SetScript( "OnEvent", function( _, e )
        self[e]( self )
    end )

    resetDefaultSize( self )
    bringToFront( self )
end

function WindowClass:ADDONS_UNLOADING()
    if self.settings == nil then
        return
    end

    local _, _, a, xpos, ypos = self.frame:GetPoint( 1 )
    self.settings.anchor = a
    self.settings.x = floor( xpos + 0.5 )
    self.settings.y = floor( ypos + 0.5 )
end

function WindowClass:DISPLAY_SIZE_CHANGED()
    -- TODO: resize the window in proportion to the size change, also make sure
    -- its bounds are still within the screen
    self.frame:SetScale( zu.ScaleFactor() )
    self:relayout( true )
end

function WindowClass:UI_SCALE_CHANGED()
    -- TODO: same as above, but in proportion to the effective scale change
    self.frame:SetScale( zu.ScaleFactor() )
    self:relayout( true )
end

local function heightForWidth( self, w )
    w = floor( w + 0.5 )
    w = max( w, self.minWidth )

    if self.maxWidth then
        w = min( w, self.maxWidth )
    end

    -- Check if our content wants a specific height for
    -- the given width
    self.hfw = self.decorator:heightForWidth( w )
    return self.hfw
end

-- Wow sometimes calls this a bunch of times during the same
-- engine frame (with different sizes, depending on how fast the
-- mouse is moving), so recomputing the layout here can result
-- in awful frame drops when resizing the window when moving
-- the mouse too fast.
-- Instead, we just record the requested size and perform the
-- relayout once for the engine frame in OnUpdate.
function WindowClass:sizeChanged( w, h )
    self.newW = w
    self.newH = h
    self.needlayout = true
end

function WindowClass:onUpdate()
    if self.needlayout then
        local w = floor( self.newW + 0.5 )
        if self.maxWidth then
            w = min( w, self.maxWidth )
        end

        local hfw = heightForWidth( self, w )
        if hfw and self.newH <= hfw then
            self.frame:SetHeight( hfw )
            self.newH = hfw
        end

        self:layout( self.newW, self.newH )
        self.needlayout = false
    end
end

function WindowClass:layout( w, h )
    w = floor( w + 0.5 )
    h = floor( h + 0.5 )

    w = max( w, self.minWidth )

    if self.maxWidth then
        w = min( w, self.maxWidth )
    end

    local minHeight = self.minHeight
    local maxHeight = self.maxHeight

    local decorator = self.decorator

    -- Check if our content wants a specific height for
    -- the given width
    local hfw = self.hfw

    -- If we have a heightForWidth from our content,
    -- Ensure that our min size it at least that.
    -- In addition, if we have a max size, override
    -- it with the hfw.
    if hfw then
        minHeight = math.max( minHeight, hfw )

        if maxHeight then
            maxHeight = hfw
        end
    end

    self.frame:SetResizeBounds( self.minWidth, minHeight )

    h = max( h, minHeight )

    if maxHeight then
        h = min( h, maxHeight )
    end

    decorator:layout( 0, 0, w, h )

    if self.settings then
        self.settings.width = w
        self.settings.height = h
    end
end

function WindowClass:show()
    if not self.frame then
        self:create()
    end

    self.frame:Show()
    self.decorator:show()
    bringToFront( self )
end

function WindowClass:hide()
    if self.overlay then
        return
    end

    if self.frame then
        self.frame:Hide()
        self.decorator:hide()
    end
end

function WindowClass:relayout( force )
    if not self.frame then
        return
    end

    calcMinMax( self )

    local oldWidth = self.frame:GetWidth()
    local width = oldWidth
    width = max( width, self.minWidth )

    if self.maxWidth then
        width = min( width, self.maxWidth )
    end

    local hfw = heightForWidth( self, width )

    local oldHeight = self.frame:GetHeight()
    local height = oldHeight
    height = max( height, self.minHeight )

    if self.maxHeight then
        height = min( height, self.maxHeight )
    end

    if hfw then
        height = max( height, self.hfw )
    end

    if not force and oldWidth ~= width or oldHeight ~= height then
        self.frame:SetSize( width, height )
    else
        self:layout( width, height )
    end

    bringToFront( self )
end

function WindowClass:setDecoratorClass( decClass )
    self.decoratorClass = decClass

    if self.frame then
        createDecorator( self, decClass )
    end
end

function WindowClass:setContent( widget )
    self.content = widget

    if not self.frame then
        return
    end

    widget:create( self.frame )

    if self.overlay then
        self.overlayLayout[3] = widget
        widget.parent = self.overlayLayout
    else
        widget.parent = self
        self.decorator:setClient( widget )
    end

    calcMinMax( self )
    local width = self.frame:GetWidth()
    heightForWidth( self, width )
    resetDefaultSize( self )
    bringToFront( self )

    local height = self.frame:GetHeight()
    self:layout( width, height )
end

function WindowClass:setOverlay( widget )
    if self.overlay == widget then
        return
    end

    if self.overlay then
        self.overlay:hide()
        self.overlayTpl[2]:hide()
    end

    self.overlay = widget
    if not self.overlayLayout then
        self.overlayTpl = { {}, zu.Filler{ 0, 0, 0, 0.6, shouldEatMouseEvents = true }, {} }
        self.overlayLayout = zu.OverlayLayout( self.overlayTpl )
        self.overlayLayout.parent = self
    end

    self.content.parent = widget and self.overlayLayout or self

    if widget then
        self.overlayTpl[1] = widget
        self.overlayTpl[3] = self.content
        self.overlayLayout:create( self.frame )
        self.overlayLayout:show()
        self.content:freeze()
        self.decorator:setClient( self.overlayLayout )
    else
        self.content:unfreeze()
        self.decorator:setClient( self.content )
    end

    calcMinMax( self )
    heightForWidth( self, self.frame:GetWidth() )
    local width = self.frame:GetWidth()
    local height = self.frame:GetHeight()
    self:layout( width, height )
    bringToFront( self )
end

function WindowClass:startSizing( direction )
    self:freeze()
    self.frame:StartSizing( direction )
end

function WindowClass:stopMovingOrSizing()
    self.frame:StopMovingOrSizing()
    self:unfreeze()
end

function WindowClass:setPosition( x, y, frame, pointToAttach, pointToAttachTo )
    if self.frame then
        pointToAttach = pointToAttach or "TOPLEFT"
        pointToAttachTo = pointToAttachTo or "TOPLEFT"
        self.frame:ClearAllPoints()
        self.frame:SetPoint( pointToAttach, frame, pointToAttachTo, x, -y )
    end
end

function WindowClass:getSkin()
    return self.skin
end

function WindowClass:freeze()
    if self.overlay then
        self.overlay:freeze()
    else
        self.content:freeze()
    end

    self.frontFrame:EnableMouse( true )
end

function WindowClass:unfreeze()
    if self.overlay then
        self.overlay:unfreeze()
    else
        self.content:unfreeze()
    end

    self.frontFrame:EnableMouse( false )
end

function WindowClass:isMouseOver()
    return self.frontFrame and self.frontFrame:IsMouseOver()
end

function WindowClass:about()
    if self.onAbout and not self.overlay then
        self.onAbout()
    end
end
